/*
 * @(#)DataFile.java
 *
 * Copyright (c) 1998-1999 by Trapezium Development LLC.  All Rights Reserved.
 *
 * The information in this file is the property of Trapezium Development LLC
 * and may be used only in accordance with the terms of the license granted
 * by Trapezium.
 */
package com.trapezium.browser;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.URL;
import java.net.URLConnection;
import java.util.*;
import java.util.zip.*;

import com.trapezium.factory.FactoryData;
import com.trapezium.factory.FactoryDataUpdater;
import com.trapezium.factory.FactoryChain;
import com.trapezium.factory.QueuedRequestFactory;
import com.trapezium.factory.TokenStreamFactory;
import com.trapezium.factory.FactoryResponseListener;
import com.trapezium.factory.ParserFactory;
import com.trapezium.parse.InputStreamFactory;
import com.trapezium.util.GlobalProgressIndicator;
import com.trapezium.util.ProgressIndicator;
import com.trapezium.util.RemoteUrlGenerator;
import com.trapezium.vrml.*;
import com.trapezium.vrml.grammar.*;
import com.trapezium.edit.TokenEditor;
import com.trapezium.edit.EditLintVisitor;
import com.trapezium.parse.TokenEnumerator;
import com.trapezium.html.HTMLgenerator;

import com.trapezium.edit.Document;

/** a DataFile object is a FactoryData object that knows how to load
    itself, reload itself from its original source, and save itself.
    It also knows how to retrieve other data from the same source, */
public class DataFile extends FactoryData implements Serializable, Saveable {

    Document doc;

    /** way to create the data given the source */
    FactoryChain loader;

    // for the moment, ParserFactory is not thread-safe, only have one for entire system
    QueuedRequestFactory parser;
    QueuedRequestFactory tokenFactory;

    public QueuedRequestFactory getParserFactory() {
        return( parser );
    }

    public QueuedRequestFactory getTokenFactory() {
        return( tokenFactory );
    }

    public DataFile() {
        this(null);
    }

    static int id = 1;
    int myId;

    boolean newlyOpened;
    public boolean isNewlyOpened() {
        return( newlyOpened );
    }

    public void notNewlyOpened() {
        newlyOpened = false;
    }


    int versionNumber;

    /**  construct a DataFile, load it and parse it */
    public DataFile(String source) {
        myId = id;
        newlyOpened = true;
        id++;
        versionNumber = 1;
        setUrl(source);

        // for now, loader is FactoryChain with no ProgressIndicator
        loader = new FactoryChain( null );  
        tokenFactory = new TokenStreamFactory();
        parser = new ParserFactory();
        loader.addFactory(tokenFactory);
        loader.addFactory(parser);
    }

    /** Get the version number for the file */
    public int getVersion() {
        return( versionNumber );
    }

    /** Get the version number as a String */
    public String getVersionString() {
        return( "v" + versionNumber + ":  " );
    }

    /** Set the version number for the file */
    public void setVersion( int versionNo ) {
        versionNumber = versionNo;
        Scene scene = getScene();
        if ( scene != null ) {
            GlobalProgressIndicator.setProgressIndicator( FactoryDataUpdater.getDefaultProgressIndicator(), "Serializing...", scene.getVrmlElementCount() );
            getDocument().setViewerVersion( versionNo, scene );
        }
    }

    /** set the name of the local file for storing this object as a String */
    public void setLocalName(String name) {
        setFile(new File(name));
    }
        

    /** save the object locally */
    public void save(int method) {
        switch (method) {
            case Saveable.TEXT:
                asciiSave(getFile(), false);
                break;
            case Saveable.GZIP:
                asciiSave(getFile(), true);
                break;
            case Saveable.JAR:
                // not yet implemented
                break;
            default:
                throw new IllegalArgumentException("Unrecognized save method.");
        }
    }

   
    public void initProgressIndicator( String title ) {
        Scene scene = getScene();
        GlobalProgressIndicator.setProgressIndicator( FactoryDataUpdater.getDefaultProgressIndicator(), title, scene.getVrmlElementCount() );
    }


    /** DocumentLoader interface */
    public void reload( Object scene ) {
        System.out.println( "Reloading scene..." );
        if ( scene instanceof Scene ) {
            Scene s = (Scene)scene;
            TokenEnumerator te = s.getTokenEnumerator();
            if ( te instanceof TokenEditor ) {
                getDocument().setLines( (TokenEditor)te );
            }
            setScene( s );
        }
    }

    FactoryResponseListener parsingListener = null;
    public void setParsingListener( FactoryResponseListener f ) {
        parsingListener = f;
        tokenFactory.setListener( f );
        parser.addListener( f );
    }

    public void mergeErrors( BitSet errorMarks, BitSet warningMarks, BitSet nonconformanceMarks ) {
        mergeLintErrors( parsingListener, errorMarks, warningMarks, nonconformanceMarks );
    }

    /** Set who is going to be called when each stage of the loader is done */
    public void setDoneListener( FactoryResponseListener f ) {
        loader.addListener( f );    // should be safe, since it rejects duplicates
    }


    public int getId() {
        return( myId );
    }

    /** Generate local name for file */
    public String generateName() {
        return( generateName( false ));
    }

    /** Generate a name for file.
     *
     *  @param stripNumber strip off "_<N>" portion of name.
     */
    public String generateName( boolean stripNumber ) {
        File file = getFile();
        if ( file != null ) {
            String name = file.getName();
            if (( saveAsCounter > 1 ) && stripNumber ) {
                int underscoreIdx = name.lastIndexOf( "_" );
                int dotIdx = name.lastIndexOf( "." );
                if (( underscoreIdx != -1 ) && ( dotIdx != -1 ) && ( underscoreIdx < dotIdx )) {
                    String beforeUnderscore = name.substring( 0, underscoreIdx );
                    String dotAndAfter = name.substring( dotIdx );
                    return( beforeUnderscore + dotAndAfter );
                }
            }
            return( name );
        } else {
            return( "unknown.wrl" );
        }
    }

    public String getLocalName() {
        File file = getFile();
        if (file != null) {
            return file.getName();
        } else {
            return null;
        }
    }

    /** generate a name for "save as", just the file name, with an appended
     *  save as counter (local to the processed file).
     */
    int saveAsCounter = 1;
    public String generateSaveAsName() {
        String name = generateName( true );
        if ( name.indexOf( "." ) == -1 ) {
            name = name + ".wrl";
        }
        String beforeDot = name.substring( 0, name.indexOf( "." ));
        String dotAndAfter = name.substring( name.indexOf( "." ));
        saveAsCounter++;
        String newName = beforeDot + "_" + saveAsCounter + dotAndAfter;
        return( newName );
    }

    public void addFactory( QueuedRequestFactory f ) {
        loader.addFactory( f );
    }

    public void addFactory( QueuedRequestFactory f, QueuedRequestFactory prev ) {
        loader.insertFactory( f, prev );
        //load();
    }

    public void insertFactory( QueuedRequestFactory f ) {
        loader.insertFactory( f, tokenFactory );
    }

    public String getBaseFilePath() {
        String name = getUrl();
        if ( name != null ) {
            if ( name.lastIndexOf( '/' ) != -1 ) {
                name = name.substring( 0, name.lastIndexOf( '/' ));
            } else if ( name.lastIndexOf( '\\' ) != -1 ) {
                name = name.substring( 0, name.lastIndexOf( '\\' ));
            } else {
                name = null;
            }
        }
        return( name );

    }

    public String getNameWithoutPath() {
        String name = getUrl();
        if ( name != null ) {
            if ( name.lastIndexOf( '/' ) != -1 ) {
                return( name.substring( name.lastIndexOf( '/' ) + 1 ));
            } else if ( name.lastIndexOf( '\\' ) != -1 ) {
                return( name.substring( name.lastIndexOf( '\\' ) + 1 ));
            } else {
                return( name );
            }
        }
        return( null );
    }

    public String getNameWithoutVersion() {
        String name = getUrl();
        File file = getFile();
        if (file != null) {
            name = file.getName();
        }
        return name;
    }

    public String getName() {
        String name = getNameWithoutVersion();
        return getVersionString() + name;
    }

    /** Get the Document associated with this data */
    public Document getDocument() {
        return( doc );
    }

    /** Set the Document associated with this data */
    public void setDocument( Document doc ) {
        this.doc = doc;
        if ( doc != null ) {
            doc.setNodeSelector( getScene() );
        }
    }

    /** kick off the loader chain of processing */
    Date startTime;
    boolean started = false;
    public void load() {
        started = true;
        load( true );
    }

    /** Start running loader factories.
     *
     *  @param resetStartTime if true, "startTime" is set to current time
     */
    void load( boolean resetStartTime ) {
        if ( resetStartTime ) {
            startTime = new Date();
        }
        loader.submit(this);
    }

    public Date getStartTime() {
        return( startTime );
    }


    /** get the scene, waiting if necessary */
    public Scene getScene(boolean wait) {
        Scene scene = getScene();
    
        if (wait && started) {
            try {
                while (!getAborted() && scene == null) {
                    Thread.sleep(100);
                    scene = getScene();
                }
            } catch (InterruptedException ie) {
                ;
            }
        }

        return scene;
    }
    

    public void asciiSave() {
        asciiSave( getFile(), false );
    }

    public void gzipSave() {
        asciiSave( getFile(), true );
    }


    class SaveThread extends Thread {
        boolean gzip;
        int numberLines;
        PrintStream out;
        TokenEditor sceneTokenEditor;

        SaveThread( TokenEditor sceneTokenEditor, boolean gzip, int numberLines, PrintStream out ) {
            this.sceneTokenEditor = sceneTokenEditor;
            this.gzip = gzip;
            this.numberLines = numberLines;
            this.out = out;
        }

        public void run() {
            Properties props = BrowserProperties.getProperties();
            String unixFormatProperty = props.getProperty("workspace.saveInUnixFormat");
            boolean saveInUnixFormat = (unixFormatProperty == null) ? false : ("true".equalsIgnoreCase(unixFormatProperty));
            if ( gzip ) {
                for ( int i = 0; i < numberLines; i++ ) {
                    GlobalProgressIndicator.markProgress();
                    if ( saveInUnixFormat ) {
                        out.print( sceneTokenEditor.getNospaceLineAt( i ));
                        out.print( '\n' );
                    } else {
                        out.println( sceneTokenEditor.getNospaceLineAt( i ));
                    }
                }
            } else {
                for ( int i = 0; i < numberLines; i++ ) {
                    GlobalProgressIndicator.markProgress();
                    if ( saveInUnixFormat ) {
                        out.print( sceneTokenEditor.getTabLineAt( i ));
                        out.print( '\n' );
                    } else {
                        out.println( sceneTokenEditor.getTabLineAt( i ));
                    }
                }
            }
            out.flush();
            out.close();
            System.out.println( "Finished saving file, saved " + numberLines + " lines" );
            GlobalProgressIndicator.reset();
            saving = false;
        }
    }

    boolean saving;
    SaveThread saver;
    public void asciiSave( File file, boolean gzip ) {
        if ( saving ) {
            return;
        }
        saving = true;
        setGzip( gzip );
        Scene scene = getScene();
        System.out.println( "Saving file..." );
        try {
            FileOutputStream fos = new FileOutputStream(file);
            OutputStream fo = fos;
            if ( gzip ) {
                fo = new GZIPOutputStream( fo );
            }
            PrintStream out = new PrintStream( fo );
            TokenEditor sceneTokenEditor = (TokenEditor) scene.getTokenEnumerator();
            int numberLines = sceneTokenEditor.getNumberLines();
            String progressStr = "Saving " + file.getName() + " ";
            if ( gzip ) {
                progressStr = "GZIP saving " + file.getName() + " ";
            }
            GlobalProgressIndicator.setProgressIndicator( FactoryDataUpdater.getDefaultProgressIndicator(), progressStr, numberLines );
            SaveThread saveThread = new SaveThread( sceneTokenEditor, gzip, numberLines, out );
            saveThread.start();
            saver = saveThread;
        } catch( Exception e ) {
            e.printStackTrace();
            saving = false;
        }
    }

    public void waitForSave() {
        if ( saver != null ) {
            try {
                saver.join();
            } catch ( Exception e ) {
            }
        }
    }
    

    /** Get the data from a url indicated by a string, and return a byte array containing it */
    public byte[] retrieveData(String name) {
        RetrieveDataThread retriever = new RetrieveDataThread(name);
        return retriever.getData();
    }
        

    class RetrieveDataThread extends Thread {
        String urlstring;
        boolean loading;
        byte[] buffer;

        RetrieveDataThread(String datafile) {
            loading = true;
            RemoteUrlGenerator urlgen = new RemoteUrlGenerator( getBaseFilePath() );
            urlstring = urlgen.createUrlToFetch( datafile );
            start();
        }


        public void run() {
/*
            URL url;
            try {
                url = new URL(urlstring);
            } catch (Exception e) {
e.printStackTrace();
                System.out.println("Unable to retrieve " + urlstring);
                return;
            }
*/
            retrieve(); //url);
            loading = false;
        }

        synchronized byte[] getData() {
            if (loading) {
                try {
                       join();
                } catch (InterruptedException ie) {
                    ;
                }
            }

            return buffer;
        }

        synchronized void retrieve() { //URL url) {
            try {
                InputStream is = InputStreamFactory.getInputStream( urlstring );
                ByteArrayOutputStream bos = new ByteArrayOutputStream();
                byte[] buf = new byte[4096];
                int count;
                while (( count = is.read( buf )) > 0 ) {
                    if ( count == 4096 ) {
                        bos.write( buf );
                    } else {
                        bos.write( buf, 0, count );
                    }
                }
                bos.flush();
                buffer = bos.toByteArray();
System.out.println( "buffer len is " + buffer.length );
                
/*
                DataInputStream dis;
                int contentLength = -1; 
                URLConnection urlc = url.openConnection();
                urlc.connect();
                urlc.getContentType();
                dis = new DataInputStream(urlc.getInputStream());
                contentLength = urlc.getContentLength();
                if (contentLength < 0) {
                    contentLength = 0;
                }
                buffer = new byte[contentLength];
                if (contentLength > 0) {
                    dis.readFully(buffer, 0, contentLength);
                } 
*/
                 
                yield();
            } catch (IOException ie) {
                ie.printStackTrace();
                System.out.println("Unable to retrieve " + urlstring);
            }
        }
    }

    

    public void synchTokenEditor() {
        Scene scene = getScene();
        TokenEditor sceneTokenEditor = (TokenEditor)scene.getTokenEnumerator();
        setTokenEditor( sceneTokenEditor );
        if ( sceneTokenEditor.isDirty() ) {
            sceneTokenEditor.retokenize();
            Scene newScene = new Scene( scene.getUrl(), sceneTokenEditor );
            VRML97parser parser = new VRML97parser();
            parser.Build( sceneTokenEditor, newScene );
            setScene( newScene );
        }
    }

    void dump( String loc ) {
        Scene s = getScene();
        if ( s != null ) {
            TokenEnumerator te = s.getTokenEnumerator();
            if ( te != null ) {
                System.out.println( loc + ", have " + te.getNumberLines() + " lines" );
            } else {
                System.out.println( loc + ", token enumerator is null" );
            }
        } else {
            System.out.println( loc + ", scene is null" );
        }
    }
}


