/*
 * @(#)ViewerPanel.java
 *
 * Copyright (c) 1999-2000 by Trapezium Development LLC.  All Rights Reserved.
 *
 * The information in this file is the property of Trapezium Development LLC
 * and may be used only in accordance with the terms of the license granted
 * by Trapezium.
 *
 */
package com.trapezium.browser;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.util.Vector;
import java.util.Enumeration;
import javax.swing.Box;
import javax.swing.*;
import javax.swing.event.*;
import javax.media.j3d.*;
import javax.vecmath.*;

import java.io.*;
import com.trapezium.vrml.Scene;

import com.trapezium.vizwiz.VizModelInfo;
import com.trapezium.browser.geometry.Page2D;
import com.trapezium.browser.vrml.node.Viewpoint;
import com.trapezium.util.BaseAction;
import com.trapezium.util.JWheelControl;
import com.trapezium.util.JAutoRepeatButton;
import com.trapezium.util.BarLayout;
import com.trapezium.util.ResourceManager;
import com.sun.j3d.utils.universe.*;
import com.sun.j3d.utils.geometry.*;
import com.sun.j3d.utils.behaviors.mouse.MouseTranslate;

/** the panel displaying the data */
/*
 *  The ViewerPanel contains a BrowserCanvas3D which is the 3D display.
 *  That 3D display has a hierarchy of objects containing objects:
 *  <PRE>
 *
 *    ViewerPanel -- has reference back to BrowserCanvas3D
 *      PlatformDeskTop, "desktop"
 *        PlatformGeometry, "pg"
 *          BranchGroup, a child of PlatformGeometry
 *            TransformGroup, a child of BranchGroup
 *
 *  </PRE>
 *  The ViewerPanel is contained in the ViewerFrame, which is the viewer
 *  JFrame in Brillig.
 *
 *  The PlatformDesktop implements the "heads up display" i.e. geometry
 *  which is not part of the scene and stays fixed in relation to the 
 *  current viewpoint.
 */

public class ViewerPanel extends JPanel implements ViewableContainer {

    static final int PI_UNITS = 34;
    static final int PREF_SLIDER_LENGTH = 146;

    Window fullScreenWindow = null;
    boolean fullScreen = false;
    Container saveParent;
    int saveDividerLocation;

    Viewable contents = null;
    
    BrowserCanvas3D canvas3D = null;
    BrowserUniverse universe = null;

    PlatformDesktop desktop;

    JPanel bottomBar;
    JPanel sideBar;
    JPanel cornerBox;
    
    boolean sideShowRotator = true;
    boolean bottomShowRotator = true;

    ViewerZoomer zoomer;
    
    ViewerXRotator rX;
    ViewerYRotator rY;
    ViewerZRotator rZ;  // doesn't do anything yet -- maybe not necessary
    
    ViewerTilter tX;
    ViewerPanner tY;

    ViewerSpeedControl speedSet;
    
    JToggleButton sideSelectRotate;
    JToggleButton sideSelectTilt;
    JToggleButton bottomSelectRotate;
    JToggleButton bottomSelectPan;
    
    JToggleButton fullScreenToggle;
    
    JAutoRepeatButton moveUp;
    JAutoRepeatButton moveDown;
    JAutoRepeatButton moveLeft;
    JAutoRepeatButton moveRight;

    float moveSpeed = 0.5f;  // meters per second

    WheelControlListener listener;
    WheelControlStateButton initialViewpoint;
    ViewerKeyListener keyListener;
    ViewerMouseListener  mouseListener;

    public ViewerPanel(BrowserUniverse u) {
        // do our own double buffering
        super(false);
        
        this.universe = u;

        setLayout(null);
        setOpaque(false);
        
        listener = new WheelControlListener(this);
        keyListener = new ViewerKeyListener();
        addKeyListener(keyListener);
        mouseListener = new ViewerMouseListener();
        addMouseListener(mouseListener);

        bottomBar = new MarginPanel(new Insets(2, 4, 2, 0));
        bottomBar.setLayout(new BarLayout(BarLayout.CENTER | BarLayout.EQUALIZE_BUTTONS));
        add(bottomBar);
        
        sideBar = new MarginPanel(new Insets(4, 2, 0, 2));
        sideBar.setLayout(new BarLayout(BarLayout.VCENTER | BarLayout.EQUALIZE_BUTTONS));
        add(sideBar);

        cornerBox = new JPanel();
        cornerBox.setLayout(new BorderLayout());
        add(cornerBox);
        
        // button margins
        Insets bm = new Insets(1, 1, 1, 1);

        fullScreenToggle = new JToggleButton(new ImageIcon(ResourceManager.getImage("fullscreen")));
    	fullScreenToggle.setMargin(bm);
        fullScreenToggle.setActionCommand("fullscreen");
        fullScreenToggle.addActionListener(listener);
        listener.addCommand("fullscreen");
        add(fullScreenToggle);

        // sidebar controls 

        zoomer = new ViewerZoomer();
        zoomer.addChangeListener(listener);
    	sideBar.add(zoomer);
       
        rX = new ViewerXRotator();
        rX.addChangeListener(listener);
    	sideBar.add(rX);

        tX = new ViewerTilter();
        tX.addChangeListener(listener);
    	sideBar.add(tX);

    	ButtonGroup sideGroup = new ButtonGroup();
    	
    	sideSelectRotate = new JToggleButton(new ImageIcon(ResourceManager.getImage("rotatex")));
    	sideSelectRotate.setMargin(bm);
        sideSelectRotate.setActionCommand("rotatex");
        sideSelectRotate.addActionListener(listener);
        listener.addCommand("rotatex");
    	sideBar.add(sideSelectRotate);
    	sideGroup.add(sideSelectRotate);

    	sideSelectTilt = new JToggleButton(new ImageIcon(ResourceManager.getImage("tilt")));
    	sideSelectTilt.setMargin(bm);
        sideSelectTilt.setActionCommand("tilt");
        sideSelectTilt.addActionListener(listener);        
        listener.addCommand("tilt");
    	sideBar.add(sideSelectTilt);
    	sideGroup.add(sideSelectTilt);

        speedSet = new ViewerSpeedControl();
        speedSet.addChangeListener(listener);
    	sideBar.add(speedSet);

        // bottom bar controls
        WheelControlState initialState = new WheelControlState();
        initialState.rotateXstate = -1.5708;
        JButton topbutton =  new WheelControlStateButton( "topview", listener, initialState );
        topbutton.setMargin(bm);
        initialState.rotateXstate = 0; //-0.65;
        initialViewpoint = new WheelControlStateButton( "frontview", listener, initialState );
        initialViewpoint.setMargin(bm);
        initialState.rotateYstate = -0.83;
        JButton button = new WheelControlStateButton( "leftview", listener, initialState );
        button.setMargin(bm);
        bottomBar.add( button );

        // add front and topview buttons after leftview so leftview
        // can be on the left
        bottomBar.add( initialViewpoint );
        bottomBar.add( topbutton );
        
        initialState.rotateYstate = 0.83;
        button = new WheelControlStateButton( "rightview", listener, initialState );
        button.setMargin(bm);
        bottomBar.add( button );
        initialState.rotateYstate = 0.0;

        rY = new ViewerYRotator();
    	rY.addChangeListener(listener);
    	bottomBar.add(rY);

        tY = new ViewerPanner();
        tY.addChangeListener(listener);
    	bottomBar.add(tY);

        ButtonGroup bottomGroup = new ButtonGroup();

    	bottomSelectRotate = new JToggleButton(new ImageIcon(ResourceManager.getImage("rotatey")));
    	bottomSelectRotate.setMargin(bm);
        bottomSelectRotate.setActionCommand("rotatey");
        bottomSelectRotate.addActionListener(listener);        
        listener.addCommand("rotatey");
    	bottomBar.add(bottomSelectRotate);
    	bottomGroup.add(bottomSelectRotate);

    	bottomSelectPan = new JToggleButton(new ImageIcon(ResourceManager.getImage("pan")));
    	bottomSelectPan.setMargin(bm);
        bottomSelectPan.setActionCommand("pan");
        bottomSelectPan.addActionListener(listener);        
        listener.addCommand("pan");
    	bottomBar.add(bottomSelectPan);
    	bottomGroup.add(bottomSelectPan);

    	button = new JButton(new ImageIcon(ResourceManager.getImage("resetview")));
        button.setMargin(bm);
        button.setActionCommand("resetview");
        button.addActionListener(listener);        
        listener.addCommand("resetview");
        bottomBar.add( button );


        // move buttons
        moveUp = new JAutoRepeatButton(new ImageIcon(ResourceManager.getImage("up")));
    	moveUp.setMargin(bm);
        moveUp.setActionCommand("up");
        moveUp.addActionListener(listener);        
        listener.addCommand("up");
    	sideBar.add(moveUp, "Top");

        moveDown = new JAutoRepeatButton(new ImageIcon(ResourceManager.getImage("down")));
    	moveDown.setMargin(bm);
        moveDown.setActionCommand("down");
        moveDown.addActionListener(listener);        
        listener.addCommand("down");
    	sideBar.add(moveDown, "Bottom");

        moveLeft = new JAutoRepeatButton(new ImageIcon(ResourceManager.getImage("left")));
    	moveLeft.setMargin(bm);
        moveLeft.setActionCommand("left");
        moveLeft.addActionListener(listener);        
        listener.addCommand("left");
    	bottomBar.add(moveLeft, "Left");

        moveRight = new JAutoRepeatButton(new ImageIcon(ResourceManager.getImage("right")));
    	moveRight.setMargin(bm);
        moveRight.setActionCommand("right");
        moveRight.addActionListener(listener);        
        listener.addCommand("right");
    	bottomBar.add(moveRight, "Right");


        // corner box control 

        rZ = new ViewerZRotator();
    	rZ.addChangeListener(listener);
    	cornerBox.add(rZ, BorderLayout.CENTER);

    
        sideShowRotator = true;
        bottomShowRotator = true;

        showSelectedControls();
    }
    
    /** set the initial viewpoint selection */
    public void setInitialView() {
      initialViewpoint.doClick();
    }  

    public void setFieldOfView(double fov) {
        if (canvas3D != null) {
            canvas3D.setFieldOfView(fov);
        }
    }
    

    // show the correct navigation controls on the side and bottom
    void showSelectedControls() {
        rX.setVisible(sideShowRotator);
        tX.setVisible(!sideShowRotator);

        rY.setVisible(bottomShowRotator);                                     
        tY.setVisible(!bottomShowRotator);

        sideSelectRotate.setSelected(sideShowRotator);
        sideSelectTilt.setSelected(!sideShowRotator);
        bottomSelectRotate.setSelected(bottomShowRotator);
        bottomSelectPan.setSelected(!bottomShowRotator);
        
        sideBar.invalidate();
        //sideBar.doLayout();
        bottomBar.invalidate();
        //bottomBar.doLayout();
        validate();
    }
    
    /** Create a 3D viewer and add it to this ViewerPanel */
    public void createCanvas3D() {
        Dimension size = getSize();      
        canvas3D = new BrowserCanvas3D( universe );
        canvas3D.setBounds(0, 0, size.width, size.height);
        canvas3D.addKeyListener(keyListener);
        desktop = new PlatformDesktop();
        //// just a test, don't need this yet
        //Page2D page = new Page2D("Browser Console", 
        //    new Color3f(0.0f, 0.0f, 0.0f), "SanSerif", 24, Font.BOLD);
        //desktop.addPage(page);
        add(canvas3D);
    }

    
    /** add a listener to the list of thingies that need to know when the 
        content changes */
    Vector contentChangeListeners = new Vector();
    public void addChangeListener(ChangeListener listener) {
        contentChangeListeners.addElement(listener);
    }
    public void removeChangeListener(ChangeListener listener) {
        contentChangeListeners.removeElement(listener);
    }
    
    
    public void setFullScreen(boolean fullScreen) {
        System.out.println("setFullScreen(" + fullScreen + ")");
        if (fullScreen != this.fullScreen) {
            this.fullScreen = fullScreen;
            fullScreenToggle.setSelected(fullScreen);
            Container parent = getParent();
            if (fullScreen) {
                if (parent != null && parent != fullScreenWindow) {
                    saveParent = parent;
                    if (saveParent instanceof JSplitPane) {
                       saveDividerLocation = ((JSplitPane)saveParent).getDividerLocation();
                    }
                    
                    parent.remove(this);
                }
                if (fullScreenWindow == null) {
                
                    while (parent != null && !(parent instanceof Frame)) {
                        parent = parent.getParent();
                    }
                
                    if (parent == null) {
                        return;
                    }
                    fullScreenWindow = new Window((Frame) parent);
                }
                    
                Dimension size = canvas3D.getScreen3D().getSize();
                fullScreenWindow.setBounds(0, 0, size.width, size.height);
                fullScreenWindow.add(this);
                fullScreenWindow.setVisible(true);
                fullScreenWindow.toFront();
                fullScreenWindow.requestFocus();
                requestFocus();
            } else {
                if (fullScreenWindow != null && parent == fullScreenWindow && saveParent != null) {
                    fullScreenWindow.setVisible(false);
                    fullScreenWindow.remove(this);
                    fullScreenWindow.dispose();
                    fullScreenWindow = null;
                    saveParent.add(this);
                    if (saveParent instanceof JSplitPane) {
                       ((JSplitPane)saveParent).setDividerLocation(saveDividerLocation);
                    }
                    saveParent.validate();
                }

            }        

            //doLayout();
        }
    }

    Rectangle getCanvasBounds() {
        Dimension size = getSize();
        return new Rectangle(0, 0, size.width - sideBar.getPreferredSize().width, size.height - bottomBar.getPreferredSize().height);
    }        


    public void doLayout() {
        Dimension size = getSize();        
        if (size.width <= 0 || size.height <= 0) {
            return;
        } else if (canvas3D == null) {
            createCanvas3D();
        }

        Rectangle rect = getCanvasBounds();

        int sideWidth = size.width - rect.width;
        int bottomHeight = size.height - rect.height;

        int sideX = rect.width;
        int bottomY = rect.height;
        
        Dimension buttonSize = moveUp.getPreferredSize();
        int sideY = buttonSize.height + 6;      // room for extra button at top right

        fullScreenToggle.setBounds(sideX + 2, 2, buttonSize.width, buttonSize.height);                    
                    
        canvas3D.setBounds(0, 0, sideX, bottomY);
        sideBar.setBounds(sideX, sideY, sideWidth, bottomY - sideY);
        bottomBar.setBounds(0, bottomY, sideX, bottomHeight);
        cornerBox.setBounds(sideX, bottomY, sideWidth, bottomHeight);
        //moveUp.setBounds(sideX + 2, 6 + buttonSize.height, sideWidth - 4, buttonSize.height);
        //moveDown.setBounds(sideX + 2, bottomY - buttonSize.height - 2, sideWidth - 4, buttonSize.height);
        //moveLeft.setBounds(2, bottomY + 2, buttonSize.width, bottomHeight - 4);
        //moveRight.setBounds(sideX - bottomX + 2, bottomY + 2, buttonSize.width, bottomHeight - 4);
        
    }

    class PlatformDesktop {
        
        PlatformGeometry pg;
        
        public PlatformDesktop() {
            // This will move the ViewPlatform back a bit so the
            // objects in the scene can be viewed.

            //ViewingPlatform vp = canvas3D.getViewingPlatform();
            //vp.setNominalViewingTransform();
            
            pg = new PlatformGeometry();
            pg.setCapability(Group.ALLOW_CHILDREN_READ);
            pg.setCapability(Group.ALLOW_CHILDREN_EXTEND);
            canvas3D.setPlatformGeometry(pg);
        }
        
        final float startX = -0.4f;   final float deltaX = 0.05f;   final float maxX = 0.2f;        
        final float startY = -0.4f;   final float deltaY = 0.05f;   final float maxY = 0.2f;        
        final float startZ = -0.9f;  final float deltaZ = -0.01f;  final float minZ = -1.5f;        

        float nextX = startX;   
        float nextY = startY;   
        float nextZ = startZ;   

        public void addPage(Page2D page) {
            Transform3D pageTranslation = new Transform3D();
            pageTranslation.setTranslation(new Vector3f(nextX, nextY, nextZ));
            nextX = (nextX < maxX ? nextX + deltaX : startX);
            nextY = (nextY < maxY ? nextY + deltaY : startY);
            nextZ = (nextZ > minZ ? nextZ + deltaZ : startZ);
            TransformGroup pageGroup = new TransformGroup(pageTranslation);
            pageGroup.addChild(page);
            BranchGroup branch = new BranchGroup();
            branch.addChild(pageGroup);
            pg.addChild(branch);
        }

        // borrowed from SimpleGeometry.java (j3d demo program)
        /*
         * Create the geometry to add to the platform geometry. 
         */
        PlatformGeometry createAimer() {

            PlatformGeometry pg = new PlatformGeometry();

            // This TransformGroup will be used by the MouseTranslate
            // utiltiy to move the cylinder around the canvas.  when the
            // the user holds down mouse button 3.
            TransformGroup moveTG = new TransformGroup();
            moveTG.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
            moveTG.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
            MouseTranslate mouseT = new MouseTranslate(moveTG);
            moveTG.addChild(mouseT);
            BoundingSphere bounds =
              new BoundingSphere(new Point3d(0.0,0.0,0.0), 100.0);
            mouseT.setSchedulingBounds(bounds);
            pg.addChild(moveTG);

            // This TransformGroup is used to place the cylinder in the scene.
            // The cylinder will be rotated 90 degrees so it will appear as
            // a circle on the screen (could be made into a nice gun site...).
            // The cylinder is also displaced a little in Z so it is in front
            // of the viewer.
            Transform3D xForm = new Transform3D();
            xForm.rotX(Math.PI/2.0);
            xForm.setTranslation(new Vector3d(0.0, 0.0, -0.7));
            TransformGroup placementTG = new TransformGroup(xForm);
            moveTG.addChild(placementTG);

            // Create the cylinder - make it thin and transparent.
            Appearance cylinderAppearance = new Appearance();
            TransparencyAttributes transAttrs =
               new TransparencyAttributes(TransparencyAttributes.FASTEST, 0.5f);
            cylinderAppearance.setTransparencyAttributes(transAttrs);
            Cylinder aimer = new Cylinder(0.06f, 0.005f, 0, cylinderAppearance);
            placementTG.addChild(aimer);

            return pg;
        }
    }
    
    /** Set the viewer contents.  The contents have a VRML Scene graph, which 
     *  is loaded into the BrowserCanvas3D.
     *
     *  @param obj contents to set in viewer
     */
    public BranchGroup setViewableContents(Viewable obj, LoaderListener loaderListener) {
        contents = obj;
        if (contents != null) {
            
            // inform change listeners
            Enumeration listeners = contentChangeListeners.elements();
            ChangeEvent ev = new ChangeEvent(this);
            while (listeners.hasMoreElements()) {
                ChangeListener cl = (ChangeListener) listeners.nextElement();
                cl.stateChanged(ev);
            }

            Scene scene = contents.get3DDepiction();
            
            //  try this out for fun, borrowed from Info2Vrml
            try {
                PrintWriter pw = new PrintWriter( new OutputStreamWriter( new FileOutputStream( "current_view.wrl" )));
                scene.generateVrml( pw );
                pw.flush();
                pw.close();
            } catch ( Exception e ) {
                System.out.println( "Exception generating VRML in ViewerPanel: " + e );
            }

            return( canvas3D.loadScene( scene, loaderListener) );
        } else {
            return( null );
        }
    }

    /** Add contents to the viewer.
     *
     *  @param obj contents to set in viewer
     */
    public void addViewableContents(Viewable obj, LoaderListener loaderListener) {
        if (canvas3D == null) {
            createCanvas3D();
            if (canvas3D == null) {
                System.out.println("Oh no! Here comes the jolly green giant!");
                System.exit(0);
            }
        }
          
        if (obj != null) {
            Scene scene = obj.get3DDepiction();
            
            //  try this out for fun, borrowed from Info2Vrml
            /***
            try {
                PrintWriter pw = new PrintWriter( new OutputStreamWriter( new FileOutputStream( "most_recent_addition.wrl" )));
                scene.generateVrml( pw );
                pw.flush();
                pw.close();
            } catch ( Exception e ) {
                System.out.println( "Exception generating VRML in ViewerPanel: " + e );
            }***/
            
            canvas3D.addScene(scene, loaderListener);
        }
    }

    public Viewable getViewableContents() {
        return contents;
    }
    
    /** Reset the view controls to their initial position */
    public void resetViewControls() {
        rX.setValue(0);
        rY.setValue(0);
        rZ.setValue(0);
        tX.setValue(0);
        tY.setValue(0);
        zoomer.setValue( zoomer.zoomAngleToValue(0.785) );
    }
    
    /** Set the view controls to particular values */
    public void reflectState(WheelControlState state) {
        rX.setValue( angleToValue( state.rotateXstate ) );
        rY.setValue( angleToValue( state.rotateYstate ) );
        tX.setValue( angleToValue( state.tiltXstate ) );
        tY.setValue( angleToValue( state.tiltYstate ) );
        zoomer.setValue( zoomer.zoomAngleToValue( state.FOVstate ) );
    }

    int angleToValue( double angle ) {
        // currently values range from -68 to 68
        // so -68 maps to 2*pi and 68 maps to -2*pi
        //System.out.println( "val is " + val );
     
        /** convert to radians */
        int result = (int) (angle / -0.0924);     // convert -68,68 to 2*pi,-2*pi
        return( result );
    }

    class ViewerKeyListener implements KeyListener {
        public void keyTyped(KeyEvent e) {
            System.out.println("ViewerPanel got keyTyped from " + e.getSource().getClass().getName());
        }
        public void keyReleased(KeyEvent e) {
            System.out.println("ViewerPanel got keyReleased from " + e.getSource().getClass().getName());
        }
        public void keyPressed(KeyEvent e) {
            System.out.println("ViewerPanel got keyPressed from " + e.getSource().getClass().getName());
            if (e.getKeyCode() == KeyEvent.VK_ESCAPE) {
                setFullScreen(false);
            }
        }
    }
    class ViewerMouseListener implements MouseListener {
        public void mouseClicked(MouseEvent e) {
            System.out.println("ViewerPanel got mouseClicked from " + e.getSource().getClass().getName());
        }
        public void mouseEntered(MouseEvent e) { }
        public void mouseExited(MouseEvent e) { }
        public void mousePressed(MouseEvent e) {
            System.out.println("ViewerPanel got mousePressed from " + e.getSource().getClass().getName());
            requestFocus();
        }
        public void mouseReleased(MouseEvent e) { }
    }
    
}

class AmalgamatedCanvas extends JComponent {
    public AmalgamatedCanvas() {
    }
    public void paint(Graphics g) {
        super.paint(g);
        Dimension size = getSize();
        
        g.setColor(Color.pink);
        g.fillRect(0, 0, size.width, size.height);
        g.setColor(Color.black);
        g.drawString("Amalgamated here!", 300, 300);
    }
    
}

class MarginPanel extends JPanel {
    Insets margin;
    public MarginPanel(Insets margin) {
        super();
        this.margin = margin;
    }
    
    public Insets getInsets() {
        Insets insets = super.getInsets();
        return new Insets(insets.top + margin.top, insets.left + margin.left, insets.bottom + margin.bottom, insets.right + margin.right);
    }
}

class HorizontalSlider extends JSlider {
    static final int units = ViewerPanel.PI_UNITS;
    public HorizontalSlider(int min, int max, int value) {
        super(JSlider.HORIZONTAL, min, max, value);
    }
    public Dimension getPreferredSize() {
        Dimension dim = super.getPreferredSize();
        dim.width = ViewerPanel.PREF_SLIDER_LENGTH;
        return dim;
    }
    
    public Dimension getMinimumSize() {
        Dimension dim = super.getMinimumSize();
        dim.width = ViewerPanel.PREF_SLIDER_LENGTH/2;
        return dim;
    }

    public Dimension getMaximumSize() {
        Dimension dim = super.getMaximumSize();
        dim.width = ViewerPanel.PREF_SLIDER_LENGTH * 2;
        return dim;
    }
}

class VerticalSlider extends JSlider {
    static final int units = ViewerPanel.PI_UNITS;
    public VerticalSlider(int min, int max, int value) {
        super(JSlider.VERTICAL, min, max, value);
    }
    public Dimension getPreferredSize() {
        Dimension dim = super.getPreferredSize();
        dim.height = ViewerPanel.PREF_SLIDER_LENGTH;
        return dim;
    }
    
    public Dimension getMinimumSize() {
        Dimension dim = super.getMinimumSize();
        dim.height = ViewerPanel.PREF_SLIDER_LENGTH/2;
        return dim;
    }

    public Dimension getMaximumSize() {
        Dimension dim = super.getMaximumSize();
        dim.height = ViewerPanel.PREF_SLIDER_LENGTH * 2;
        return dim;
    }
}

class ViewerZoomer extends VerticalSlider {  // zoom control

    // zoom
    static final int ZoomUnits = zoomAngleToValue(0.0);
    static final int ZoomStart = zoomAngleToValue(0.785);
    public ViewerZoomer() {
    	super( 0, ZoomUnits, ZoomStart );

    	getAccessibleContext().setAccessibleName("Zoom Control");
    	getAccessibleContext().setAccessibleDescription("Zoom in or out");
    }
    
    static int zoomAngleToValue(double angle) {
        angle = 3.14159 - angle;
        int result = (int) ( (angle * angle) / 0.0924 );
        //System.out.println("zoomToVal: " + angle + " -> " + result);
        return( result );
    }
}



// 31416 = 8 * 3 * 7 * 11 * 17 = about (pi * 10000)  

class ViewerXRotator extends JWheelControl {  // pitch
    static final int Xunits = 68;  // was 17
    static final int Xpreferred = 146;  // was 78;
    public ViewerXRotator() {
    	super(JWheelControl.VERTICAL, -Xunits, Xunits, 0);

    	getAccessibleContext().setAccessibleName("X-Axis Rotator");
    	getAccessibleContext().setAccessibleDescription("Rotate the viewpoint around the origin's X axis");
    }
    
    public Dimension getPreferredSize() {
        Dimension dim = super.getPreferredSize();
        dim.height = Xpreferred;
        return dim;
    }
    
    public Dimension getMinimumSize() {
        Dimension dim = super.getMinimumSize();
        dim.height = Xpreferred/2;
        return dim;
    }

    public Dimension getMaximumSize() {
        Dimension dim = super.getMaximumSize();
        dim.height = Xpreferred*2;
        return dim;
    }
   
}

class ViewerYRotator extends JWheelControl {  // yaw
    static final int Yunits = 68;       // was 34
    static final int Ypreferred = 146;
    public ViewerYRotator() {
    	super(JWheelControl.HORIZONTAL, -Yunits, Yunits, 0);

    	getAccessibleContext().setAccessibleName("Y-Axis Rotator");
    	getAccessibleContext().setAccessibleDescription("Rotate the viewpoint around the origin's Y axis");
    }

    public Dimension getPreferredSize() {
        Dimension dim = super.getPreferredSize();
        dim.width = Ypreferred;
        return dim;
    }
    
    public Dimension getMinimumSize() {
        Dimension dim = super.getMinimumSize();
        dim.width = Ypreferred/2;
        return dim;
    }
    
    public Dimension getMaximumSize() {
        Dimension dim = super.getMaximumSize();
        dim.width = Ypreferred;
        return dim;
    }
}


class ViewerZRotator extends JPanel {   // roll
    ChangeListener changeListener = null;

    public ViewerZRotator() {
    }

    public void addChangeListener(ChangeListener listener) {
        changeListener = listener;
    }
    
    public void setValue(int val) {
    }
}

class ViewerTilter extends VerticalSlider {  // pitch
    public ViewerTilter() {
    	super( -units, units, 0 );

    	getAccessibleContext().setAccessibleName("Tilt Control");
    	getAccessibleContext().setAccessibleDescription("Rotate the viewpoint around the X axis");
    }
}


class ViewerPanner extends HorizontalSlider {  // yaw
    public ViewerPanner() {
    	super( -units, units, 0);

    	getAccessibleContext().setAccessibleName("Pan Control");
    	getAccessibleContext().setAccessibleDescription("Rotate the viewpoint around the Y axis");
    }
}

class ViewerSpeedControl extends VerticalSlider {  // yaw

    // speed is calculated logarithmically, speed in cm/sec = 10^(val/100)
    static final int ZERO_SPEED = 0;   // special case: stop
    static final int MAX_SPEED = 400; // 10000 cm/sec, or about about 200mph
    static final int JJ_MAX_SPEED = 200;
    
    public ViewerSpeedControl() {
    	super( ZERO_SPEED, JJ_MAX_SPEED, ZERO_SPEED );
    }
}

class WheelControlStateButton extends JButton implements ActionListener {
    boolean set;
    WheelControlState state;
    WheelControlListener listener;
    
    public WheelControlStateButton( String s, WheelControlListener listener ) {
        super( new ImageIcon(ResourceManager.getImage(s)) );
        this.listener = listener;
        set = false;
        state = new WheelControlState();
        addActionListener( this );
        // this allows a menu command to activate this button
        BaseAction.delegateAction(s, this);
    }
    
    public WheelControlStateButton( String s, WheelControlListener listener, WheelControlState initialState ) {
       this( s, listener );
       state.setState( initialState );
       set = true;
    }
       
       
    public void actionPerformed( ActionEvent event ) {
        if ( !set ) {
            listener.getState( state );
            set = true;
        } else {
            listener.setState( state );
        }
    }
}

class WheelControlState {
    public double rotateXstate;
    public double rotateYstate;
    public double tiltXstate;
    public double tiltYstate;
    public double FOVstate = 0.785;
    
    public void getState( WheelControlState result ) {
        result.rotateXstate = rotateXstate;
        result.rotateYstate = rotateYstate;
        result.tiltXstate = tiltXstate;
        result.tiltYstate = tiltYstate;
        result.FOVstate = FOVstate;
    }
    
    public void setState( WheelControlState state ) {
        rotateXstate = state.rotateXstate;
        rotateYstate = state.rotateYstate;
        tiltXstate = state.tiltXstate;
        tiltYstate = state.tiltYstate;
        FOVstate = state.FOVstate;
    }
}


/**
 *  The WheelControlListener gets called when any of the viewer sliders is moved.
 */


class WheelControlListener extends BaseAction implements ChangeListener, ItemListener {
    ViewerPanel target;
    WheelControlState wcs;

    public WheelControlListener(ViewerPanel target) {
        this.target = target;
        wcs = new WheelControlState();
    }
    
    public synchronized void getState( WheelControlState result ) {
        wcs.getState( result );
    }
    
    public synchronized void setState( WheelControlState state ) {
        wcs.setState( state );
        Viewpoint viewpoint = target.canvas3D.getViewpoint();
        if (viewpoint != null) {
            //viewpoint.resetUserTransform();
            viewpoint.userRotateX(wcs.rotateXstate);
            viewpoint.userRotateY(wcs.rotateYstate);
            viewpoint.userTiltX(wcs.tiltXstate);
            viewpoint.userTiltY(wcs.tiltYstate);
            target.setFieldOfView(wcs.FOVstate);
            target.reflectState(wcs);
        }
   }  

    double valueToAngle( int val ) {
        // currently values range from -68 to 68
        // so -68 maps to 2*pi and 68 maps to -2*pi
        //System.out.println( "val is " + val );
     
        /** convert to radians */
        double result = (double) val * -0.0924;     // convert -68,68 to 2*pi,-2*pi
        return( result );
    }

    double valueToZoomAngle( int val ) {
        double result = Math.sqrt((double) val * 0.0924);
        result = 3.14159 - result;
        result = Math.max(0.05, Math.min(3.1, result));
        //System.out.println("valToZoom: " + val + " -> " + result);
        return( result );
    }

    public void actionPerformed(ActionEvent e) {
        String command = e.getActionCommand();
        
        if ("rotatex".equals(command)) {
            target.sideShowRotator = true;
        } else if ("tilt".equals(command)) {
            target.sideShowRotator = false;
        } else if ("rotatey".equals(command)) {
            target.bottomShowRotator = true;
        } else if ("pan".equals(command)) {
            target.bottomShowRotator = false;
        } else if ("fullscreen".equals(command)) {
            target.setFullScreen(!target.fullScreen);
            return;

        } else {
            Viewpoint viewpoint = target.canvas3D.getViewpoint();
            if (viewpoint != null) {
                float delta = target.moveSpeed;
                if ("up".equals(command)) {
                    viewpoint.userMove(0.0f, delta, 0.0f);
                } else if ("down".equals(command)) {
                    viewpoint.userMove(0.0f, -delta, 0.0f);
                } else if ("left".equals(command)) {
                    viewpoint.userMove(-delta, 0.0f, 0.0f);
                } else if ("right".equals(command)) {
                    viewpoint.userMove(delta, 0.0f, 0.0f);
                } else if ("resetview".equals(command)) {
                    viewpoint.reset();
                } else if ("fullscreen".equals(command)) {
                }
            }
        }
        target.showSelectedControls();
    }

    public void stateChanged(ChangeEvent e) {
        
        JSlider slider = (JSlider) e.getSource();
        if ( slider instanceof ViewerSpeedControl ) {
            VizModelInfo.shiftGrid( "slidingGrid", slider.getValue() );
//            System.out.println( "speed value " + slider.getValue() );
            return;
        }
        String str;    
        Viewpoint viewpoint = target.canvas3D.getViewpoint();
        if (viewpoint != null) {
            if (slider instanceof ViewerXRotator) {
                wcs.rotateXstate = valueToAngle( slider.getValue() );
                viewpoint.userRotateX(wcs.rotateXstate);
            } else if (slider instanceof ViewerYRotator) {
                wcs.rotateYstate = valueToAngle( slider.getValue() );
                viewpoint.userRotateY(wcs.rotateYstate);

            } else if (slider instanceof ViewerTilter) {
                wcs.tiltXstate = -valueToAngle( slider.getValue() );
                viewpoint.userTiltX(wcs.tiltXstate);
            } else if (slider instanceof ViewerPanner) {
                wcs.tiltYstate = -valueToAngle( slider.getValue() );
                viewpoint.userTiltY(wcs.tiltYstate);

            } else if (slider instanceof ViewerZoomer) {
                wcs.FOVstate = valueToZoomAngle( slider.getValue() );
                
             //System.out.println("Zoom control to " + wcs.FOVstate);

                target.setFieldOfView(wcs.FOVstate);
            }
        }
    }
	public void itemStateChanged(ItemEvent e) {
	    AbstractButton b = (AbstractButton) e.getSource();
        String command = b.getActionCommand();
        if ("rotatex".equals(command)) {
            target.sideShowRotator = b.isSelected();;
        } else if ("tilt".equals(command)) {
            target.sideShowRotator = !b.isSelected();
        } else if ("rotatey".equals(command)) {
            target.bottomShowRotator = b.isSelected();
        } else if ("pan".equals(command)) {
            target.bottomShowRotator = !b.isSelected();
        } else if ("fullscreen".equals(command)) {
            target.setFullScreen(b.isSelected());
            return;
        }
        target.showSelectedControls();
    }
}

