/*
 * @(#)RealizableNode.java
 *
 * Copyright (c) 1999 by Trapezium Development LLC.  All Rights Reserved.
 *
 * The information in this file is the property of Trapezium Development LLC
 * and may be used only in accordance with the terms of the license granted
 * by Trapezium.
 *
 */

package com.trapezium.browser.vrml.node;
import com.trapezium.browser.LoaderListener;
import com.trapezium.browser.VrmlPlayer;
import com.trapezium.browser.VrmlNavigation;
import com.trapezium.parse.TokenEnumerator;
import com.trapezium.vrml.Scene;
import com.trapezium.vrml.node.InvalidFieldException;
import com.trapezium.vrml.grammar.*;
import com.trapezium.vrml.fields.*;
import java.util.StringTokenizer;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;
import javax.media.j3d.*;
import javax.vecmath.Point3d;

public abstract class RealizableNode extends com.trapezium.vrml.node.Node {
    static Bounds infiniteBounds = new BoundingSphere(new Point3d(), Double.MAX_VALUE);
    static BoundingLeaf infiniteBoundingLeaf = new BoundingLeaf(infiniteBounds);
    static {
        infiniteBoundingLeaf.setCapability(BoundingLeaf.ALLOW_REGION_READ);
    }
    static Bounds zeroBounds = new BoundingSphere(new Point3d(), -1.0);

    vrml.BaseNode baseNode;
    boolean realized;

    // some nodes wait for a thread to complete, this flag indicates that a
    // subclass has created a thread that has to complete before the node
    // can be fully realized.  

    protected boolean waitingForThread;

    /** Finish whatever details necessary for realization.
     *  Default implementation waits for a thread (possibly should push to subclasses,
     *  and have default implementation do nothing...
     *  Subclasses may override to do whatever else they need.
     */
    public void waitForThreadCompletion() {
        while ( waitingForThread ) {
            System.out.println( getClass().getName() + " waiting for thread..." );
            try { Thread.sleep( 1000 ); } catch( Exception e )  { }
        }
    }

    public RealizableNode() {
        super();
        if (isScript()) {
            baseNode = new vrml.node.Script();
        } else {
            baseNode = new vrml.node.Node();
        }
        baseNode.setNode(this);
        realized = false;
        waitingForThread = false;
    }

    static Hashtable browserTable = new Hashtable();

    /** only nodes which do not belong to a scene need to call this */
    public void setBrowser(vrml.Browser browser) {
        browserTable.put(this, browser);
    }      

    /** Any subclass which waits for completsion in any form calls this */
    public void updateWaiter() {
        Scene s = (Scene)getScene();
        Vector waiter = s.getWaitingCompletionTable();
        waiter.addElement( this );
    }

    /** template method, subclasses override if they have no corresponding Java 3d object */
    public boolean hasJava3dObject() {
        return( true );
    }

    /** template method, subclasses which need access to transform group call this..
        maybe better alternative would be to provide this in "realize"?
     */
    public void setTransformGroup() {
System.out.println( "RealizableNode template method setTransformGroup" );
    }
    
    public vrml.Browser getBrowser() {
        Scene scene = (Scene) getScene();
        if (scene != null) {
            return scene.getBrowser();
        } else {
            return (vrml.Browser) browserTable.get(this);
        }
    }

    public VrmlPlayer getPlayer() {
        vrml.Browser browser = getBrowser();
        if (browser != null) {
            return browser.getPlayer();
        } else {
            return null;
        }
    }
    

    public void updateField(String fieldName, double time) {
        // subclasses do all the work
    }

    public void realize( LoaderListener loaderListener ) {
        realize();
    }
    
    public void realize() {
        
        if (realized) {
//System.out.println("No need to realize " + toString().substring(toString().lastIndexOf('.') + 1));
            return;
        }
        
        // initialize all runtime fields    
        Hashtable fields = VRML97.getFieldTable( getNodeName() );

//System.out.println("Realizing " + toString().substring(toString().lastIndexOf('.') + 1));

        Enumeration enum = fields.keys();
        while (enum.hasMoreElements()) {
            // RuntimeField.getRuntimeField() creates the runtime field and 
            // initializes it the first time a field is referenced
            String key = (String) enum.nextElement();
//System.out.print("  ...realize field " + key);
            RuntimeField.getRuntimeField(this, key);
        }       
    
        realized = true;
    }

    public boolean isRealized() {
        return realized;
    }
    
    public void reset() {
        baseNode = null;
        realized = false;
    }


    // called by parent after node is added to parent
    public void updateParent(javax.media.j3d.Node j3dParent) {
    }

    public abstract javax.media.j3d.SceneGraphObject getSceneGraphObject();
    
    /** for convenience, an alternative to getSceneGraphObject which assumes
      * the object is a Java3D node and casts accordingly.
      */
    final public javax.media.j3d.Node getJ3dNode() {
        return (javax.media.j3d.Node) getSceneGraphObject();
    }


    static Hashtable sharedGroups = new Hashtable();
   
    public SharedGroup getSharedGroup() {
        javax.media.j3d.Node node = getJ3dNode();
if ( node == null ) {
System.out.println( "my node is null, i am " + this );
}
        SharedGroup sg = (SharedGroup) sharedGroups.get(node);
        if (sg == null) {
System.out.println( "AddChild " + node.getClass().getName() );
System.out.println( ".. node is " + node );
System.out.println( ", isLive " + node.isLive() );
System.out.println( ", isCompiled " + node.isCompiled() );
//if ( ! ( node instanceof VrmlNavigation )) {
//System.out.println( ", parent is " + node.getParent() );
//System.out.println( ",, its not a VrmlNavigation node!" );
            sg = new SharedGroup();
            sg.addChild(node);
            sharedGroups.put(node, sg);
//} else System.out.println( "it is a VrmlNavigation node!" );
        }
        return sg;
    }
    
    
    // I don't think this is needed
    //public vrml.Field getVrmlField(String fieldName) {
    //    RuntimeField rf = getRuntimeField(fieldName);
    //    return rf.getVrmlField();
    //}
    
    /** Indicates that this node is exposed to scripts through a vrml.BaseNode object */
    public boolean hasBaseNode() {
        return( true );
    }

    public vrml.BaseNode getBaseNode() {
        return baseNode;
    }

    public void setBaseNode(vrml.BaseNode baseNode) {
        this.baseNode = baseNode;
    }

    // methods used for accessing changeable fields at runtime 
    public RuntimeField getRuntimeField( String fieldName ) {
        return RuntimeField.getRuntimeField(this, fieldName);
    }

    
    public com.trapezium.vrml.node.Node getUpdatedNodeValue( String fieldName ) {
        RuntimeField rf = getRuntimeField(fieldName);
        return rf.getNodeValue();
    }

    public com.trapezium.vrml.node.Node[] getUpdatedNodeArray( String fieldName ) {
        RuntimeField rf = getRuntimeField(fieldName);
        return rf.getNodeArray();
    }

    public boolean getUpdatedBooleanValue( String fieldName ) {
        RuntimeField rf = getRuntimeField(fieldName);
        return rf.getBooleanValue();
    }

    public String getUpdatedStringValue( String fieldName ) {
        RuntimeField rf = getRuntimeField(fieldName);
        return rf.getStringValue();
    }
    
    public String[] getUpdatedStringArray( String fieldName ) {
        RuntimeField rf = getRuntimeField(fieldName);
        return rf.getStringArray();
    }
    
    public float getUpdatedFloatValue( String fieldName ) {
        RuntimeField rf = getRuntimeField(fieldName);
        return rf.getFloatValue();
    }

    public float[] getUpdatedFloatArray( String fieldName ) {
        RuntimeField rf = getRuntimeField(fieldName);
        return rf.getFloatArray();
    }

    /** Update a boolean field value.  
     *
     *  @param fieldName name of field
     *  @param value boolean value
     */
    public void updateValue( String fieldName, boolean value ) {
        updateValue(fieldName, new Boolean(value));
    }
        
    public void updateValue( String fieldName, int value ) {
        updateValue(fieldName, new Integer(value));
    }

    /** Update a float field value.
     *
     *  @param fieldName name of field
     *  @param value float value
     */
    public void updateValue( String fieldName, float value ) {
        updateValue(fieldName, new Float(value));
    }
    
    public void updateValue( String fieldName, double value ) {
        updateValue(fieldName, new Double(value));
    }

    public void updateValue( String fieldName, Object value ) {
        RuntimeField rf = getRuntimeField(fieldName);
        rf.propagateValue(getPlayer(), value);
    }
    
    // trigger routes without calling setValue
    public void propagateValue( String fieldName, Object value ) {
        RuntimeField rf = getRuntimeField(fieldName);
        rf.propagateValue(getPlayer(), value);
    }

    // convenience routines for accessing fields

    public boolean getBooleanValue( String fieldName ) {
        // Node has a function to do this already
        return getBoolean(fieldName);
    }


    /** Get the string value of a field, return default value if the field is
      * not present.
      */
    public String getStringValue( String fieldName ) {
        Field field = getField( fieldName );
        if ( field == null ) {
            //System.out.println( "getNodeName() is " + getNodeName() + ", field " + fieldName );
            FieldDescriptor fd = VRML97.getFieldDescriptor( getNodeName(), fieldName );
            //System.out.println( "fd is " + fd );
            return fd.getDefaultValue();
        } else {
            FieldValue val = field.getFieldValue();
            if (val instanceof SFStringValue) {
                return ((SFStringValue) val).getStringValue();
            } else {
                TokenEnumerator te = ((Scene)getScene()).getTokenEnumerator();
                return te.toString(val.getFirstTokenOffset());
            }
        }
    }

    /** Return an array of all the strings in a field, or null if the
      * field is not present.
      */
    public String[] getStringArray( String fieldName ) {
        Field field = getField( fieldName );
        if ( field != null ) {
            return( field.getStringArray() );
        }
        return null;
    }


    /** Get the float value of a field, return default value if the field is
      * not present.
      */
    public float getFloatValue( String fieldName ) {
        Field field = getField( fieldName );
        if ( field == null ) {
            FieldDescriptor fd = VRML97.getFieldDescriptor( getNodeName(), fieldName );
            return Float.valueOf(fd.getDefaultValue()).floatValue();
        } else {
            FieldValue val = field.getFieldValue();
            TokenEnumerator te = ((Scene)getScene()).getTokenEnumerator();
            return( te.getFloat(val.getFirstTokenOffset()) );
        }
    }

    /** Return an array of all the floats representing a field, or the default
      * value if the field is not present.
      */
    public float[] getFloatArray( String fieldName ) {
        return getFloatArray(fieldName, -1);
    }


    /** Return a specified number of floats representing a field, or the default
      * value if the field is not present.
      */
    public float[] getFloatArray( String fieldName, int n ) {
        float[] floats;
        Field field = getField( fieldName );
        if ( field == null ) {
            FieldDescriptor fd = VRML97.getFieldDescriptor( getNodeName(), fieldName );
            StringTokenizer tok = new StringTokenizer(fd.getDefaultValue(), "\t\n\r ,[]");
            if (n == -1) {
                n = tok.countTokens();
            }
            floats = new float[n];
            for (int i = 0; i < n; i++) {
                floats[i] = Float.valueOf(tok.nextToken()).floatValue();
            }
        } else {
            TokenEnumerator te = ((Scene)getScene()).getTokenEnumerator();
            FieldValue fv = field.getFieldValue();
            if (fv instanceof SFFieldValue) {
                if (n == -1) {
                    if (fv instanceof SFVec2fValue) {
                        n = 2;
                    } else if (fv instanceof SFVec3fValue || fv instanceof SFColorValue) {
                        n = 3;
                    } else if (fv instanceof SFRotationValue) {
                        n = 4;
                    } else {
                        n = 1;
                    }
                }
                floats = new float[n];
                int offset = te.getNextNumber(fv.getFirstTokenOffset());
                for (int i = 0; i < n; i++) {
                    floats[i] = te.getFloat(offset);
                    offset = te.getNextNumber( offset + 1 );
                }
            } else {
                MFFieldValue val = (MFFieldValue) field.getFieldValue();
                if (n == -1) {
                    n = val.getRawValueCount();
                }
                floats = new float[n];
                for (int i = 0; i < n; i++) {
                    FieldValue v = val.getFieldValueAt(i);
                    floats[i] = te.getFloat(v.getFirstTokenOffset());
                }
            }
        }
        return floats;
    }

    /** Get the double value of a field, return default value if the field is
      * not present.
      */
    public double getDoubleValue( String fieldName ) {
        Field field = getField( fieldName );
        if ( field == null ) {
            FieldDescriptor fd = VRML97.getFieldDescriptor( getNodeName(), fieldName );
            return Double.valueOf(fd.getDefaultValue()).doubleValue();
        } else {
            FieldValue val = field.getFieldValue();
            TokenEnumerator te = ((Scene)getScene()).getTokenEnumerator();
            return( te.getDouble(val.getFirstTokenOffset()) );
        }
    }


    
    /** Return an array of all the doubles representing a field, or the default
      * value if the field is not present.
      */
    public double[] getDoubleArray( String fieldName ) {
        return getDoubleArray(fieldName, -1);
    }


    /** Return a specified number of doubles representing a field, or the default
      * value if the field is not present.
      */
    public double[] getDoubleArray( String fieldName, int n ) {
        double[] doubles;
        Field field = getField( fieldName );
        if ( field == null ) {
            FieldDescriptor fd = VRML97.getFieldDescriptor( getNodeName(), fieldName );
            StringTokenizer tok = new StringTokenizer(fd.getDefaultValue(), "[],\t\n\r ,");
            if (n == -1) {
                n = tok.countTokens();
            }
            doubles = new double[n];
            for (int i = 0; i < n; i++) {
                doubles[i] = Double.valueOf(tok.nextToken()).doubleValue();
            }
        } else {
            TokenEnumerator te = ((Scene)getScene()).getTokenEnumerator();
            MFFieldValue val = (MFFieldValue) field.getFieldValue();
            if (n == -1) {
                n = val.getRawValueCount();
            }
            doubles = new double[n];
            for (int i = 0; i < n; i++) {
                FieldValue fv = val.getFieldValueAt(i);
                doubles[i] = te.getDouble(fv.getFirstTokenOffset());
            }
        }
        return doubles;
    }

}