/*
 * @(#)RuntimeField.java
 *
 * Copyright (c) 1999 by Trapezium Development LLC.  All Rights Reserved.
 *
 * The information in this file is the property of Trapezium Development LLC
 * and may be used only in accordance with the terms of the license granted
 * by Trapezium.
 *
 */

package com.trapezium.browser.vrml.node;

import vrml.BaseNode;
import com.trapezium.browser.VrmlPlayer;
import com.trapezium.parse.TokenEnumerator;
import com.trapezium.vrml.*;
import com.trapezium.vrml.node.*;
import com.trapezium.vrml.grammar.*;
import com.trapezium.vrml.fields.*;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;
import java.lang.reflect.*;


/** This class encapsulates the four possible forms a field value can take
    at runtime: an instance variable in the running script, a field whose
    value is stored in a node class member, a field whose value is stored
    in the scene graph, or a default value for a field not specified.
    
    RuntimeField objects are not allocated by the program; rather, they are 
    automatically allocated as needed and retrieved via static methods.
    
  */
public class RuntimeField /*extends Group*/  {

    /** if there's a runtime variable for this field, it's here */
    java.lang.reflect.Field runtimeField = null;

    /** if there's a value for this field in the source file, it's here */
    com.trapezium.vrml.fields.Field staticField = null;

    /** if there's neither a runtime variable nor a source file entry for
        this field, the value is generated here */
    com.trapezium.vrml.grammar.FieldDescriptor defaultField = null;


    Node node;
    String name;

    
    final static String delim = "\t\n\r ,[]";

    Vector connections = null;
    
    private static Hashtable globalFields = new Hashtable();
    
    public static RuntimeField getRuntimeField(vrml.BaseNode node, String fieldName) {
        return getRuntimeField(node.getNode(), fieldName);
    }
    
    public static RuntimeField getRuntimeField(Node node, String fieldName) {
        String runtimeName = getRuntimeName(node, fieldName);
//System.out.print("  runtimeName " + runtimeName);
        Scene scene = (Scene) node.getScene();

        Hashtable runtimeFields = (scene == null ? globalFields : scene.getRuntimeFieldTable());
 //System.out.println("  table " + (runtimeFields == null ? "NULL!!!" : (runtimeFields == globalFields ? "global" :  "scene")));

        RuntimeField rf = (RuntimeField) runtimeFields.get(runtimeName);
        if (rf == null) {
            rf = new RuntimeField(node, fieldName);
            runtimeFields.put(runtimeName, rf);
        }        
        return rf;
    }

    public static RuntimeField createRuntimeField(boolean b) {
        return new BooleanField(b);
    }
    
    public static RuntimeField createRuntimeField(int i) {
        return new IntField(i);
    }
    
    public static RuntimeField createRuntimeField(int[] i) {
        return new ArrayField(i);
    }

    public static RuntimeField createRuntimeField(float f) {
        return new FloatField(f);
    }

    public static RuntimeField createRuntimeField(float x, float y) {
        float[] f = new float[2];
        f[0] = x;
        f[1] = y;
        return new ArrayField(f);
    }

    public static RuntimeField createRuntimeField(float x, float y, float z) {
        float[] f = new float[3];
        f[0] = x;
        f[1] = y;
        f[2] = z;
        return new ArrayField(f);
    }

    public static RuntimeField createRuntimeField(float x, float y, float z, float a) {
        float[] f = new float[4];
        f[0] = x;
        f[1] = y;
        f[2] = z;
        f[3] = a;
        return new ArrayField(f);
    }

    public static RuntimeField createRuntimeField(float[] f) {
        return new ArrayField(f);
    }

    public static RuntimeField createRuntimeField(float[][] f) {
        return new ArrayField(f);
    }

    public static RuntimeField createRuntimeField(float[] f, int tupleSize) {
        int len = f.length / tupleSize;
        float[][] ff = new float[len][];
        for (int i = 0; i < len; i++) {
            ff[i] = new float[tupleSize];
            System.arraycopy(f, i * tupleSize, ff[i], 0, tupleSize);
        }
        return new ArrayField(ff);
    }

    public static RuntimeField createRuntimeField(double d) {
        return new DoubleField(d);
    }
    
    public static RuntimeField createRuntimeField(double[] d) {
        return new ArrayField(d);
    }
    
    public static RuntimeField createRuntimeField(String s) {
        return new StringField(s);
    }

    public static RuntimeField createRuntimeField(String[] s) {
        return new ArrayField(s);
    }

    public static RuntimeField createRuntimeField(BaseNode node) {
        return new BaseNodeField(node);
    }

    public static RuntimeField createRuntimeField(BaseNode[] node) {
        return new ArrayField(node);
    }

    /** catch-all generic object field */
    public static RuntimeField createRuntimeField(Object obj) {
        if (obj.getClass().isArray()) {
            return new ArrayField(obj);
        } else {
            return new ObjectField(obj);
        }
    }


    /** class constructor */
    protected RuntimeField(String fieldName) {
        this(null, fieldName);
    }

    /** class constructor */
    protected RuntimeField(Node node, String fieldName) {
        
boolean debug = false;
//if ( fieldName.compareTo( "url" ) == 0 ) debug = true;
        this.node = node;
        this.name = fieldName;

        java.lang.reflect.Field[] fields;
        if (node != null) {
            staticField = node.getField(fieldName);
if ( debug ) if ( staticField != null ) System.out.println( "Node " + node.getNodeName() + " has field " + fieldName );
            defaultField = VRML97.getFieldDescriptor(node.getNodeName(), fieldName);
            fields = node.getClass().getFields();

        } else {
            staticField = null;
            defaultField = null;
            fields = this.getClass().getFields();
        }
if ( debug )
System.out.print("RuntimeField for node " + node.getNodeName() + ".....checking for field " + fieldName);
        int i;
        for (i = 0; i < fields.length; i++) {
            if (fieldName.equals(fields[i].getName())) {
if ( debug )
System.out.print("     >>-- FOUND ---<<");          
                runtimeField = fields[i];
                reset();
                break;
            }
        }
if ( debug ) {
if (i == fields.length) System.out.print(" <<>>-- NOT found ---<<>>");          
System.out.println("       ...done");
}
    }

    /** reset the runtime variable to its initial value (from file or default) */
    public void reset() {
        
        if (runtimeField == null) {
            return;
        }
        
        Object obj = null;
        if ( staticField != null ) {
//System.out.print("     %%%%%%%%%%% reset using staticField...");            

            FieldValue fv = staticField.getFieldValue();
//System.out.println( "fv is " + fv );
            obj = getFieldValueObject(fv);
//System.out.println( "obj is " + obj );
        } else if ( defaultField != null ) {
//System.out.print("     %%%%%%%%%%% reset using defaultField...");            
            String value = defaultField.getDefaultValue();
//System.out.print("  ...value " + value);            
            int type = defaultField.getFieldType();
        
            if (type == VRML97.SFBool) {
                obj = new Boolean(value);
                
            } else if (type == VRML97.SFString) {
                obj = value;
            } else if (type == VRML97.MFString) {
                String[] strings = new String[1];
                strings[0] = value;
                obj = strings;
                
            } else if ( type == VRML97.SFImage || 
                        type == VRML97.SFNode || 
                        type == VRML97.MFNode || 
                        type == VRML97.MFTime) {
                if (value != null && !value.equalsIgnoreCase("null")) {
                    // not sure what to do about these yet
                    System.out.println("Unsupported field type " + type + " in RuntimeField.reset()");
                }
                obj = null;

            } else {
                // remaining types can't handle null strings
                if (value == null) {
                    value = "0";
                }
                
                if (type == VRML97.SFInt32) {
                    obj = new Integer(value);
                    
                } else if (type == VRML97.SFTime) {
                    obj = new Double(value);

                } else if (type == VRML97.SFFloat) {
                    obj = new Float(value);

                } else {
                    StringTokenizer tok = new StringTokenizer(value, delim);
                    int n = tok.countTokens();
                    
                    if (type == VRML97.MFInt32) {
                        int[] ints = new int[n];
                        for (int i = 0; i < n; i++) {
                            ints[i] = Integer.valueOf(tok.nextToken()).intValue();
                        }
                        obj = ints;
                    } else {
                        if ( type == VRML97.MFColor || 
                             type == VRML97.MFRotation || 
                             type == VRML97.MFVec2f || 
                             type == VRML97.MFVec3f ) {
                            int dim = getValueDimension(defaultField);
                            n /= dim;
                            float[][] tuples = new float[n][dim];
                            for (int i = 0; i < n; i++) {
                                for (int j = 0; j < dim; j++) {
                                    tuples[i][j] = Float.valueOf(tok.nextToken()).floatValue();
                                }
                            }
                            obj = tuples;
                        } else {
                            float[] floats = new float[n];
                            for (int i = 0; i < n; i++) {
                                floats[i] = Float.valueOf(tok.nextToken()).floatValue();
                            }
                            obj = floats;
                        }
                    }
                }
            }
        }
//System.out.print(" >> setValue: " + valueString(obj) + " << ");            
        setValue( obj );
    }
    
    Object getFieldValueObject(FieldValue fv) {
        Object obj = null;
        TokenEnumerator te = ((Scene)node.getScene()).getTokenEnumerator();
    
        if ( fv instanceof MFFieldValue ) {
            MFFieldValue mfv = (MFFieldValue) fv;
            int numvalues = mfv.getRawValueCount();
            if ( mfv instanceof MFFloatValue ) {
                obj = getFieldValueFloatArray(mfv, te, numvalues); 

            } else if ( mfv instanceof MFInt32Value ) {
                obj = getFieldValueIntArray(mfv, te, numvalues); 

            } else if ( mfv instanceof MFColorValue ) {
                obj = getFieldValueTupleArray(mfv, te, numvalues/3, 3); 
               
            } else if ( mfv instanceof MFRotationValue ) {
                obj = getFieldValueTupleArray(mfv, te, numvalues/4, 4); 
               
            } else if ( mfv instanceof MFVec2fValue ) {
                obj = getFieldValueTupleArray(mfv, te, numvalues/2, 2); 
               
            } else if ( mfv instanceof MFVec3fValue ) {
                obj = getFieldValueTupleArray(mfv, te, numvalues/3, 3); 

            } else if ( mfv instanceof MFNodeValue ) {
                Node[] nodes = new Node[numvalues];
                int numchildren = mfv.numberChildren();
                int n = 0;
                for ( int i = 0; i < numchildren; i++ ) {
                    if (n >= numvalues) {
                        break;
                    }
                    
                    VrmlElement child = mfv.getChildAt( i );
                    if ( child instanceof Node ) {
                        Node node = (Node) child;
                        if ( node.isDEForUSE() ) {
                            node = node.getNode();
                        }
                        if (node instanceof RealizableNode) {
                            RealizableNode rnode = (RealizableNode) node;
                            if (!rnode.isRealized()) {
                                rnode.realize();
                            }
                        }
                        nodes[n++] = node;
                    }
                }
                obj = nodes;
            } else if ( mfv instanceof MFStringValue ) {
//System.out.println( "hmm.. numvalues is " + numvalues );
                obj = getFieldValueStringArray( mfv, te );
            }
        } else if ( fv instanceof SFColorValue ) {
            obj = getFieldValueFloatArray(fv, te, 3); 
           
        } else if ( fv instanceof SFRotationValue ) {
            obj = getFieldValueFloatArray(fv, te, 4); 
           
        } else if ( fv instanceof SFVec2fValue ) {
            obj = getFieldValueFloatArray(fv, te, 2); 
           
        } else if ( fv instanceof SFVec3fValue ) {
            obj = getFieldValueFloatArray(fv, te, 3); 

        } else if ( fv instanceof SFBoolValue ) {
            obj = new Boolean(((SFBoolValue) fv).getValue());

        } else if ( fv instanceof SFInt32Value ) {
            obj = Integer.valueOf(te.toString(fv.getFirstTokenOffset()));
    
        } else if ( fv instanceof SFFloatValue ) {
            obj = new Float(te.getFloat(fv.getFirstTokenOffset()));
    
        } else if ( fv instanceof SFTimeValue ) {
            obj = new Double(te.getDouble(fv.getFirstTokenOffset()));
    
        } else if (fv instanceof SFStringValue) {
            obj = ((SFStringValue) fv).getStringValue();

        } else if ( fv instanceof SFImageValue ) {
            // to do
System.out.println( "Hmm.. an SFImageValue" );
        } else if ( fv instanceof SFNodeValue ) {
            Node node = ((SFNodeValue) fv).getNode();
            if ( node.isDEForUSE() ) {
                node = node.getNode();
            }
            if (node instanceof RealizableNode) {
                RealizableNode rnode = (RealizableNode) node;
                if (!rnode.isRealized()) {
                    rnode.realize();
                }
            }
            obj = node;
        }
    
        return obj;
    }

    Object getFieldValueStringArray(MFFieldValue mfv, TokenEnumerator te) { 
        int nvalues = mfv.numberValues();
        if ( nvalues > 0 ) {
            String[] strs = new String[nvalues];
            for ( int i = 0; i < nvalues; i++ ) {
                FieldValue fv = mfv.getFieldValueAt( i );
                if ( fv instanceof SFStringValue ) {
                    SFStringValue sfv = (SFStringValue)fv;
                    strs[i] = sfv.getStringValue();
                }
            }
            return( strs );
        }
        return null;
    }
    Object getFieldValueIntArray(FieldValue fv, TokenEnumerator te, int n) { 
        int[] ints = new int[n];
        int offset = te.getNextNumber(fv.getFirstTokenOffset());
        for (int i = 0; (i < n) && (offset != -1); i++) {
            ints[i] = te.getIntValue(offset);
            offset = te.getNextNumber( offset + 1 );
        }
        return ints;
    }
    
    Object getFieldValueFloatArray(FieldValue fv, TokenEnumerator te, int n) { 
//System.out.println( "getFieldValueFloatArray for " + n + " values" );
        float[] floats = new float[n];
        int idx = te.getNextNumber(fv.getFirstTokenOffset());
        for (int i = 0; i < n; i++) {
            floats[i] = te.getFloat(idx);
            idx = te.getNextNumber( idx + 1 );
//System.out.println(" ++++++++++++ floats[" + i + "]=" + floats[i]);                
        }
        return floats;
    }
    
    /** Extract a tuple array from the scene graph */
    Object getFieldValueTupleArray(FieldValue fv, TokenEnumerator te, int n, int dim) { 

        float[][] tuples = new float[n][dim];
        int idx = te.getNextNumber( fv.getFirstTokenOffset() );
        for (int i = 0; i < n; i++) {
            for (int j = 0; j < dim; j++) {
                tuples[i][j] = te.getFloat( idx );
                idx = te.getNextNumber( idx + 1 ); 
//System.out.println(" ++++++++++++ tuples[" + i + "][" + j + "]=" + tuples[i][j]); 
            }
        }
        return tuples;
    }

    private vrml.Field vrmlField = null;
    public vrml.Field getVrmlField() {
        if (vrmlField == null) {
            vrmlField = vrml.field.FieldFactory.createField(this);
        }
        return vrmlField;
    }
    
    public vrml.BaseNode getVrmlNode() {
        return node.getBaseNode();
    }
    
    public String getName() {
        return name;
    }
    
    public Node getNode() {
        return node;
    }
    
    public String[] getFieldNames() {
        java.lang.reflect.Field[] fields = node.getClass().getFields();
        String[] names = new String[fields.length];
        
        for (int i = 0; i < fields.length; i++) {
            names[i] = fields[i].getName();
        }
        return names;
    }

    
    /** Return the number of values in this field */
    public int getSize() {
        if (runtimeField != null) {
            Object[] fields; 
            try {
                fields = (Object[]) runtimeField.get(node);
                return fields.length;
            } catch (Exception e) {
            }
        }

        if ( staticField != null ) {
            FieldValue fv = staticField.getFieldValue();
            if (fv instanceof MFFieldValue) {
                return ((MFFieldValue) fv).getRawValueCount();
            } else {
                return 1;
            }            
        }

        if ( defaultField != null ) {
            return 1;
        }
        
        return 0;
    }
    
    public void clear() {
        if (runtimeField != null) {
            Class c = runtimeField.getType();
            if (c.isArray()) {
                // don't know how to do this... the following is wrong
                //runtimeField = Array.newInstance(c.getComponentType(), 0);
            }
        } else if (staticField != null) {

            // to do            
            
        }
        // can't clear out default field
    }
    
        
    public void delete(int index) {
    
        // to do
    }
    

    /** get one field out of a MF-type field */    
    //public Field get1Field(int index) {
    //}

   
    
    public BaseNode getBaseNodeValue() {
        Node n = getNodeValue();
        if (n != null) {
            return n.getBaseNode();
        } else {
            return null;
        }
    }

    public Node getNodeValue() {

        if (runtimeField != null) {
            try {
                return (Node) runtimeField.get(node);                
            } catch (Exception e) {
            }
        }

        if (staticField != null) {
            FieldValue fv = staticField.getFieldValue();
            if ( fv instanceof SFNodeValue ) {
                SFNodeValue sfnv = (SFNodeValue) fv;
                return sfnv.getNode();
            }
        }
    
        if (defaultField != null) {
            // we don't handle this yet -- maybe never needed
        }

       return null;
    }

    public Node[] getNodeArray() {
        if (runtimeField != null) {
            try {
                return (Node[]) runtimeField.get(node);
            } catch (Exception e) {
            }
        }

        if (staticField != null) {
            return staticField.getNodeArray();
        }
        
        if (defaultField != null) {
            // we don't handle this yet -- maybe never needed
        }

        return (new Node[0]);
    }

    
    public BaseNode[] getBaseNodeArray() {

        if (runtimeField != null) {
            try {
                Node[] nodes = (Node[]) runtimeField.get(node);
                BaseNode[] basenodes = new BaseNode[nodes.length];
                for (int i = 0; i < nodes.length; i++) {
                    basenodes[i] = nodes[i].getBaseNode();
                }
                return basenodes;
            } catch (Exception e) {
            }
        }

        if (staticField != null) {
            return staticField.getBaseNodeArray();
        }
    
        if (defaultField != null) {
            // we don't handle this yet -- maybe never needed
        }

        return (new BaseNode[0]);
    }


    public boolean getBooleanValue() {
        
        if (runtimeField != null) {
            try {
                return (boolean) runtimeField.getBoolean(node);
            } catch (Exception e) {
            }
        }

        if (staticField != null) {
            FieldValue fv = staticField.getFieldValue();
            if ( fv instanceof SFBoolValue ) {
                SFBoolValue sfbv = (SFBoolValue)fv;
                return sfbv.getValue();
            }
        }
       
        if (defaultField != null) {
            return defaultField.getDefaultValue().equals("TRUE");
        }

        return false;
    }


    /** Get the string value of a field, return default value if the field is
      * not present.
      */
    public String getStringValue() {
        if (runtimeField != null) {
            try {
                return (String) runtimeField.get(node);
            } catch (Exception e) {
            }
        }

        if (staticField != null) {
            FieldValue fv = staticField.getFieldValue();
            if (fv instanceof SFStringValue) {
                return ((SFStringValue) fv).getStringValue();
            } else {
                TokenEnumerator te = ((Scene)node.getScene()).getTokenEnumerator();
                return te.toString(fv.getFirstTokenOffset());
            }
        }

        if (defaultField != null) {
            return defaultField.getDefaultValue();
        }
        return null;
    }
    
    /** Get the integer value of a field, return default value if the field is
      * not present.
      */
    public int getIntValue() {
        if (runtimeField != null) {
            try {
                return runtimeField.getInt(node);
            } catch (Exception e) {
            }
        }

        if (staticField != null) {
            FieldValue fv = staticField.getFieldValue();
            TokenEnumerator te = ((Scene)node.getScene()).getTokenEnumerator();
            return( te.getIntValue(fv.getFirstTokenOffset()) );
        }

        if ( defaultField != null ) {
            return Integer.valueOf(defaultField.getDefaultValue()).intValue();
        }
    
        return 0;
    }
    
    /** Get the float value of a field, return default value if the field is
      * not present.
      */
    public float getFloatValue() {
        if (runtimeField != null) {
            try {
                return runtimeField.getFloat(node);
            } catch (Exception e) {
            }
        }

        if (staticField != null) {
            FieldValue fv = staticField.getFieldValue();
            TokenEnumerator te = ((Scene)node.getScene()).getTokenEnumerator();
            return( te.getFloat(fv.getFirstTokenOffset()) );
        }

        if ( defaultField != null ) {
            return Float.valueOf(defaultField.getDefaultValue()).floatValue();
        }
    
        return 0.0f;
    }

    /** Get the double value of a field, return default value if the field is
      * not present.
      */
    public double getDoubleValue() {
        if (runtimeField != null) {
            try {
                return runtimeField.getDouble(node);
            } catch (Exception e) {
            }
        }

        if (staticField != null) {
            FieldValue fv = staticField.getFieldValue();
            TokenEnumerator te = ((Scene)node.getScene()).getTokenEnumerator();
            return( te.getDouble(fv.getFirstTokenOffset()) );
        }

        if ( defaultField != null ) {
            return Double.valueOf(defaultField.getDefaultValue()).doubleValue();
        }
    
        return 0.0;
    }

    /** Return an array of all the strings in a field, or null if the
      * field is not present.
      */
    public String[] getStringArray() {
        String[] strings = null;

        if (runtimeField != null) {
            try {
                return (String[]) runtimeField.get(node);
            } catch (Exception e) {
            }
        }

        if ( staticField != null ) {
            MFFieldValue val = (MFFieldValue) staticField.getFieldValue();
            int n = val.getRawValueCount();
            strings = new String[n];
            TokenEnumerator te = ((Scene)node.getScene()).getTokenEnumerator();
            for (int i = 0; i < n; i++) {
                FieldValue fv = val.getFieldValueAt(i);
                strings[i] = te.toString(fv.getFirstTokenOffset());
                // trim the double quotes if present
                if (( strings[i].charAt( 0 ) == '"' ) && 
                    ( strings[i].charAt( strings[i].length() - 1 ) == '"' )) {
                    strings[i] = strings[i].substring( 1, strings[i].length() - 1 );
                }
            }
            return strings;
        }
        if (defaultField != null) {
            strings = new String[1];
            strings[0] = defaultField.getDefaultValue();
        }
        return strings;
    }


    /** Return an array of all the ints representing a field, or the default
      * value if the field is not present.
      */
    public int[] getIntArray() {
        return getIntArray(-1);
    }

    /** Return a specified number of floats representing a field, or the default
      * value if the field is not present.
      */
    public int[] getIntArray( int n ) {
        int[] ints = null;
        
        if (runtimeField != null) {
            try {
                return (int[]) runtimeField.get(node);
            } catch (Exception e) {
            }
        }
 
        if ( staticField != null ) {
            MFFieldValue val = (MFFieldValue) staticField.getFieldValue();
            if (n == -1) {
                n = val.getRawValueCount();
            }
            ints = new int[n];
            TokenEnumerator te = ((Scene)node.getScene()).getTokenEnumerator();
            for (int i = 0; i < n; i++) {
                FieldValue fv = val.getFieldValueAt(i);
                ints[i] = te.getIntValue(fv.getFirstTokenOffset());
            }
            return ints;
        }

        if (defaultField != null) {        
            StringTokenizer tok = new StringTokenizer(defaultField.getDefaultValue(), delim);
            if (n == -1) {
                n = tok.countTokens();
            }
            ints = new int[n];
            for (int i = 0; i < n; i++) {
                ints[i] = Float.valueOf(tok.nextToken()).intValue();
            }
        }
        return ints;
    }

    /** Return an array of all the floats representing a field, or the default
      * value if the field is not present.
      */
    public float[] getFloatArray() {
        return getFloatArray(-1);
    }

    /** Return a specified number of floats representing a field, or the default
      * value if the field is not present.
      */
    public float[] getFloatArray( int n ) {
        float[] floats = null;
    
        if (runtimeField != null) {
            try {
                return (float[]) runtimeField.get(node);
            } catch (Exception e) {
            }
        }

        if ( staticField != null ) {
            TokenEnumerator te = ((Scene)node.getScene()).getTokenEnumerator();
            FieldValue fv = staticField.getFieldValue();
            int dim = getValueDimension(fv);
            if (fv instanceof MFFieldValue) {
                if (n == -1) {
                    n = fv.numberChildren();
                } else {
                    n /= dim;
                }
                int numfloats = dim * n;
                floats = new float[numfloats];
                int ix = 0;
                for (int i = 0; i < n && ix < numfloats; i++) {
                    int j = 0;
                    int jx = 0;
                    while (jx < dim && ix < numfloats) {
                        VrmlElement child = fv.getChildAt( j++ );
                        if ( child instanceof FieldValue ) {
                            floats[ix++] = te.getFloat(child.getFirstTokenOffset());
                            jx++;
                        }
                    }
                }
                return floats;

            } else if (fv instanceof SFFieldValue) {
                if (n == -1) {
                    n = dim;
                } else {
                    n = Math.min(n, dim);
                }
                floats = new float[n];
                if (n == 1) {
                    floats[0] = te.getFloat(fv.getFirstTokenOffset());
                } else {
                    for (int i = 0; i < n; i++) {
                        VrmlElement child = fv.getChildAt( i );
                        floats[i] = te.getFloat(child.getFirstTokenOffset());
                    }
                }
                return floats;
            }
        }
        
        if (defaultField != null) {        
            StringTokenizer tok = new StringTokenizer(defaultField.getDefaultValue(), delim);
            if (n == -1) {
                n = tok.countTokens();
            }
            floats = new float[n];
            for (int i = 0; i < n; i++) {
                floats[i] = Float.valueOf(tok.nextToken()).floatValue();
             }
        }
        return floats;
    }

    
    
    /** Return a tuple array, which is a two-dimensional array of floats
      */
    public float[][] getTupleArray() {
        float[][] tuples = null;
        int n;
        int dim;
    
        if (runtimeField != null) {
            try {
                return (float[][]) runtimeField.get(node);
            } catch (Exception e) {
            }
        }

        if ( staticField != null ) {
            TokenEnumerator te = ((Scene)node.getScene()).getTokenEnumerator();
            FieldValue fv = staticField.getFieldValue();
            dim = getValueDimension(fv);
            if (fv instanceof MFFieldValue) {
                n = fv.numberChildren() / dim;
                tuples = new float[n][dim];
                int ix = 0;
                for (int i = 0; i < n; i++) {
                    int j = 0;
                    for (j = 0; j < dim; j++) {
                        VrmlElement child = fv.getChildAt( ix++ );
                        tuples[i][j] = te.getFloat(child.getFirstTokenOffset());
                    }
                }
                return tuples;

            } else if (fv instanceof SFFieldValue) {
                tuples = new float[1][dim];
                for (int i = 0; i < dim; i++) {
                    VrmlElement child = fv.getChildAt( i );
                    tuples[0][i] = te.getFloat(child.getFirstTokenOffset());
                }
                return tuples;
            }
        }
       
        if (defaultField != null) {        
            dim = getValueDimension(defaultField);

            StringTokenizer tok = new StringTokenizer(defaultField.getDefaultValue(), delim);
            n = tok.countTokens() / dim;
            tuples = new float[n][dim];
            for (int i = 0; i < n; i++) {
                for (int j = 0; j < dim; j++) {
                    tuples[i][j] = Float.valueOf(tok.nextToken()).floatValue();
                }
            }
        }
        return tuples;
    }


    /** Return an array of all the doubles representing a field, or the default
      * value if the field is not present.
      */
    public double[] getDoubleArray() {
        return getDoubleArray(-1);
    }

    /** Return a specified number of doubles representing a field, or the default
      * value if the field is not present.
      */
    public double[] getDoubleArray( int n ) {
        double[] doubles = null;
    
        if (runtimeField != null) {
            try {
                return (double[]) runtimeField.get(node);
            } catch (Exception e) {
            }
        }

        if ( staticField != null ) {
            TokenEnumerator te = ((Scene)node.getScene()).getTokenEnumerator();
            FieldValue fv = staticField.getFieldValue();
            int dim = getValueDimension(fv);
            if (fv instanceof MFFieldValue) {
                if (n == -1) {
                    n = fv.numberChildren();
                } else {
                    n /= dim;
                }
                int numdoubles = dim * n;
                doubles = new double[numdoubles];
                int ix = 0;
                for (int i = 0; i < n && ix < numdoubles; i++) {
                    int j = 0;
                    int jx = 0;
                    while (jx < dim && ix < numdoubles) {
                        VrmlElement child = fv.getChildAt( j++ );
                        if ( child instanceof FieldValue ) {
                            doubles[ix++] = te.getDouble(child.getFirstTokenOffset());
                            jx++;
                        }
                    }
                }
                return doubles;

            } else if (fv instanceof SFFieldValue) {
                if (n == -1) {
                    n = dim;
                } else {
                    n = Math.min(n, dim);
                }
                doubles = new double[n];
                if (n == 1) {
                    doubles[0] = te.getDouble(fv.getFirstTokenOffset());
                } else {
                    for (int i = 0; i < n; i++) {
                        VrmlElement child = fv.getChildAt( i );
                        doubles[i] = te.getDouble(child.getFirstTokenOffset());
                    }
                }
                return doubles;
            }
        }
        
        if (defaultField != null) {        
            StringTokenizer tok = new StringTokenizer(defaultField.getDefaultValue(), delim);
            if (n == -1) {
                n = tok.countTokens();
            }
            doubles = new double[n];
            for (int i = 0; i < n; i++) {
                doubles[i] = Double.valueOf(tok.nextToken()).doubleValue();
            }
        }
        return doubles;
    }

    
    // helper functions for getting values of various types from field values
    // of various types
    
    /** extract one or more floats from a field value according to its type */
    int getFieldValueFloats( float[] floats, int ix, FieldValue fv, TokenEnumerator te) {
        int n = 0;       
        if (fv instanceof MFFieldValue) {           // get an array of floats
            MFFieldValue val = (MFFieldValue) fv;
            int count = val.getRawValueCount();
            for (int i = 0; i < count && (ix + n) < floats.length; i++) {
                n += getFieldValueFloats( floats, ix + n, val.getFieldValueAt(i), te);
            }

        } else {
            // get the appropriate number of values for the type
            n = getValueDimension(fv);
            for ( int i = 0; i < n; i++ ) {
                VrmlElement vle = fv.getChildAt( i );
                floats[ix + i] = te.getFloat(vle.getFirstTokenOffset());
            }
        }
        return n;
    }
            
    // get the number of values for a single item of a given type
    static int getValueDimension(FieldValue fv) {
        if (fv instanceof SFVec2fValue || fv instanceof MFVec2fValue) {
            return 2;
        } else if (fv instanceof SFVec3fValue || fv instanceof MFVec3fValue) {
            return 3;
        } else if (fv instanceof SFColorValue || fv instanceof MFColorValue) {
            return 3;
        } else if (fv instanceof SFRotationValue || fv instanceof MFRotationValue) {
            return 4;
        } else {
            return 1;
        }
    }
    
    // get the number of values for a single item of a given type given its FieldDescriptor
    static int getValueDimension(FieldDescriptor fd) {
        int type = fd.getFieldType();
        if (type == VRML97.SFVec2f || type == VRML97.MFVec2f) {
            return 2;
        } else if (type == VRML97.SFVec3f || type == VRML97.MFVec3f) {
            return 3;
        } else if (type == VRML97.SFColor || type == VRML97.MFColor) {
            return 3;
        } else if (type == VRML97.SFRotation || type == VRML97.MFRotation) {
            return 4;
        } else {
            return 1;
        }
    }


    public Object getValue() throws NoSuchFieldException, IllegalAccessException {
        Object obj = null;
        if (runtimeField != null) {
            obj = runtimeField.get(node);
            
        } else if (staticField != null) {            
            FieldValue fv = staticField.getFieldValue();
            obj = getFieldValueObject(fv);

        } else {
            throw new NoSuchFieldException("getValue() only works with runtime fields; no runtime field '" + name + "' defined.");
        }
        return obj;
    }
       

    public void setValue(Object value) {
        
        if (runtimeField != null) {
            try {
//if ( value == null ) System.out.println( "value at 1 is null" );
                if (value instanceof Node) {
                    Node childnode = (Node) value;
                    if (childnode.isDEForUSE()) {
                        value = childnode.getNode();
//if ( value == null ) System.out.println( "value at 2 is null" );
                    }
                }
//if ( node == null ) System.out.println( "node is null" );
//if ( value == null ) System.out.println( "value is null" );
                runtimeField.set(node, value);
            } catch (Exception e) {
                System.out.println("Unable to set value of field " + name + ": " + e);           
e.printStackTrace();
            }
        } else if (staticField != null) {
            try {
                String fieldName = staticField.getName();
                String preSeparator = " ";
                String postSeparator = " ";
                if (staticField instanceof MFField) {
                    preSeparator = " [ ";
                    postSeparator = " ] ";
                } else {
                    preSeparator = postSeparator = " ";
                }
                System.out.println( "node.setField: " + fieldName + preSeparator + value + postSeparator ); 
                node.setField(fieldName + preSeparator + value + postSeparator);
                // reset the field
                staticField = node.getField(fieldName);
                
            } catch (InvalidFieldException ife) {
            }
        }
    }

    public void set1Value(int index, Object value) {
        if (runtimeField != null) {
            try {
                Array.set(runtimeField.get(node), index, value);
            } catch (IllegalAccessException iae) {
            }
        } else if (staticField != null) {
            // don't have a good way of doing this yet
            System.out.println("Cannot dynamically alter single elements of this field");
        }
    }


    private transient double lastUpdate = 0.0;

    
    public void propagateValue(VrmlPlayer player, Object value) {
        double eventTime = player.getEventTime();
        player.beginRoute(eventTime);

        setValue(value);
        if (node != null && node instanceof RealizableNode) {
            RealizableNode rnode = (RealizableNode) node;
            rnode.updateField(name, eventTime);
        }
            
        if (connections != null) {
            // Test time to avoid sending two events with same timestamp from
            // a single EventOut
            if (lastUpdate != eventTime) {
                lastUpdate = eventTime;

                Enumeration e = connections.elements();
                if (e != null) {
                    while (e.hasMoreElements()) {
                        RuntimeField target = (RuntimeField) e.nextElement();
                        target.propagateValue(player, value);
                    }
                }
            }
        }
        player.endRoute();
    }

    
    public void connectToField(RuntimeField field) {
        if (connections == null) {
            connections = new Vector();
        }
        if (!connections.contains(field)) {
            connections.addElement(field);
        }
    }

    public void deleteConnection(RuntimeField field) {
        connections.removeElement(field);
    }
    
  
   
    /** static method for constructing a standard key for a field */
    static public String getRuntimeName(Node node, String fieldName) {
        String name = node.toString();
        int n = name.lastIndexOf('.');
        if (n > 0) {
            name = name.substring(n + 1);
        }
        return (name + '.' + fieldName);
    }

    // convenience routines for converting tuple fields into strings
    static public String valueString(Object val) {
        if (val == null) {
            return "NULL";
        } else if (val instanceof float[]) {
            return tupleString((float[]) val);
        } else if (val instanceof float[][]) {
            return tupleString((float[][]) val);
        } else if (val instanceof int[]) {
            return tupleString((int[]) val);
        } else {
            return val.toString();
        }
    }
      
    static public String tupleString(float[][] f) {
        return tupleString(f, (f == null || f.length == 0 ? 0 : f[0].length));
    }
    static public String tupleString(float[][] f, int size) {
        if (f == null || f.length == 0 || size == 0) {
            return("# empty ");
        }
        StringBuffer ts = new StringBuffer("[ ");

        for (int i = 0; i < f.length; i++) {
            ts.append(Float.toString(f[i][0]));
            for (int j = 1; j < size; j++) {
                ts.append(" ");
                ts.append(Float.toString(f[i][j]));
            }
            if (i < f.length - 1) {
                ts.append(", ");
            }
        }
        ts.append(" ]");
        return ts.toString();
    }
    static public String tupleString(float[] f) {
        return tupleString(f, f.length);
    }
    static public String tupleString(float[] f, int size) {
        if (f == null || size == 0) {
            return("# empty ");
        }

        StringBuffer ts = new StringBuffer();
        ts.append(Float.toString(f[0]));

        for (int i = 1; i < size; i++) {
            ts.append(" ");
            ts.append(Float.toString(f[i]));
        }
        return ts.toString();
    }
    static public String tupleString(int[] ints) {
        if (ints == null || ints.length == 0) {
            return("# empty ");
        }

        StringBuffer ts = new StringBuffer( "[ " );
        ts.append(Integer.toString(ints[0]));

        for (int i = 1; i < ints.length; i++) {
            ts.append(" ");
            ts.append(Integer.toString(ints[i]));
        }
        ts.append( " ]" );
        return ts.toString();
    }


}

class FloatField extends RuntimeField {
    public float floatValue;
    public FloatField(float f) {
        super("floatValue");
        floatValue = f;
    }
}

class DoubleField extends RuntimeField {
    public double doubleValue;
    public DoubleField(double d) {
        super("doubleValue");
        doubleValue = d;
    }
}

class IntField extends RuntimeField {
    public int intValue;
    public IntField(int i) {
        super("intValue");
        intValue = i;
    }
}

class BooleanField extends RuntimeField {
    public boolean boolValue;
    public BooleanField(boolean b) {
        super("boolValue");
        boolValue = b;
    }
}

class StringField extends RuntimeField {
    public String stringValue;
    public StringField(String s) {
        super("stringValue");
        stringValue = s;
    }
}

class BaseNodeField extends RuntimeField {
    public BaseNode baseNode = null;
    public BaseNodeField(BaseNode node) {
        super("baseNode");
        baseNode = node;
    }
}

class ArrayField extends RuntimeField {
    public Object valueArray;
    public ArrayField(Object objArray) {
        super("valueArray");

        int len =  Array.getLength(objArray);
        // is this right?
        valueArray = Array.newInstance(objArray.getClass().getComponentType(), len);
        // how about this?
        System.arraycopy(objArray, 0, valueArray, 0, len);
    }
}

class ObjectField extends RuntimeField {
    public Object obj = null;
    public ObjectField(Object o) {
        super("obj");
        obj = o;
    }
}
  
    