/*
 * @(#)ChiselResources.java
 *
 * Copyright (c) 1998 by Trapezium Development LLC.  All Rights Reserved.
 *
 * The information in this file is the property of Trapezium Development LLC
 * and may be used only in accordance with the terms of the license granted
 * by Trapezium.
 *
 */
package com.trapezium.chisel;

import java.util.*;
import java.io.File;
import java.awt.Image;
import java.awt.Toolkit;
import java.net.URL;

/**
 *  Localizable resources for Chisel.
 *
 *  @author          Michael St. Hippolyte
 *  @version         1.1, 29 Jan 1998
 *
 *  @since           1.0
 */
public class ChiselResources extends ListResourceBundle {

    public final static int LOADFROMFILE = 0;
    public final static int LOADFROMCLASSPATH = 1;

    public static int loadMode = LOADFROMCLASSPATH;
    public static void setLoadFromJar(boolean loadFromJar) {
        loadMode = loadFromJar ? LOADFROMCLASSPATH : LOADFROMFILE;
    }


    //
    // private stuff
    //

	private static ResourceBundle resources;
    static {
		System.out.println("Loading resources...");
		try {
            resources = ResourceBundle.getBundle("com.trapezium.chisel.ChiselResources", Locale.getDefault());
        } catch (MissingResourceException mre) {
            System.err.println("Chisel resource class not found");
   			System.exit(1);
        }
    }

	public static ResourceBundle getDefaultResourceBundle() {
		return resources;
	}

    public static String getResourceString(String nm) {
        String str;
        try {
            str = resources.getString(nm);
        } catch (MissingResourceException mre) {
            str = null;
        }
        return str;
    }

    /** Get a URL for a resource, look in JAR file if possible, or disk if not */
    public static URL getResource(String key) {
        String name = getResourceString(key);
        return( getResourceByName( name ));
    }
    
    public static URL getResourceByName( String name ) {
        if (name != null) {
            URL url = resourceURL(name, loadMode);
            // couldn't find it?  look in the other possible place
            if (url == null) {
                url = resourceURL(name, loadMode == LOADFROMCLASSPATH ? LOADFROMFILE : LOADFROMCLASSPATH);
            }
            return url;
        }
        return null;
    }
    
    public static URL getResourceByName( String name, int mode ) {
        if (name != null) {
            return resourceURL(name, mode);
        } else {
            return null;
        }
    }
        

    private static URL base = null;
    private static String resourceName(String file) {
        if (file.startsWith("~")) {
            file = System.getProperty("user.home") + file.substring(1);
        }
        return "file:" + (new File(file)).getAbsolutePath().replace(File.separatorChar, '/');
    }

    private static String resourceURLString(String file) {
        URL url = resourceURL(file);
        if (url == null) {
            return null;
        }
        return url.toExternalForm();
    }

    /** Get a URL of a file using the current load mode -- LOADFROMCLASSPATH or LOADFROMDISK */
    public static URL resourceURL(String file) {
        return resourceURL(file, loadMode);
    }
    
    public static URL resourceURL(String file, int mode) {
        if (mode == LOADFROMCLASSPATH) {
//            System.out.println( "LOADFROMCLASSPATH trying resource '/" + file + "'" );
            return file.getClass().getResource("/" + file);
        } else {
            String resource = resourceName(file);
//            System.out.println( "LOADFROMFILE converted '" + file + "' to '" + resource + "'" );
            URL url;
            try {
                url = new URL(base, resource);
            } catch (Exception e) {
                url = null;
            }
            return url;
        }
    }

    /**
     * Suffix applied to the key used in resource file
     * lookups for an image.
     */
    private static final String imageSuffix = "Image";

    /**
     * Suffix applied to the key used in resource file
     * lookups for a label.
     */
    private static final String labelSuffix = "Label";

    /**
     * Suffix applied to the key used in resource file
     * lookups for an action.
     */
    private static final String actionSuffix = "Action";

    /**
     * Suffix applied to the key used in resource file
     * lookups for tooltip text.
     */
    private static final String tipSuffix = "Tooltip";
    /**
     * Suffix applied to the key used in resource file
     * lookups for initial item state.
     */
    private static final String stateSuffix = "State";


    public static String getImageName(String key) {
        return getResourceString(key + imageSuffix);
    }

    public static String getLabel(String key) {
        return getResourceString(key + labelSuffix);
    }

    public static String getActionCommand(String key) {
        return getResourceString(key + actionSuffix);
    }

    public static String getTip(String key) {
        return getResourceString(key + tipSuffix);
    }

    public static String getState(String key) {
        return getResourceString(key + stateSuffix);
    }

    static Image chiselImage = null;
    public static Image getImage(String key) {
        if ( key.compareTo( "chisel" ) == 0 ) {
            if ( chiselImage != null ) {
                return( chiselImage );
            }
        }
        URL url = getResource(key + imageSuffix);
        if (url == null) {
            System.out.println("No image resource for " + key);
            return null;
        } else {
            //System.out.println("Loading image resource for " + key + " URL: " + url.toExternalForm());
            Image image = null;
            try {
                image = Toolkit.getDefaultToolkit().createImage((java.awt.image.ImageProducer) url.getContent());
            } catch (Exception e) {
                System.out.println("Error loading image " + url.toExternalForm());
                image = null;
            }
            if ( key.compareTo( "chisel" ) == 0 ) {
                chiselImage = image;
            }
            return image;
        }
    }

    public Object[][] getContents() {
        return contents;
    }

    static final Object[][] contents = {
        // icon
        { "chiselImage", "images/chisel.gif" },
        
        //{ "menubar", "file edit action" },
        { "menubar", "file edit process view help" },

        { "file", "open fetch - save saveas gzipsave gzipsaveas - genhtml - exit" },
        //{ "file", "new open fetch - save saveas gzipsave gzipsaveas - exit" },
        { "fileLabel", "File" },
        { "openLabel", "Open..." },
        { "openImage", "images/open.gif" },
        { "openAction", "open" },
        { "openTooltip", "Open a file" },
        { "newLabel", "New" },
        { "newImage", "images/new.gif" },
        { "newAction", "new" },
        { "newTooltip", "Create a new file" },
        { "saveLabel", "Save" },
        { "saveImage", "images/save.gif" },
        { "saveAction", "save" },
        { "saveTooltip", "Save current file" },
        { "saveasLabel", "Save As..." },           // lame
        { "saveasImage", "images/saveas.gif" },
        { "saveasTooltip", "Save under new name"},
        { "saveAsLabel", "Save As" },
        { "saveAsImage", "images/saveas.gif" },
        { "saveAsAction", "saveas" },
        { "gzipsaveLabel", "GZIP Save" },       // lame
        { "gzipsaveImage", "images/gzipsave.gif" },
        { "gzipsaveTooltip", "Save gzipped"},
        { "gzipSaveLabel", "GZIP Save" },
        { "gzipSaveImage", "images/gzipsave.gif" },
        { "gzipSaveAction", "gzipsave" },
        { "gzipsaveasLabel", "GZIP Save As..." },  // really lame
        { "gzipsaveasImage", "images/gzipsaveas.gif" },
        { "gzipsaveasTooltip", "Save gzipped under new name"},
        { "gzipSaveAsLabel", "GZIP Save As" },
        { "gzipSaveAsImage", "images/gzipsaveas.gif" },
        { "gzipSaveAsAction", "gzipsaveas" },
        { "fetchLabel", "Fetch..." },
        { "fetchImage", "images/fetch.gif" },
        { "fetchAction", "fetch" },
        { "fetchTooltip", "Retrieve a URL" },
        { "genhtmlLabel", "Generate HTML report" },
        { "genhtmlAction", "genhtml" },
        { "exitLabel", "Exit" },
        { "exitImage", "images/exit.gif" },
        { "exitAction", "exit" },

        { "action", "chisel" },
        { "actionLabel", "Action" },
        { "chiselLabel", "Run chisels" },
        { "chiselImage", "images/chisel.gif" },

        { "process", "stop validate format clean condense reduce reorganize" },
        { "processLabel", "Process" },

        { "edit", "undo redo - copy" },
        { "editLabel", "Edit" },
        { "undoLabel", "Undo" },
        { "undoAction", "undo-last-command" },
        { "undoImage", "images/undo.gif" },
        { "undoTooltip", "Undo" },
        { "redoLabel", "Redo" },
        { "redoAction", "redo-last-command" },
        { "redoImage", "images/redo.gif" },
        { "redoTooltip", "Redo" },
        { "copyLabel", "Copy" },
        { "copyAction", "copy-to-clipboard" },
        { "copyImage", "images/copy.gif" },
        { "copyTooltip", "Copy selection to clipboard" },

        // not used right now
        { "cutLabel", "Cut" },
        { "cutAction", "cut-to-clipboard" },
        { "cutImage", "images/cut.gif" },
        { "cutTooltip", "Move selection to clipboard" },
        { "pasteLabel", "Paste" },
        { "pasteAction", "paste-from-clipboard" },
        { "pasteImage", "images/paste.gif" },
        { "pasteTooltip", "Paste clipboard to selection" },

        { "stopLabel", "Stop" },
        { "stopAction", "stop" },
        { "stopImage", "images/stop.gif" },
        { "stopTooltip", "Stop the current process" },
        { "validateLabel", "Validate" },
        { "validateAction", "validate" },
        { "validateImage", "images/validate.gif" },
        { "validateTooltip", "Look for errors in current world" },
        { "formatLabel", "Format" },
        { "formatAction", "format" },
        { "formatImage", "images/format.gif" },  // note, no such image
        { "formatTooltip", "Format text in current world" },
        { "cleanLabel", "Clean" },
        { "cleanAction", "clean" },
        { "cleanImage", "images/clean.gif" },
        { "cleanTooltip", "Remove unnecessary and unusable code" },
        { "condenseLabel", "Condense" },
        { "condenseAction", "condense" },
        { "condenseImage", "images/condense.gif" },
        { "condenseTooltip", "Reduce the size of the current file" },
        { "reduceLabel", "Reduce" },
        { "reduceAction", "reduce" },
        { "reduceImage", "images/reduce.gif" },
        { "reduceTooltip", "Reduce the number of polygons in the current world" },
//        { "mutateLabel", "Mutate" },
//        { "mutateAction", "mutate" },
//        { "mutateImage", "images/mutate.gif" },
//        { "mutateTooltip", "Modify the geometry in the current world" },
        { "reorganizeLabel", "Reorganize" },
        { "reorganizeAction", "reorganize" },
        { "reorganizeImage", "images/reorganize.gif" },
        { "reorganizeTooltip", "Reorganize the nodes and files that make up the current world" },

        { "view", "preverror nexterror - options" },
        { "viewLabel", "View" },
        { "showtoolbarLabel", "Toolbar" },
        { "showtoolbarAction", "show-toolbar" },
        { "showtoolbarState", "checked" },
        { "showstatusLabel", "Status line" },
        { "showstatusAction", "show-statusline" },
        { "showstatusState", "checked" },
        { "preverrorLabel", "Previous Error" },
        { "preverrorAction", "prev-error" },
        { "preverrorImage", "images/preverror.gif" },
        { "preverrorTooltip", "Go to previous error" },
        { "nexterrorLabel", "Next Error" },
        { "nexterrorAction", "next-error" },
        { "nexterrorImage", "images/nexterror.gif" },
        { "nexterrorTooltip", "Go to next error" },
        { "optionsLabel", "Options" },
        { "optionsAction", "options" },
        { "optionsImage", "images/options.gif" },
        { "optionsTooltip", "Set font size and other options" },

        { "help", "about - vcheck" },
        { "helpLabel", "Help" },
        { "aboutLabel", "About..." },
        { "aboutAction", "about" },
        { "vcheckLabel", "Check version..." },
        { "vcheckAction", "vcheck" },
        
        { "unlockmenu", "unlock" },
        { "unlockmenuLabel", "Unlock" },
        { "unlockAction", "unlock" },

        { "toolbar", "open fetch - save saveas gzipsave gzipsaveas - copy - redo undo - stop - preverror nexterror" },
        //{ "toolcell", "open fetch" },

        // button labels for dialogs
        { "okLabel", "OK" },
        { "cancelLabel", "Cancel" },
        { "purchaseLabel", "Purchase..." },
        { "unlockLabel", "Unlock..." },

        // specific dialog labels
        { "fetchDialogTitle", "Retrieve URL" },
        { "urlLabel", "URL:" },
        { "recurseLabel", "Retrieve referenced files" },
        { "unlockDialogTitle", "Unlock Chisel" },
        { "unlockLine0", "Please enter the following to verify your" },
        { "unlockLine1", "registration.  Your key may be found in" },
        { "unlockLine2", "the email message you received when" },
        { "unlockLine3", "you purchased Chisel." },
        { "unlockFailedLine0", "One or more of fields you entered could not be" },
        { "unlockFailedLine1", "validated; please reenter." },
        { "unlockFailedLine2", " " },
        { "unlockFailedLine3", "If you are sure all the information you entered" },
        { "unlockFailedLine4", "is correct, contact support@com.trapezium.com" },
        { "nameLabel", "Name:" },
        { "emailLabel", "Email:" },
        { "keyLabel", "Key:" },
        { "optionsDialogTitle", "Options" },
        { "font.nameLabel", "Font" },
        { "font.styleLabel", "Style" },
        { "font.sizeLabel", "Size" },
        { "loadWorkspaceLabel", "Reload workspace on launch" },
        { "saveInUnixFormat", "Save in Unix format" },
        { "automaticClean", "Automatically clean after processing" },


		// adapted from mpedit list
        { "PromptFont", "Font:" },
        { "DialogGoto", "Go To" },
        { "TextDirty", "This file has not been saved." },
        { "DialogFind", "Find" },
        { "MenuRedo", "Redo" },
        { "DialogSaveAs", "Save file as" },
        { "MenuClose", "Close" },
        { "MenuOptions", "Options" },
        { "MenuOpen", "Open" },
        { "DialogNotFound", "String Not Found" },
        { "ChoiceBold", "Bold" },
        { "MenuFile", "File" },
        { "DialogReplace", "Find and Replace" },
        { "MenuSaveAs", "Save As" },
        { "PromptGoto", "Enter Line Number:" },
        { "ButtonSave", "Save" },
        { "PromptReplace", "Replace:" },
        { "ButtonCancel", "Cancel" },
        { "ChoiceBoldItalic", "Bold Italic" },
        { "PromptSize", "Size:" },
        { "MenuWindow", "Window" },
        { "MenuSave", "Save" },
        { "MenuNew", "New" },
        { "MenuHelp", "Help" },
        { "ButtonGoto", "Go" },
        { "ChoiceItalic", "Italic" },
        { "ButtonClose", "Close" },
        { "MenuGoto", "Go To" },
        { "UntitledDoc", "Untitled" },
        { "MenuProps", "Properties" },
        { "PromptFind", "Find:" },
        { "ChoicePlain", "Plain" },
        { "MenuPrint", "Print" },
        { "MenuEdit", "Edit" },
        { "PromptTab", "Tab Size:" },
        { "MenuFind", "Find" },
        { "MenuAbout", "About" },
        { "ButtonReplace", "Replace" },
        { "MenuCloneWin", "Clone Window" },
        { "MenuUndo", "Undo" },
        { "PromptStyle", "Style:" },
        { "MenuCut", "Cut" },
        { "MenuPaste", "Paste" },
        { "MenuReplace", "Replace" },
        { "PromptLocal", "Apply changes to:" },
        { "ButtonDontSave", "Don't Save" },
        { "DialogProps", "Properties" },
        { "ButtonOk", "OK" },
        { "ButtonFindNext", "Find Next" },
        { "DialogDirty", "Warning!" },
        { "DialogOpen", "Open text file" },
        { "MenuCopy", "Copy" },
        { "CheckThisDoc", "This Document" },
        { "PromptSyntax", "Use Syntax Coloring When Possible" },
        //{ "DialogAbout", "About mpEDIT" },
        { "CheckAllDocs", "All Documents" },
        { "UnbalancedCharacter", "Unbalanced character" },
        { "NoAnchors", "No anchors dropped at the moment" },
        { "InvalidAnchor", "Anchor no longer is valid" },
        { "NoPreviousAnchor", "No previous anchor" },
        { "action.anchor-drop.name", "Drop Anchor" },
        { "action.anchor-drop.short", "Drops anchor at given line, so you can later return there" },
        { "action.anchor-goto-last.name", "Last anchor.." },
        { "action.anchor-goto-last.short", "Jumps to last anchor" },
        { "action.anchor-goto-dialog.name", "Goto anchor.." },
        { "action.anchor-goto-dialog.short", "Pops up dialog with anchors you can jump to" },
        { "action.selection-indent.name", "Indent block" },
        { "action.selection-indent.short", "Adds tab before each line of selected region" },
        { "action.selection-unindent.name", "Unindent block" },
        { "action.selection-unindent.short", "Deletes first character of each line of selected regionaction.mode-readonly-switch.name", "ReadOnly" },
        { "action.mode-readonly-switch.short", "Switches between normal and read-only modes" },
        { "action.line-swap.name", "Swap lines" },
        { "action.line-swap.short", "Swaps current and next line" },
        { "action.cursor-up.name", "Cursor Up" },
        { "action.cursor-up.short", "Moves cursor one line up" },
        { "action.cursor-down.name", "Cursor Down" },
        { "action.cursor-down.short", "Moves cursor one line down" },
        { "action.cursor-forward.name", "Cursor Forward" },
        { "action.cursor-forward.short", "Moves cursor one character forward" },
        { "action.cursor-backward.name", "Cursor Backward" },
        { "action.cursor-backward.short", "Moves cursor one character backward" },
        { "action.cursor-word-forward.name", "Next Word" },
        { "action.cursor-word-forward.short", "Moves cursor to begin of next word" },
        { "action.cursor-word-backward.name", "Previous Word" },
        { "action.cursor-word-backward.short", "Moves cursor to begin of previous word" },
        { "action.cursor-line-begin.name", "Line Begin" },
        { "action.cursor-line-begin.short", "Moves cursor to begin of current line" },
        { "action.cursor-line-end.name", "Line End" },
        { "action.cursor-line-end.short", "Move cursor to end of current line" },
        { "action.cursor-page-begin.name", "Page Begin" },
        { "action.cursor-page-begin.short", "Moves cursor to begin of current page" },
        { "action.cursor-page-end.name", "Page End" },
        { "action.cursor-page-end.short", "Moves cursor to end of current page" },
        { "action.cursor-document-begin.name", "Document Begin" },
        { "action.cursor-document-begin.short", "Moves cursor to begin of document" },
        { "action.cursor-document-end.name", "Document End" },
        { "action.cursor-document-end.short", "Moves cursor to end of document" },
        { "action.page-up.name", "Page Up" },
        { "action.page-up.short", "Moves viewport one page up together with cursor" },
        { "action.page-down.name", "Page Down" },
        { "action.page-down.short", "Moves viewport one page down together with cursor" },
        { "action.find-next-forward.name", "Find Next" },
        { "action.find-next-forward.short", "Find next matching pattern in text" },
        { "action.find-next-backward.name", "Find Prev" },
        { "action.find-next-backward.short", "Find previous matching pattern in text" },
        { "action.brace-match-forward.name", "Forward Match" },
        { "action.brace-match-forward.short", "Seeks for matching brace or character forward" },
        { "action.brace-match-backward.name", "Backward Match" },
        { "action.brace-match-backward.short", "Seeks for matching brace ot character backward" },
        { "action.character-delete-forward.name", "Delete" },
        { "action.character-delete-forward.short", "Delete next character" },
        { "action.character-delete-backward.name", "Backspace" },
        { "action.character-delete-backward.short", "Delete previous character" },
        { "action.line-break.name", "Newline" },
        { "action.line-break.short", "Insert newline with possible indent" },
        { "action.line-clone.name", "Clone line" },
        { "action.line-clone.short", "Insert identical line after current" },
        { "action.line-delete.name", "Delete line" },
        { "action.line-delete.short", "Delete current line" },
        { "action.document-save.name", "Save" },
        { "action.document-save.short", "Saves document under current name" },
        { "action.frame-close.name", "Close" },
        { "action.frame-close.short", "Closes window" },
        { "action.mode-autoindent-switch.name", "Autindent On/Off" },
        { "action.mode-autoindent-switch.short", "Turns autoindenting on or off" },
        { "action.document-new.name", "New File" },
        { "action.document-new.short", "Creates new frame with empty document" },
        { "action.document-open-dialog.name", "Open File.." },
        { "action.document-open-dialog.short", "Opens file in a new frame" },
        { "action.document-save-as-dialog.name", "Save As.." },
        { "action.document-save-as-dialog.short", "Saves document with a new name" },
        { "action.document-print-dialog.name", "Print.." },
        { "action.document-print-dialog.short", "Prints current file" },
        { "action.find-dialog.name", "Find.." },
        { "action.find-dialog.short", "Seeks for given string" },
        { "action.goto-dialog.name", "Goto Line.." },
        { "action.goto-dialog.short", "Jumps to specified line number" },
        { "action.replace-dialog.name", "Replace.." },
        { "action.replace-dialog.short", "Searches and replaces strings" },
        { "action.selection-copy.name", "Copy" },
        { "action.selection-copy.short", "Copies current selection to buffer" },
        { "action.selection-cut.name", "Cut" },
        { "action.selection-cut.short", "Cuts current selection and puts into buffer" },
        { "action.buffer-paste.name", "Paste" },
        { "action.buffer-paste.short", "Pastes data from clipboard buffer" },
        { "action.properties-dialog.name", "Properties.." },
        { "action.properties-dialog.short", "Shows up properties dialog" },
        { "action.frame-clone.name", "Clone Window" },
        { "action.frame-clone.short", "Clones current window" },
        { "action.help-about-dialog.name", "About" },
        { "action.help-about-dialog.short", "Shows 'About' help dialog" },
        { "action.keytable-save.name", "Save Keytable" },
        { "action.keytable-save.short", "Saves keytable to external file" },
        { "action.keytable-load.name", "Load Keytable" },
        { "action.keytable-load.short", "Loads keytable from external file" },
        { "action.mode-readonly-switch.name", "Readonly" },
        { "action.mode-readonly-switch.short", "Switches between normal and read-only modes" },
        { "action.undo.name", "Undo" },
        { "action.undo.short", "Cancels last operation" },
        { "action.redo.name", "Redo" },
        { "action.redo.short", "Redoes last undone operation" },
        { "action.cursor-select-forward.name", "Select character forward" },
        { "action.cursor-select-forward.short", "Select character forward" },
        { "action.cursor-select-backward.name", "Select character backward" },
        { "action.cursor-select-backward.short", "Select character backward" },
        { "action.cursor-select-up.name", "Select line up" },
        { "action.cursor-select-up.short", "Select line up" },
        { "action.cursor-select-down.name", "Select line down" },
        { "action.cursor-select-down.short", "Select line down" },
        { "action.cursor-select-word-backward.name", "Select word backward" },
        { "action.cursor-select-word-backward.short", "Select word backward" },
        { "action.cursor-select-word-forward.name", "Select word forward" },
        { "action.cursor-select-word-forward.short", "Select word forward" },
        { "action.cursor-select-document-begin.name", "Select to beginning of document" },
        { "action.cursor-select-document-begin.short", "Select to beginning of document" },
        { "action.cursor-select-document-end.name", "Select to end of document" },
        { "action.cursor-select-document-end.short", "Select to end of document" },
        { "action.cursor-select-line-begin.name", "Select to beginning of line" },
        { "action.cursor-select-line-begin.short", "Select to beginning of line" },
        { "action.cursor-select-line-end.name", "Select to end of line" },
        { "action.cursor-select-line-end.short", "Select to end of line" },
        { "action.next-error.name", "Next error" },
        { "action.next-error.short", "Go to next error line" },
        { "action.prev-error.name", "Previous error" },
        { "action.prev-error.short", "Go to previous error line" },
    };
}
