/*
 * @(#)Generator.java
 *
 * Copyright (c) 1999 by Trapezium Development LLC.  All Rights Reserved.
 *
 * The information in this file is the property of Trapezium Development LLC
 * and may be used only in accordance with the terms of the license granted
 * by Trapezium.
 *
 */

package com.trapezium.generator;

import com.trapezium.parse.*;
import com.trapezium.vrml.grammar.VRML97parser;
import com.trapezium.vrml.Scene;

import java.io.*;

public abstract class Generator implements Runnable, java.io.Serializable {
    
    /** construct a generator given a name.  The name could either be
     *  a serialized generator or a class name.
     */
    public static Generator getGenerator( String name ) {
        Generator gen = null;

        // strip off the prefix if present to get file or class name
        if (name.startsWith("gen:")) {
            name = name.substring(4);
        } else if (name.startsWith("generator:")) {
            name = name.substring(10);
        }
        
        try {
            System.out.println("Attempting to load previously saved instance of " + name + "...");
            ObjectInputStream ois = new ObjectInputStream( new FileInputStream( name ) );
            gen = (Generator) ois.readObject();
            ois.close();
            System.out.println("Previously saved instance of " + name + " loaded." );
        
        } catch (Exception efile) {
            
            System.out.println(name + " not a valid file, checking if valid class name");
            try {
                Class genclass = Class.forName( name ); 
                gen = (Generator) genclass.newInstance();
                System.out.println("Instance of " + name + " created with default properties." );
    			
    	    } catch (Exception eclass) {
    	        gen = null;
                System.out.println(name + " not a recognized class name; returning null");
    	    }
        }
        return gen;
    }
    
    
    /** empty constructor */
    public Generator() {
    }

    abstract public void generate( PrintWriter out );

    /** generate up to a particular nesting depth, used by container
        generators */

    public void generate( PrintWriter out, int level ) {
        generate( out );
    }

    public void generateFile( String fileName ) {
        try {
            PrintWriter pw = new PrintWriter( new OutputStreamWriter( new FileOutputStream( fileName ) ) );
            generate( pw );
            pw.flush();
            pw.close();
        } catch ( Exception e ) {
            System.out.println( "Error generating file: " + e );
        }
    }

    public String getUrl() {
        return ("generator:" + getClass().getName());
    }


    public void run() {
        getScene();
    }


    public InputStream getInputStream() {
        InputStream in;
        try {
   	        //System.out.print("Launching generator " + getClass().getName());
            GeneratorThread t = new GeneratorThread(this);
            t.start();
			
            // wait for the generator to start cranking
  	        Thread.sleep(10);
            in = t.getInputStream();
	        while (in == null) {
                //System.out.print(".");
 		        Thread.sleep(100);
		        in = t.getInputStream();
	        }
  	        //System.out.println("...");

        } catch (Exception e) {
            System.out.println(" Unable to get input stream from generator");
            in = null;
        }

        return in;
    }
        

    public Scene generateScene() {
      	InputStream in = getInputStream();
    	if (in == null) {
   	        return null;
    	}

        Scene scene;

        try {
    	    if (in.available() == 0) {
  	            //System.out.print("waiting for scene creation");
    	        while (in.available() == 0) {
                    //System.out.print(".");
		            Thread.sleep(500);
   	            }
      	        //System.out.println("...");
	        }
			
	        //System.out.println(in.available() + " bytes available, proceeding to read");
			
            String pseudoUrl = getUrl();
            TokenEnumerator te = new TokenEnumerator(in, pseudoUrl);
            scene = new Scene( pseudoUrl, te );
            VRML97parser parser = new VRML97parser();
            parser.Build( te, scene, -1 );
        
//DumpVisitor dv = new DumpVisitor(System.out, scene.getTokenEnumerator());
//scene.traverse( dv );

        } catch (Exception e) {
            System.out.println("Unable to create scene: " + e);
            scene = null;
        }
        return scene;
    }

    // if fileName is nonnull, then the generator first tries to load the 
    // object from the file.  If fileName is null, or the file doesn't exist,
    // the object is generated procedurally

    transient Scene cachedScene = null;
    /** generate a scene object, or return a cached copy if this method was
     *  called previously with the cache flag = true
     */
    public Scene getScene()  {
        if (cachedScene == null) {
            cachedScene = generateScene();
        }
        return cachedScene;
    }
    
    public void clearCache() {
        cachedScene = null;
    }
}


class GeneratorThread extends Thread {
    Generator gen;
    PipedOutputStream outPipe = null;
    PipedInputStream inPipe = null;

    public GeneratorThread(Generator gen) {
        this.gen = gen;
        inPipe = new PipedInputStream();
    }

    public InputStream getInputStream() throws Exception {
        // to prevent eternal waiting and other problems, eject if a problem
        // occurred during scene generation
    	if (problem != null) {
	        System.out.println(" GeneratorThread.getInputStream reporting exception: " + problem);
	        throw problem;
	    }
		
	    // reading the input pipe before it is connected to the output pipe
	    // can make bad things happen, so only divulge the input pipe after
	    // outPipe has been created

	    if (outPipe == null) {
	        System.out.println(" output pipe not ready, GeneratorThread.getInputStream returning null");
  	        return null;
	    }
	    return inPipe;
    }

    Exception problem = null;
	
    public void run() {
	    try {
            outPipe = new PipedOutputStream(inPipe);
            PrintWriter pw = new PrintWriter( new OutputStreamWriter( outPipe ) );
            gen.generate( pw );
            pw.flush();
            pw.close();
        } catch (Exception e) {
            System.out.println("Exception generating scene: " + e);
	        problem = e;
        }
    }
}
