/*
 * @(#)TerrainGenerator.java
 *
 * Copyright (c) 1999-2000 by Trapezium Development LLC.  All Rights Reserved.
 *
 * The information in this file is the property of Trapezium Development LLC
 * and may be used only in accordance with the terms of the license granted
 * by Trapezium.
 *
 */

package com.trapezium.generator;

import java.io.*;
import java.util.Vector;
import com.trapezium.vrml.*;
import com.trapezium.parse.TokenEnumerator;
import com.trapezium.browser.vrml.node.*;

public class TerrainGenerator extends Generator {
    
    protected int xDimension;
    protected float xSpacing;
    protected int zDimension;
    protected float zSpacing;
    
    protected transient float height[][];   // zDimension X xDimension
    protected transient float color[][];    // zDimenxion X (xDimension * 3)
    
    protected float midColor[] = {  0.6f, 0.3f, 0.1f };
    protected float highColor[] = { 0.85f, 0.85f, 0.9f };
    protected float lowColor[] = { 0.2f, 0.5f, 0.2f };


    public TerrainGenerator() {
        this(80, 80, 10f, 10f);
    }

    public TerrainGenerator(int hDim, int vDim, float hSpacing, float vSpacing ) {
        xDimension = hDim;
        zDimension = vDim;
        xSpacing = hSpacing;
        zSpacing = vSpacing;
    }
    
    
    protected void generateValues() {
        
        height = new float[zDimension][xDimension];
        
        float xCenter = (xDimension * xSpacing) / 2.0f;
        float zCenter = (zDimension * zSpacing) / 2.0f;
        
        float htMax = Float.NEGATIVE_INFINITY;
        float htMin = Float.POSITIVE_INFINITY;

        for (int iz = 0; iz < zDimension; iz++) {
            float z = (iz * zSpacing) - zCenter;
            for (int ix = 0; ix < xDimension; ix++) {
                //default is dampened sinusoidal wave from center
                float x = (ix * xSpacing) - xCenter;
                float ht = (float) Math.sin( x * (x - z) ) * 0.5f;
                
                if (htMax < ht) {
                    htMax = ht;
                }
                if (htMin > ht) {
                    htMin = ht;
                }

                height[iz][ix] = ht;
            }
        }

        float htAvg = (htMax + htMin) / 2;
        
        // avoid 3 dimensional array because it uses more memory
        color = new float[zDimension][xDimension * 3];
        for (int iz = 0; iz < zDimension; iz++) {
            for (int ix = 0; ix < xDimension; ix++) {
                int i = ix * 3;
                float ht = height[iz][ix];
                float portion, remainder;
                if (ht > htAvg) {
                    portion = (ht - htAvg) / (htMax - htAvg);
                    remainder = 1.0f - portion;
                    color[iz][i] = (portion * highColor[0]) + (remainder * midColor[0]);
                    color[iz][i + 1] = (portion * highColor[1]) + (remainder * midColor[1]);
                    color[iz][i + 2] = (portion * highColor[2]) + (remainder * midColor[2]);
        
                } else if (ht < htAvg) {
                    remainder = (ht - htMin) / (htAvg - htMin);
                    portion = 1.0f - remainder;
                    color[iz][i] = (portion * lowColor[0]) + (remainder * midColor[0]);
                    color[iz][i + 1] = (portion * lowColor[1]) + (remainder * midColor[1]);
                    color[iz][i + 2] = (portion * lowColor[2]) + (remainder * midColor[2]);

                } else {
                    color[iz][i] = midColor[0];
                    color[iz][i + 1] = midColor[1];
                    color[iz][i + 2] = midColor[2];
                }
            }
        }
    }        
            
            
    public void generateElevationGrid( PrintWriter pw ) {
        
        if (height == null || height.length != zDimension || height[0].length != xDimension) {
            generateValues();
        }
        
        pw.println( "geometry ElevationGrid {" );
        pw.println( "  xDimension " + xDimension);
        pw.println( "  xSpacing " + xSpacing );
        pw.println( "  zDimension " + zDimension);
        pw.println( "  zSpacing " + zSpacing );
        pw.println( "  height [" );

        for (int iz = 0; iz < zDimension; iz++) {
            for (int ix = 0; ix < xDimension; ix++) {
                pw.print( " " + height[iz][ix] );

                // every sixteen numbers, except at the end of the row, 
                // print a newline (end of row gets automatic newline below)
                if ( ((ix & 15) == 15) && (ix != xDimension - 1) ) {
                    pw.println("");
                }
            }
            pw.println("");
        }

        pw.println( "  ]" );
        pw.println( "  color Color {" );
        pw.println( "    color [" );
        for (int iz = 0; iz < zDimension; iz++) {
            for (int ix = 0; ix < xDimension; ix++) {
                int i = ix * 3;
                pw.print( " " + color[iz][i] + " " + color[iz][i + 1] + " " + color[iz][i + 2] );
                // every eight triplets, except at the end of the row, 
                // print a newline (end of row gets automatic newline below)
                if ( ((ix & 7) == 7) && (ix != xDimension - 1) ) {
                    pw.println("");
                }
            }
        }
        pw.println( "    ]" );
        pw.println( "  }" );
        pw.println( "}" );
    }
    
    public void generate( PrintWriter pw ) {
        
        float width = xSpacing * xDimension;
        float depth = zSpacing * zDimension;
        
           
        pw.println( "Transform {" );
        pw.println( "  translation " + (-width/2) + " 0 " + (-depth/2) );
        pw.println( "  children [" );
        pw.println( "    Shape {" );
 
        generateElevationGrid(pw);
        
        pw.println( "    }" );
        pw.println( "  ]" );
        pw.println( "}" );
    }
}