/*
 * @(#)InputStreamFactory.java
 *
 * Copyright (c) 1998 by Trapezium Development LLC.  All Rights Reserved.
 *
 * The information in this file is the property of Trapezium Development LLC
 * and may be used only in accordance with the terms of the license granted
 * by Trapezium.
 *
 */
package com.trapezium.parse;

import com.trapezium.generator.Generator;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.FileInputStream;
import java.io.BufferedInputStream;
import java.io.PushbackInputStream;
import java.net.URL;
import java.util.zip.GZIPInputStream;

/**
 *  Provides a static public method for creating
 *  an InputStream from a String indicating a file or URL.
 *
 *  @author          Johannes N. Johannsen
 *  @version         1.1, 16 Dec 1999
 *
 *  @since           1.0
 */
public class InputStreamFactory {
    private InputStreamFactory() {
    }
    
    /**
	 *  Creates an InputStream from a String indicating either a file, a URL,
	 *  a generator class name, or a serialized generator.  
     *
     *  If the file has the prefix "generator:" or "gen:", tries to construct
     *  a generator.  Otherwise tries to construct a URL.  If this fails, 
     *  assumes it is a file.
	 *
	 *  If using JDK 1.0.x, GZIPInputStream not available.  If a GZIP file is
	 *  encountered, we attempt to use this class, exception is caught, null
	 *  returned.
	 *
	 *  @return     InputStream of data associated with url, null if InputStream
	 *              cannot be constructed
	 *  @exception  FileNotFoundException  thrown when url cannot be created
	 */
	static public InputStream getInputStream( String fileName ) throws FileNotFoundException, IOException, NoClassDefFoundError {

		if (fileName.startsWith("gen:") || fileName.startsWith("generator:")) {
		    Generator gen = Generator.getGenerator( fileName );
		    if (gen != null) {
		        return gen.getInputStream();
		    }
		}
		
		// not a generator; see if URL or file
		
		try {
			URL urlAttempt = new URL( fileName );
			PushbackInputStream test = new PushbackInputStream( urlAttempt.openStream() );
			int testchar = test.read();
			test.unread( testchar );
			// detect gzip files by checking first character  0x1f ==> gzip
			if ( testchar == 0x1f ) {
			    try {
			        GZIPInputStream gzi = new GZIPInputStream( test );
			        return( new BufferedInputStream( gzi ));
			    } catch ( Exception e ) {
			        String lm = e.getLocalizedMessage();
			        if (( lm != null ) && ( lm.compareTo( "Not in GZIP format" ) == 0 )) {
			            System.out.println( "File not in GZIP format" );
			        }
			    }
			    return( null );
			// unsupported at the moment, P ==> zip files    
//			} else if ( testchar == 'P' ) {
			} else {
				return( new BufferedInputStream( test ));
			}
		} catch( Exception e ) {
			File source = new File( fileName );
			return( getInputStream( source ));
		}
	}
	
	/** Get an InputStream from a File, handle gzipped files */
	static public InputStream getInputStream( File source ) throws IOException {
		PushbackInputStream test = new PushbackInputStream( new FileInputStream( source ));
		int testchar = test.read();
		test.unread( testchar );
		if ( testchar == 0x1f ) {
			try {
				return( new BufferedInputStream( new GZIPInputStream( test )));
			} catch ( NoClassDefFoundError ex ) {
				System.out.println( "Cannot check gzipped file." );
				throw ex;
			} catch ( Exception ee ) {
				System.out.println( "Cannot check gzipped file." );
				return( null );
			}
//			} else if ( testchar == 'P' ) {
		} else {
			return( new BufferedInputStream( test ));
		}
	}
	
	/**
	 *  Convert a string url into an InputStream
	 *
	 *  @return  InputStream associated with the url, or null if there is any unexpected error
	 *  @exception  FileNotFoundException if the url cannot be found
	 */
	static public InputStream getRawInputStream( String fileName ) throws FileNotFoundException, IOException, NoClassDefFoundError {
		try {
			URL urlAttempt = new URL( fileName );
			PushbackInputStream test = new PushbackInputStream( urlAttempt.openStream() );
			return( new BufferedInputStream( test ));
		} catch( Exception e ) {
		    return( null );
		}
	}
}