/*
 * @(#)SpaceEntity.java
 *
 * Copyright (c) 1998 by Trapezium Development LLC.  All Rights Reserved.
 *
 * The information in this file is the property of Trapezium Development LLC
 * and may be used only in accordance with the terms of the license granted
 * by Trapezium.
 *
 */
package com.trapezium.space;

import com.trapezium.util.CompositeObject;
import com.trapezium.util.TypedObject;

/**
 *  The SpaceEntity represents any primitive aspect of a space structure.
 *  The aspects are enumerated by a set of constants.
 *
 *  At the most basic level, and as part of an optimization to prevent the creation
 *  of too many objects (which I would prefer, but which causes Java to run out of
 *  memory very quickly), a SpaceEntity is a location in space, or a list of index values.
 *
 *  The SpaceEntity below handles the "location-in-space" aspect of an entity.
 *
 *  Still haven't implemented the list of index values aspect.
 *
 *  Even these two cases aren't entirely accurate, texture coordinates are a 2d location
 *  in space (so what is below could handle it OK).
 *
 *  Might make more sense to have the SpaceEntity be the abstract base class, then
 *  subclass the various possible implementations.
 *
 *  Current thinking on this.. all SpaceEntities have a location (which may
 *  not be set in some cases).  Subclasses have the surrounding list of offset
 *  values.
 */
public class SpaceEntity implements TypedObject {
    /** Constant to represent various aspects of a SpaceEntity */
    static public final int NoType = -1;
    static public final int Vertex = 0;
    static public final int Edge = 1;
    static public final int Face = 2;
    static public final int TextureCoordinate = 3;
    static public final int Color = 4;
    static public final int Normal = 5;

    // location
    public float x;
    public float y;
    public float z;

    /** Class constructor */
    public SpaceEntity( float x, float y, float z ) {
        this.x = x;
        this.y = y;
        this.z = z;
    }

    /** Copy constructor */
    public SpaceEntity( SpaceEntity source ) {
        this.x = source.x;
        this.y = source.y;
        this.z = source.z;
    }

    /** Constructor used for locationless entities, faces/edges are
     *  initially constructed this way because most of the time we don't
     *  need their location.
     */
    public SpaceEntity() {
        x = 0f;
        y = 0f;
        z = 0f;
    }

    /** Adjust x to new location and scale */
    public void adjustX( float xToZero, float xFactor, float xShift ) {
        x += xToZero;
        x *= xFactor;
        x += xShift;
    }

    /** Adjust y to new location and scale */
    public void adjustY( float yToZero, float yFactor, float yShift ) {
        y += yToZero;
        y *= yFactor;
        y += yShift;
    }

    /** Adjust z to new location and scale */
    public void adjustZ( float zToZero, float zFactor, float zShift ) {
        z += zToZero;
        z *= zFactor;
        z += zShift;
    }

    /** Set the location */
    public void setLocation( float x, float y, float z ) {
        this.x = x;
        this.y = y;
        this.z = z;
    }

    /** Extract the location into an arry */
    public void extractLocation( float[] location ) {
        location[0] = x;
        location[1] = y;
        location[2] = z;
    }

	public float distanceTo( SpaceEntity other ) {
		float xDis = other.x - x;
		float yDis = other.y - y;
		float zDis = other.z - z;
		xDis = xDis*xDis;
		yDis = yDis*yDis;
		zDis = zDis*zDis;
		float dis = xDis + yDis + zDis;
		try {
			return( (float)Math.sqrt( dis ));
		} catch( Exception e ) {
			return( (float)0 );
		}
	}

    /** TypedObject interface, in practice the most basic types Vertex, Edge
     *  and Face, don't return a type, and keep their type in their parent
     *  SpaceEntitySet.  But for convenience, the types attached to them
     *  do keep a type, in the TypedSpaceEntity object.
     */
    public int getType() {
        return( NoType );
    }

    /** TypedObject interface, looks through attached objects for one
     *  of the specified type.
     */
    public Object getObject( int type ) {
        if ( attachedObject == null ) {
            return( null );
        } else if ( attachedObject.getType() == type ) {
            return( attachedObject );
        } else {
            return( attachedObject.getObject( type ));
        }
    }

    /** Attach an object to the SpacePrimitive.  Several different objects
     *  can be attached using this method.
     */
    TypedObject attachedObject = null;
    public void attachObject( TypedObject a ) {
        if ( attachedObject != null ) {
            attachedObject = new CompositeObject( a, attachedObject );
        } else {
            attachedObject = a;
        }
    }

    /** Get the object attached to the SpacePrimitive */
    public TypedObject getAttachedObject() {
        return( attachedObject );
    }
}
