/** See header below for original info...
    Converted to Java by Joh Johannsen, jojo@com.trapezium.com
    */
//$Id: geodesic.cpp 4.60 1996/09/21 23:00:23 RICK Released RICK $

/*
GEODESIC.CPP - C++ class for the calculation of geodesic dome properties

	Copyright (C) 1995, 1996 Richard J. Bono

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	 Please direct inquiries, comments and modifications to:
	 Richard J. Bono
	 44 Augusta Rd.
	 Brownsville, TX 78521

	 email: rjbono@hiline.net

Revision history:

$Log: geodesic.cpp $
'Revision 4.60  1996/09/21  23:00:23  RICK
'Released Version
'-Added Buckyball VRML support
'-Added wire-frame for DXF & VRML
'-Added face & axial Angle Calcs to DAT
'-Added full sphere Class II support
'-Enhanced POV-ray output

'Revision 4.20  1996/01/27  23:23:39  RICK
'Production Release
'Added elliptical support
'Added enhanced buckyball constructs
'Streamlined Code
'
'Revision 4.0  1995/12/31  18:59:34  RICK
'-Changed data structures to array based linked-lists
'-Many function changes to accomodate linked lists
'-Added Buckball support
'-Added VRML output functions
'-Added face functions
'-Added faces to POV data.
'-DXF data saves face data instead of chord data
'-Source code split into modules
'-Added time passage display while calculating
'-File name is displayed when execution is complete.
'-deleted make_sphere function
'
'Revision 1.1  1995/10/01  16:49:59  RICK
'Initial.
'

Acknowledgements & References:
The main reference used in the creation of this code was "Geodesic Math & How
to Use It" by Hugh Kenner, 1976, University of California Press.
ISBN 0-520-02924-0; Library of Congress Catalog Card Number: 74-27292. Many
thanks to Hugh for putting this data in an accessible format.

Also, many thanks to:
	J. F. Nystrom
	My wife Sonia
	My daughter Kathy
	Chris Fearnley
	Kirby Urner
		&
	R. Buckminster Fuller for changing the way I view Universe.

Compliation:
This program was complied and tested using Borland C++ 4.52 using the large
memory model on a Gateway 2000 4DX2-50V.

The release executables were complied using Borland C++ 3.1. The program will
now run on any IBM PC class machine with or without a math coprocessor.

Support is included for non-Borland compilation. Unix patches provided thanks
to Chris Fearnley & John Kirk.
*/

package com.trapezium.space.dome;

import java.io.FileOutputStream;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.StringWriter;

public class Geodesic implements MathConversions {
	//Various indices, and temporary variables
	int i, j, k;
	int index;
	double sphi, stheta, ephi, etheta, X, Y, Z, E;
	//actual values used in calculations set dependent on class type
	int freq_calc, vertex_calc, face_calc, edges_calc;
	int status_count;

	public int frequency, vertex, faces, edges;

	//structure containing intermediate coordinates
	class Sphere{
		//vertex coordinates: 0,0,0 equals zenith
		int xprime;
		int yprime;
		int zprime;
	}

	Sphere sphere;

	//structure containing vertex labels
	class Label {
		//vertex labels: 0,0 equals zenith
		public int A;
		public int B;
	}

	Label label;

	//structure defines geodesic coordinate points
	class Coordinates{
		//vertex spherical coordinates - This is the main data structure
		double phi;
		double theta;
		double radius;
	}

	Coordinates coordinates;

	//structure defines chord data
	class Chords{
		//Chords are defined by the point array index
		//This structure stores the vertex array index for the chord start
		//and end points.
		int	start;		//chord start point
		int 	end;			//chord end point
	}

	Chords chords;

	//structure defines face data
	class Polygon_face{
		//chords are defined by the point array index
		//This structure stores the vertex array index for the three
		//points required to define a triangular face.
		int cornerA;
		int cornerB;
		int cornerC;
	}

	Polygon_face polygon_face;

	//portability note: huge keyword was needed in order to allocate memory for
	//arrays from the far heap in a MSDOS far memory model.
	Sphere[] pnt_calc;		//pointer to spherical point variable array
	Label[] pnt_label;		//pointer to label array
	Coordinates[] pntcrd;  	//pointer to coordinate array
	Chords[] edgepts;      	//pointer to chord array
	Polygon_face[] polyface;	//pointer to face array
	Coordinates[] sphere_pnt; //pointer to sphere array


//--------------geodesic constructor
public Geodesic( int frequency )
{

	//The constructor handles all calculation when an instance is called.
	//The input parameters are the frequency of subdivision, the class
	//type, the polygon type and a flag to determine whether to generate
	//spherical data. This is all that is needed to specify a geodesic dome!
	//(Note: I have not included radius as size is special case and scaling
	//factors can be applied. The program assumes unit radius domes).

	//initialize class variables
	this.frequency = frequency;
	vertex = 0;
	faces = 0;
	edges = 0;

	//start calculations
	topology();			//calculate symmetry triangle topological abundance
	init_crd();			//initialize vertex coordinates and data structures
	spherical(); 		//calculate spherical coordinates
	chord_factor();	//determine chord factors
	face_factor();		//determine face factors

}

float fmod( float n, int m ) {
	if ( n > 0 ) {
		while ( n >= m ) {
			n -= m;
		}
	} else if ( n < 0 ) {
		int mneg = -1*m;
		while ( n < mneg ) {
			n += m;
		}
	}
	return( n );
}

//--------------calculate topological abundance of symmetry triangle
void topology()
{

	//This class function calculates the topological abundance
	//of the symmetry triangle.

	//calculate the number of vertexia & faces in symmetry triangle
	//This is based on the class type
	//First the polyhedron face...

	for(i=1; i<=frequency; i++){
		if(i == 1)
			vertex = 3;
		else
			vertex += (i + 1);
	}
	faces = frequency * frequency;
	edges = (vertex + faces) - 1;
}
//-------------initialize coordinate array by setting vertex points
void init_crd()
{

	//Set the topology dependent on class type
	vertex_calc = vertex;
	edges_calc = edges;
	freq_calc = frequency;
	face_calc = faces;
}
//-------------calculate spherical coordinates for given polygon type & class
void spherical()
{
	//calculate spherical coordinates of geodesic vertexia
	//given coordinates values
	//care must be taken to avoid generating a trig error

	//Memory is allocated for the full polyhedron face. Create array object for vertexia
	//First allocate memory for label array
	pnt_label = new Label[vertex_calc + 1];
	for (i=0; i<(vertex_calc+1); i++) {
		pnt_label[i] = new Label();
	}

	//now allocate space for the actual vertex coordinate array
	pntcrd = new Coordinates[vertex_calc + 1];
	for (i=0; i<(vertex_calc+1); i++) {
		pntcrd[i] = new Coordinates();
	}

	//create spherical array for one polyhedron face
	sphere_pnt = new Coordinates[vertex_calc + 1];
	for (i=0; i<(vertex_calc+1); i++ ) {
		sphere_pnt[i] = new Coordinates();
	}

	//now allocate space for temporary variables
	pnt_calc = new Sphere[vertex_calc + 1];
	for (i=0; i<(vertex_calc+1); i++) {
		pnt_calc[i] = new Sphere();
	}

	//Initialize vertex labels and coordinates
	index = 1;
	for(i=0; i<=freq_calc; i++){
		for(j=0; j<=i; j++){
			pnt_label[index].A = i;
			pnt_label[index].B = j;
			pnt_calc[index].xprime = j;
			pnt_calc[index].yprime = i - j;
			pnt_calc[index].zprime = frequency - i;
			index++;
		}
	}

	for(i=1; i<=vertex_calc; i++){
		//set-up x,y,z coordinates for spherical calulations
		//See page 75 of Kenner's "Geodesic Math & How to use it"
		//for formulae used below
		//Icosahedron
		X = pnt_calc[i].xprime * Math.sin(72.0 * Geodesic.DEG_TO_RAD);
		Y = pnt_calc[i].yprime + (pnt_calc[i].xprime * Math.cos(72.0 * DEG_TO_RAD));
		Z = frequency/2.0 + (pnt_calc[i].zprime / (2.0 * Math.cos(36.0 * DEG_TO_RAD)));


		//Calculate phi while avoiding trig errors
		if(Y == 0 && X == 0)
			pntcrd[i].phi = 0.0;
		else if(Y == 0)
			pntcrd[i].phi = 90.0;
		else
			pntcrd[i].phi = Math.atan(X / Y) * RAD_TO_DEG;

		//adjust value to correct quadrant
		if(pntcrd[i].phi < 0.0)
			pntcrd[i].phi += 180.0;

		//now calculate theta...this is class dependent...
		if(Z == 0)
			pntcrd[i].theta = 90.0;
		else
			//All class I types use this equation for theta
			pntcrd[i].theta = Math.atan((Math.pow(Math.pow(X, 2.0) + Math.pow(Y, 2.0), 0.5)) / Z) * RAD_TO_DEG;

		//make sure the right quadrant is used
		if(pntcrd[i].theta < 0.0)
			pntcrd[i].theta += 180.0;

		//Set Radius = 1
		pntcrd[i].radius = 1.0;
	}
	pnt_calc = null;
}
//-------------------calculate chord factors
void chord_factor()
{
	//Function cycles through each vertex and determines the
	//chord connections. The use of linked lists have simplified this code
	//considerably.

	index = 1;

	//create array object for edges
	edgepts = new Chords[(edges + 1)];
	for (i = 0; i < (edges+1); i++ ) {
		edgepts[i] = new Chords();
	}

	for(i=1; i<vertex_calc; i++){
		if(pnt_label[i].A == 0 && pnt_label[i].B == 0){
			//point is the zenith vertex; add 10 & 11
			//line #1 start point definition
			edgepts[index].start = i;
			for(j=i; j<=vertex_calc; j++){
				if(pnt_label[j].A == pnt_label[i].A + 1 && pnt_label[j].B == pnt_label[i].B)
					break;
			}
			//end point
			edgepts[index].end = j;

			index++;

			//line #2 start point definition
			edgepts[index].start = i;
			for(j=i; j<=vertex_calc; j++){
				if(pnt_label[j].A == pnt_label[i].A + 1 && pnt_label[j].B == pnt_label[i].B + 1)
					break;
			}
			//end point
			edgepts[index].end =j;

			index++;

		}
		else if(pnt_label[i].A == freq_calc && pnt_label[i].B == 0){
			//point is a left vertex; add 01
			//line #1 start point definition
			edgepts[index].start = i;
			for(j=i; j<=vertex_calc; j++){
				if(pnt_label[j].A == pnt_label[i].A && pnt_label[j].B == pnt_label[i].B + 1)
					break;
			}
			//end point
			edgepts[index].end = j;
			index++;
		}
		else if(pnt_label[i].A == pnt_label[i].B){
			//point is a right edge; add 10 & 11
			//line #1 start point definition
			edgepts[index].start = i;
			for(j=i; j<=vertex_calc; j++){
				if(pnt_label[j].A == pnt_label[i].A + 1 && pnt_label[j].B == pnt_label[i].B)
					break;
			}
			//end point
			edgepts[index].end = j;

			index++;

			//line #2 start point definition
			edgepts[index].start = i;
			for(j=i; j<=vertex_calc; j++){
				if(pnt_label[j].A == pnt_label[i].A + 1 && pnt_label[j].B == pnt_label[i].B + 1)
					break;
			}
			//end point
			edgepts[index].end = j;

			index++;
		}
		else if(pnt_label[i].A == freq_calc){
			//point is a bottom vertex; add 01
			//line #1 start point definition
			edgepts[index].start = i;
			for(j=i; j<=vertex_calc; j++){
				if(pnt_label[j].A == pnt_label[i].A  && pnt_label[j].B == pnt_label[i].B + 1)
					break;
			}
			//end point
			edgepts[index].end = j;

			index++;
		}
		else if(pnt_label[i].B == 0){
			//point is a right edge; add 01, 10 & 11
			//line #1 start point definition
			edgepts[index].start = i;
			for(j=i; j<=vertex_calc; j++){
				if(pnt_label[j].A == pnt_label[i].A && pnt_label[j].B == pnt_label[i].B + 1)
					break;
			}
			//end point
			edgepts[index].end = j;

			index++;

			//line #2 start point definition
			edgepts[index].start = i;
			for(j=i; j<=vertex_calc; j++){
				if(pnt_label[j].A == pnt_label[i].A + 1 && pnt_label[j].B == pnt_label[i].B)
					break;
			}
			//end point
			edgepts[index].end = j;

			index++;

			//line #3 start point definition
			edgepts[index].start = i;
			for(j=i; j<=vertex_calc; j++){
				if(pnt_label[j].A == pnt_label[i].A + 1 && pnt_label[j].B == pnt_label[i].B + 1)
					break;
			}
			//end point
			edgepts[index].end = j;

			index++;
		}
		else {
			//point is an interior vertex; add 01, 10 & 11
			//line #1 start point definition
			edgepts[index].start = i;
			for(j=i; j<=vertex_calc; j++){
				if(pnt_label[j].A == pnt_label[i].A && pnt_label[j].B == pnt_label[i].B + 1)
					break;
			}
			//end point
			edgepts[index].end = j;

			index++;

			//line #2 start point definition
			edgepts[index].start = i;
			for(j=i; j<=vertex_calc; j++){
				if(pnt_label[j].A == pnt_label[i].A + 1 && pnt_label[j].B == pnt_label[i].B)
					break;
			}
			//end point
			edgepts[index].end = j;

			index++;

			//line #3 start point definition
			edgepts[index].start = i;
			for(j=i; j<=vertex_calc; j++){
				if(pnt_label[j].A == pnt_label[i].A + 1 && pnt_label[j].B == pnt_label[i].B + 1)
					break;
			}
			//end point
			edgepts[index].end = j;

			index++;
		}
	}
}
//-------------------calculate face factors
void face_factor()
{
	//Function cycles through each vertex and determines the
	//face definitions essentially going CCW about the face

	index = 1;

	//create array object for faces
	polyface = new Polygon_face[face_calc + 1];
	for (i = 0; i <(face_calc+1); i++) {
		polyface[i] = new Polygon_face();
	}

	for(i=1; i<=(vertex_calc-(freq_calc+1)); i++){
	//function only interates to the end of line frequency - 1. This is located
	//at vertex#:vertex_calc-(freq_calc+1)
		if(pnt_label[i].A == 0 && pnt_label[i].B == 0){
			//point is the zenith vertex; add 10 & 01;
			//Corner A
			polyface[index].cornerA = i;
			for(j=i; j<=vertex_calc; j++){
				if(pnt_label[j].A == pnt_label[i].A + 1 && pnt_label[j].B == pnt_label[i].B)
					break;
			}
			//Corner B
			polyface[index].cornerB = j;
			for(k=i; k<=vertex_calc; k++){
				if(pnt_label[k].A == pnt_label[j].A && pnt_label[k].B == pnt_label[j].B + 1)
					break;
			}
			//Corner C
			polyface[index].cornerC = k;

			index++;

		}
		else if(pnt_label[i].B == 0){
			//point is a left edge; add 10 & 01;
			//                      add 11; subtract 10
			//face #1
			//Corner A
			polyface[index].cornerA = i;
			for(j=i; j<=vertex_calc; j++){
				if(pnt_label[j].A == pnt_label[i].A + 1 && pnt_label[j].B == pnt_label[i].B)
					break;
			}
			//corner B
			polyface[index].cornerB = j;
			for(k=i; k<=vertex_calc; k++){
				if(pnt_label[k].A == pnt_label[j].A && pnt_label[k].B == pnt_label[j].B + 1)
					break;
			}
			//corner C
			polyface[index].cornerC = k;

			index++;

			//Face #2
			//Corner A
			polyface[index].cornerA = i;
			for(j=i; j<=vertex_calc; j++){
				if(pnt_label[j].A == pnt_label[i].A + 1 && pnt_label[j].B == pnt_label[i].B + 1)
					break;
			}
			//Corner B
			polyface[index].cornerB = j;
			for(k=i; k<=vertex_calc; k++){
				if(pnt_label[k].A == pnt_label[j].A - 1 && pnt_label[k].B == pnt_label[j].B)
					break;
			}
			//Corner C
			polyface[index].cornerC = k;

			index++;

		}
		else if(pnt_label[i].A == pnt_label[i].B){
			//point is a right edge; add 10 & 01; subtract 11
			//line #1 start point definition
			//Corner A
			polyface[index].cornerA = i;

			for(j=i; j<=vertex_calc; j++){
				if(pnt_label[j].A == pnt_label[i].A + 1 && pnt_label[j].B == pnt_label[i].B)
					break;
			}
			//Corner B
			polyface[index].cornerB = j;
			for(k=i; k<=vertex_calc; k++){
				if(pnt_label[k].A == pnt_label[j].A && pnt_label[k].B == pnt_label[j].B + 1)
					break;
			}
			//Corner C
			polyface[index].cornerC = k;

			index++;

		}
		else {
			//Interior point; add 10 & 01;
			//                add 11; subtract 10
			//face #1
			//Corner A
			polyface[index].cornerA = i;
			for(j=i; j<=vertex_calc; j++){
				if(pnt_label[j].A == pnt_label[i].A + 1 && pnt_label[j].B == pnt_label[i].B)
					break;
			}
			//Corner B
			polyface[index].cornerB = j;
			for(k=i; k<=vertex_calc; k++){
				if(pnt_label[k].A == pnt_label[j].A && pnt_label[k].B == pnt_label[j].B + 1)
					break;
			}
			//Corner C
			polyface[index].cornerC = k;

			index++;

			//Face #2
			//Corner A
			polyface[index].cornerA = i;
			for(j=i; j<=vertex_calc; j++){
				if(pnt_label[j].A == pnt_label[i].A + 1 && pnt_label[j].B == pnt_label[i].B + 1)
					break;
			}
			//Corner B
			polyface[index].cornerB = j;
			for(k=i; k<=vertex_calc; k++){
				if(pnt_label[k].A == pnt_label[j].A - 1 && pnt_label[k].B == pnt_label[j].B)
					break;
			}
			//Corner C
			polyface[index].cornerC = k;

			index++;
		}

		//faces=index;
	}
}
//--------------------clean floating point numbers
double clean_float(double fp_number)
{
		//function cleans floating point results close to but not quite zero
		if((fp_number < 0.0) & (fp_number > -0.0000000000001))
			fp_number = fabs(fp_number);

		return fp_number;
}

double fabs(double f) {
	if ( f < 0 ) {
		return( -1*f );
	} else {
		return( f );
	}
}
//-------------------time passage signal
void time_passage(int count)
{
	//This function displays the twirly to incidate the program is working
	//and that you machine is not hanging somewhere in limbo...
	if(count == 0)
		System.out.print( '\b' + "-" );
	else if(count == 1)
		System.out.print( '\b' + "/" );
	else if(count == 2)
		System.out.print( '\b' + "|" );
	else if(count == 3)
		System.out.print( '\b' + '\\' );
}
//-------------------determine number of faces per sphere
int face_quantity(int class_value, int poly_value)
{
	int return_value=0;

	//determine the number of faces required (first face = 0)
	if(class_value == 1){
		if(poly_value == 1)
			return_value = 19;
		else if(poly_value == 2)
			return_value =7;
		else if(poly_value == 3)
			return_value = 3;
	}

	return return_value;
}
//-------------------save ASCII PRN format
void save_PRN(String filename)
{
	double sX, sY, sZ, eX, eY, eZ;

	//This function saves the spherical & cartesian data to an
	//comma-delimited ASCII PRN file. This file can be imported into
	//most spreadsheets and the like.

	//Set field width

	System.out.println( "Spherical Vertexia Coordinates" );
	//output PRN data...start with vertexia in spherical coordinates...
	for(i=1; i<=vertex_calc; i++)
			System.out.println( pntcrd[i].phi + ", " + pntcrd[i].theta );

	System.out.println( "" );

	System.out.println( "Cartesian Vertexia Coordinates" );
	//...then in XYZ coordinates
	for(i=1; i<=vertex_calc; i++){
			//convert spherical to cartesian
			sX = clean_float(Math.cos(pntcrd[i].phi * DEG_TO_RAD) * Math.sin(pntcrd[i].theta * DEG_TO_RAD));
			sY = clean_float(Math.sin(pntcrd[i].phi * DEG_TO_RAD) * Math.sin(pntcrd[i].theta * DEG_TO_RAD));
			sZ = clean_float(Math.cos(pntcrd[i].theta * DEG_TO_RAD));
			//Save data
			System.out.println( sX + ", " + sY + ", " + sZ );
	}
	System.out.println( "" );

	System.out.println( "Spherical Chord Coordinates" );
	//Now do the chords...first in spherical coordinate pairs...
	for(i=1; i<=edges_calc; i++)
		System.out.println( pntcrd[edgepts[i].start].phi + ", " + pntcrd[edgepts[i].start].theta + ", "	+ pntcrd[edgepts[i].end].phi + ", " + pntcrd[edgepts[i].end].theta );

	System.out.println( "" );

	System.out.println( "Cartesian Chord Coordinates" );
	//...then in XYZ...
	for(i=1; i<=edges_calc; i++){
		//convert spherical to cartesian
		sX = clean_float(Math.cos(pntcrd[edgepts[i].start].phi * DEG_TO_RAD) *
							  Math.sin(pntcrd[edgepts[i].start].theta * DEG_TO_RAD));
		sY = clean_float(Math.sin(pntcrd[edgepts[i].start].phi * DEG_TO_RAD) *
							  Math.sin(pntcrd[edgepts[i].start].theta * DEG_TO_RAD));
		sZ = clean_float(Math.cos(pntcrd[edgepts[i].start].theta * DEG_TO_RAD));
		eX = clean_float(Math.cos(pntcrd[edgepts[i].end].phi * DEG_TO_RAD) *
							  Math.sin(pntcrd[edgepts[i].end].theta * DEG_TO_RAD));
		eY = clean_float(Math.sin(pntcrd[edgepts[i].end].phi * DEG_TO_RAD) *
							  Math.sin(pntcrd[edgepts[i].end].theta * DEG_TO_RAD));
		eZ = clean_float(Math.cos(pntcrd[edgepts[i].end].theta * DEG_TO_RAD));

		//save data
		System.out.println( sX + ", " + sY + ", " + sZ + ", " + eX + ", " + eY + ", " + eZ );
	}
}
//-------------------make an icosa face for sphere
void icosa_sphere(int facenumber)
{
	double x_phi, x_theta;
	double rot_factor = 0;

	//Given the polyhedron face number calculate Icosa
	//face and store in array. Used in creating spheres
	if(facenumber >=0 && facenumber <=4){
		//do a 72 degree "cap" rotation
		for(i=1; i<=vertex_calc; i++){
			sphere_pnt[i].phi = pntcrd[i].phi + 72.0 * facenumber;
			sphere_pnt[i].theta = pntcrd[i].theta;
			sphere_pnt[i].radius = 1.0;
		}
	}
	else if(facenumber >= 5 && facenumber <=9){
		//do a downward  rotation then rotate appropriate angle
		k = facenumber - 5;		//phi rotation factor
		for(i=1; i<=vertex_calc; i++){
			sphere_pnt[i].theta = rotate_theta(36.0, pntcrd[i].phi, (180.0 - (Math.atan(2) * RAD_TO_DEG)),
					pntcrd[i].theta);
			sphere_pnt[i].phi = (rotate_phi(36.0, pntcrd[i].phi, pntcrd[i].theta,
					sphere_pnt[i].theta) + 36.0) + 72.0 * k;
			sphere_pnt[i].radius = 1.0;
		}
	}
	else if(facenumber >= 10 && facenumber <=14){
		//do a downward rotation then rotate 72 x k phi and 180 theta
		k = facenumber - 10;		//phi rotation factor
		for(i=1; i<=vertex_calc; i++){
			sphere_pnt[i].theta = 180.0 - rotate_theta(36.0, pntcrd[i].phi, (180.0 - (Math.atan(2) * RAD_TO_DEG)),
					pntcrd[i].theta);
			sphere_pnt[i].phi = ((rotate_phi(36.0, pntcrd[i].phi, pntcrd[i].theta,
					sphere_pnt[i].theta) + 36.0) + 72.0 * k) + 36.0;
			sphere_pnt[i].radius = 1.0;
		}
	}
	else{
		//rotate bottom cap into place and the sphere is done!
		k = facenumber - 15;
		for(i=1; i<=vertex_calc; i++){
			sphere_pnt[i].phi = (pntcrd[i].phi + 72.0 * k) + 36.0;
			sphere_pnt[i].theta = 180.0 - pntcrd[i].theta;
			sphere_pnt[i].radius = 1.0;
		}
	}
}
//-------------------make a tetra face for sphere
void tetra_sphere(int facenumber)
{
	double x_phi, x_theta, rot_factor;

	//find A coordinate where angles turn
	int A_change = tetra_angle();

	//Given the polyhedron face number calculate Tetra
	//face and store in array. Used in creating spheres.
	if(facenumber >=0 && facenumber <=2){
		//Build upper three faces
		for(i=1; i<=vertex_calc; i++){
			sphere_pnt[i].phi = pntcrd[i].phi + 120.0 * facenumber;
			sphere_pnt[i].theta = pntcrd[i].theta;
			sphere_pnt[i].radius = 1.0;
		}
	}
	else if(facenumber == 3){
		//Do bottom tetra face
		for(i=1; i<=vertex_calc; i++){
			sphere_pnt[i].theta = rotate_theta(240.0, pntcrd[i].phi, Math.acos(-1.0/3.0) * RAD_TO_DEG,
					pntcrd[i].theta);
			sphere_pnt[i].phi = rotate_phi(240.0, pntcrd[i].phi, pntcrd[i].theta,
					sphere_pnt[i].theta);
				if(pnt_label[i].A >= A_change)
				sphere_pnt[i].phi = 180.0 - sphere_pnt[i].phi;
			if(sphere_pnt[i].phi < 0.0)
				sphere_pnt[i].phi += 360.0;
			sphere_pnt[i].radius = 1.0;
		}
	}
}
//-------------------calculate root E theta correction------------
double root_E(double E, double theta)
{
	//calculate the root E theta correction to even out elliptical edge length
	//distributions.

	double result, X, Y;

	if(theta == 90.0)
		result = 90.0;
	else if(theta == 270.0)
		result = 270.0;
	else{
		result = clean_float(Math.atan(Math.tan(theta * DEG_TO_RAD) / Math.pow(E, 0.5)) * RAD_TO_DEG);
		if(result >= 0.0){
			X = result;
			Y = 180.0 + result;
			if(fabs(X - theta) <= fabs(Y - theta))
				result = X;
			else
				result = Y;
		}
		else{
			X = 180.0 + result;
			Y = 360.0 + result;
			if(fabs(X - theta) <= fabs(Y - theta))
				result = X;
			else
				result = Y;
		}
	}

	return(result);
}
//-------------------calculate radius of ellipse
double ellipse_radius(double E, double theta)
{
	//calculate the radius for a given E and theta

	double result;

	result = Math.pow(Math.pow(E, 2.0) / (Math.pow(E, 2.0) * Math.pow(Math.sin(theta * DEG_TO_RAD), 2.0)
					 + Math.pow(Math.cos(theta * DEG_TO_RAD), 2.0)), 0.5);

	return(result);
}
//-------------------Determine Axial Angle A
double axial_angle_A(int index)
{
	//Calculates the first axial angle for chord "index"
	double sphi, stheta, ephi, etheta, length, result;

	sphi = pntcrd[edgepts[index].start].phi;
	stheta = pntcrd[edgepts[index].start].theta;
	ephi = pntcrd[edgepts[index].end].phi;
	etheta = pntcrd[edgepts[index].end].theta;
	length = chord_length(sphi, stheta, ephi, etheta);

	result = Math.acos((length*length+pntcrd[edgepts[index].start].radius*
				pntcrd[edgepts[index].start].radius-pntcrd[edgepts[index].end].radius*
					pntcrd[edgepts[index].end].radius)/(2*length*
						pntcrd[edgepts[index].start].radius)) * RAD_TO_DEG;

	return(result);
}
//-------------------Determine Axial Angle B
double axial_angle_B(int index, double angleA)
{
		//calculates second axial angle for chord "index" based on angle A

		double result;

		result = Math.asin((pntcrd[edgepts[index].start].radius*
						Math.sin(angleA * DEG_TO_RAD))/
							pntcrd[edgepts[index].end].radius) * RAD_TO_DEG;

		return(result);
}
//------------------Determine Face Angle A
double face_angle_A(int index)
{
	//Calculates first face angle for face "index"
	double sphi, stheta, ephi, etheta, result;
	double l1, l2, l3;

		//get face lengths
		//First A to C
		sphi = pntcrd[polyface[index].cornerA].phi;
		stheta = pntcrd[polyface[index].cornerA].theta;
		ephi = pntcrd[polyface[index].cornerC].phi;
		etheta = pntcrd[polyface[index].cornerC].theta;
		l1 = chord_length(sphi, stheta, ephi, etheta);
		//Next A to B
		sphi = pntcrd[polyface[index].cornerA].phi;
		stheta = pntcrd[polyface[index].cornerA].theta;
		ephi = pntcrd[polyface[index].cornerB].phi;
		etheta = pntcrd[polyface[index].cornerB].theta;
		l2 = chord_length(sphi, stheta, ephi, etheta);
		//Next B to C
		sphi = pntcrd[polyface[index].cornerB].phi;
		stheta = pntcrd[polyface[index].cornerB].theta;
		ephi = pntcrd[polyface[index].cornerC].phi;
		etheta = pntcrd[polyface[index].cornerC].theta;
		l3 = chord_length(sphi, stheta, ephi, etheta);
		result = Math.acos((l1*l1+l2*l2-l3*l3)/(2*l1*l2)) * RAD_TO_DEG;

		return(result);
}
//------------------Determine Face Angle B
double face_angle_B(int index, double faceA)
{
	//Calculates face angle B for face "index" based on face angle A
	double sphi, stheta, ephi, etheta, result;
	double l1, l2, l3;

		//get face lengths
		//First A to C
		sphi = pntcrd[polyface[index].cornerA].phi;
		stheta = pntcrd[polyface[index].cornerA].theta;
		ephi = pntcrd[polyface[index].cornerC].phi;
		etheta = pntcrd[polyface[index].cornerC].theta;
		l1 = chord_length(sphi, stheta, ephi, etheta);
		//...Then B to C
		sphi = pntcrd[polyface[index].cornerB].phi;
		stheta = pntcrd[polyface[index].cornerB].theta;
		ephi = pntcrd[polyface[index].cornerC].phi;
		etheta = pntcrd[polyface[index].cornerC].theta;
		l3 = chord_length(sphi, stheta, ephi, etheta);
		result = Math.asin((l1*Math.sin(faceA * DEG_TO_RAD)) / l3) * RAD_TO_DEG;

		return(result);

}
//------------------Determine Face Angle C
double face_angle_C(double faceA, double faceB)
{
	//Calculate face angle C from face angles A & B
	double result;

	result = 180.0 - faceA - faceB;

	return(result);
}
//-------------------save data to ASCII file
void save_ascii(String filename)
{
	//function to save data to ASCII data file
	//save chord data for symmetry triangle to DAT file
	//filename should include .DAT extension on MS-DOS systems
	//Only data for the symmetry triangle is saved

	double sphi, stheta, ephi, etheta;
	try {
	FileOutputStream ASCIIx = new FileOutputStream(filename);
	PrintStream ASCII = new PrintStream( ASCIIx );
	//output ASCII report data
	ASCII.println( "Geodesic Dome Data" );
	ASCII.println( "----------------Dome Parameters-------------------------" );
	ASCII.print( "Polyhedron Type: ");
	ASCII.println( "Icosahedron" );
	ASCII.print( "Class " );
	ASCII.println( "I" );
	ASCII.println( "Frequency: " + frequency );
	ASCII.println( "----------------Symmetry Triangle Data------------------" );
	ASCII.println( "Vertexia: " + vertex );
	ASCII.println( "Edges: " + edges );
	ASCII.println( "Faces: " + faces );
	ASCII.println( "-----------Symmetry Triangle Spherical Coordinates------" );
	ASCII.println( "Vertex #" + '\t' + "phi" + '\t' + "theta" );

	for(i=1; i<=vertex_calc; i++) {
		ASCII.print( i + '\t' );
		ASCII.print( pntcrd[i].phi + '\t' );
		ASCII.println( pntcrd[i].theta );
	}
	ASCII.println( "" );
	ASCII.println( "----------Symmetry Triangle Chord Data------------------" );
	ASCII.println( "Vertex" + '\n' + "start/end" + '\t' + "Chord Length" );
	for(i=1; i<=edges_calc; i++) {
		ASCII.print( edgepts[i].start + "-" + edgepts[i].end + '\t' );
		sphi = pntcrd[edgepts[i].start].phi;
		stheta = pntcrd[edgepts[i].start].theta;
		ephi = pntcrd[edgepts[i].end].phi;
		etheta = pntcrd[edgepts[i].end].theta;
		ASCII.println( chord_length(sphi, stheta, ephi, etheta));
	}
	ASCII.println( "---------Symmetry Triangle Axial Angle Data-------------" );
	ASCII.println( "Vertex" + '\n' + "start/end" + '\t' + "Axial #1" + '\t' + "Axial #2" );
	for(i=1; i<=edges_calc; i++){
		ASCII.print( edgepts[i].start + "-" + edgepts[i].end + '\t' );
		ASCII.println( axial_angle_A(i) + '\t' + axial_angle_B(i, axial_angle_A(i)) );
	}
	ASCII.println( "---------Symmetry Triangle Face Angle Data-------------" );
	ASCII.println( "Face" + '\t' + "Face Angle #1" + '\t' + "Face Angle #2" + '\t' + "Face Angle #3" );
	for(i=1; i<=face_calc; i++){
		ASCII.print( i + '\t' );
		ASCII.print( face_angle_A(i) + '\t' + face_angle_B(i, face_angle_A(i)) + '\t' );
		ASCII.println( face_angle_C(face_angle_A(i), face_angle_B(i, face_angle_A(i))) );
	}
	ASCII.println( "End Dome Data" );

	ASCII.close();
	} catch( Exception e ) {
		e.printStackTrace();
	}
}
//-------------------calculate chord lengths
double chord_length(double sphi, double stheta, double ephi, double etheta)
{
	//Function returns distance between two points given their
	//spherical coordinates. Radius is fixed at one.

	double result;

	result = Math.pow((2-2 * (Math.cos(stheta * DEG_TO_RAD) * Math.cos(etheta * DEG_TO_RAD) +
		Math.cos((sphi-ephi) *DEG_TO_RAD) * Math.sin(stheta * DEG_TO_RAD) *
			Math.sin(etheta * DEG_TO_RAD))),0.5);

	return result;

}
//------------------phi rotation function
double rotate_phi(double phi, double phi1, double theta1, double theta2)
{
		//phi half of the rotation formula given in Kenner's "Geodesic Math &
		//How to Use it" Kenner credits this formula to Professor George Owen

		double result;

		result = (Math.sin(theta1 * DEG_TO_RAD)* Math.sin((phi-phi1) * DEG_TO_RAD)) /
						Math.sin(theta2 * DEG_TO_RAD);

		//Apply correction for round-off errors
		if(result > 1.0)
			result = 1.0;
		else if(result < -1.0)
			result = -1.0;

		return Math.asin(result) * RAD_TO_DEG;

}
//------------------theta rotation function
double rotate_theta(double phi, double phi1, double theta, double theta1)
{
		//theta half of the rotation formula given in Kenner's "Geodesic Math &
		//How to Use it" Kenner credits this formula to Professor George Owen
		double result;

		result = Math.cos(theta1 * DEG_TO_RAD) * Math.cos(theta * DEG_TO_RAD) +
						Math.sin(theta1 * DEG_TO_RAD) * Math.sin(theta * DEG_TO_RAD) *
							Math.cos((phi-phi1) * DEG_TO_RAD);

		//Apply correction for round-off errors
		if(result > 1.0)
			result = 1.0;
		else if(result < -1.0)
			result = -1.0;

		return Math.acos(result) * RAD_TO_DEG;

}

//-------------Determine "A" Coordinate of bottom tetra face-----------
int tetra_angle()
{

		//This function returns the "A" coordinate of the bottom tetrahedron face
		//where correction of the phi angle must begin. It is the result of quirks
		//in professor owens rotation formulae. I don't find this code very elegant
		//but it does provide the proper results.

		double sX, sphi, stheta;
		double eX, ephi, etheta;

		//Save the 0,0 coordinate as the start point and convert to an X value
		stheta = rotate_theta(240.0, pntcrd[1].phi, Math.acos(-1.0/3.0) * RAD_TO_DEG,
						pntcrd[1].theta);
		sphi = rotate_phi(240.0, pntcrd[1].phi, pntcrd[1].theta, stheta);
		sX = clean_float(Math.cos(sphi * DEG_TO_RAD) * Math.sin(stheta * DEG_TO_RAD));

		//The goal of this routine is to find the "A" coordinate in which the
		//difference between the current X and the last X is positive. We start
		//at 0,0 then proceed down the tetra face edge until this is found
		//(i.e. 0,0 to 1,0; 1,0 to 2,0; 2,0 to 3,0;...f-1,0 to f,0)

		index = 1;
		for(i=2; i<=freq_calc; i++){
			index += (i-1);	//index given the array location of the tetra edge point

			//Save the next coordinate as the end point and convert to an X value
			etheta = rotate_theta(240.0, pntcrd[index].phi, Math.acos(-1.0/3.0) * RAD_TO_DEG,
							pntcrd[index].theta);
			ephi = rotate_phi(240.0, pntcrd[index].phi, pntcrd[index].theta, etheta);

			eX = clean_float(Math.cos(ephi * DEG_TO_RAD) * Math.sin(etheta * DEG_TO_RAD));

			if((eX - sX) > 0.0)
				return pnt_label[index].A;
			else
				sX = eX;
		}

		//Handle special cases of frequency < 5
		if(freq_calc == 4)
			return 3;
		if(freq_calc == 1)
			return 1;
		else
			return 2;
}
//--------------------display data function
void display_data()
{
		//Set the topology dependent on class type
		vertex_calc = vertex;
		edges_calc = edges;
		System.out.println( "Geodesic Dome Data" );
		System.out.println( "----------------Dome Parameters-------------------------" );
		System.out.print( "Polyhedron Type: " );
		System.out.println( "Icosahedron" );
		System.out.print( "Class " );
		System.out.println( "I" );
		System.out.println( "Frequency: " + frequency );

			System.out.println( "----------------Symmetry Triangle Data------------------" );

		System.out.println( "Vertexia: " + vertex_calc );
		System.out.println( "Edges: " + edges_calc );
		System.out.println( "Faces: " + face_calc );
		System.out.println( "" );

}

//----------------------------------------WRL File memeber Functions
//-------------------save WRL file as indexed face set

void save_WRL(String filename)
{
	//save chord data for symmetry triangle to WRL file
	//filename should include .WRL extension on MS-DOS systems

	try {
    	PrintStream WRL = new PrintStream( new FileOutputStream(filename));

    	StringWriter sw = new StringWriter();
    	PrintWriter pw  = new PrintWriter( sw );
    	pw.println( "#VRML V2.0 utf8" );
    	pw.println( "" );
    	pw.println( "#WRL File Generated by DOME 4.60: " );
    	pw.println( "#modified by Trapezium for VRML 2.0" );
    	pw.println( "" );
    	pw.println( "Transform { children [ " );
    	saveIFS( pw );
    	pw.println( "] }" );
    	pw.flush();
    	WRL.println( sw.toString() );
	} catch( Exception e ) {
		e.printStackTrace();
	}

}

public void saveIFS( PrintWriter WRL ) {
    try {
        int end_face;
    	double dX, dY, dZ;
    	int sX, sY, sZ;

        //determine the number of faces required (first face = 0)
   		end_face = 19;

    	//set up VRML file header
    	WRL.println( "Shape {" );
    	WRL.println( "  appearance Appearance { material Material {" );
    	WRL.println( "    diffuseColor 0 0 1 }}" );
    	WRL.println( "  geometry IndexedFaceSet {" );
    	WRL.println( "    coord Coordinate {" );
    	WRL.println( "      point [" );

    	//output VRML data...start with vertexia points...
    	for(j=0; j<=end_face; j++){
   			icosa_sphere(j);

    		WRL.println( "       #Face #" + (j+1) );
    		for(i=1; i<=vertex_calc; i++){

	    		//convert spherical to cartesian
    			dX = sphere_pnt[i].radius *
    				  clean_float(Math.cos(sphere_pnt[i].phi * DEG_TO_RAD) *
    								  Math.sin(sphere_pnt[i].theta * DEG_TO_RAD));
    			dY = sphere_pnt[i].radius *
    				  clean_float(Math.sin(sphere_pnt[i].phi * DEG_TO_RAD) *
    								  Math.sin(sphere_pnt[i].theta * DEG_TO_RAD));
    			dZ = sphere_pnt[i].radius *
    				  clean_float(Math.cos(sphere_pnt[i].theta * DEG_TO_RAD));

    			//Save data
    			WRL.println( dX + " " + dY + " " + dZ + "," );
    		}
    	}
    	WRL.println( "            ]" );
    	WRL.println( "          }" );

    	//Now define faces...
	    WRL.println( "       coordIndex [" );

    	for(j=0; j<=end_face; j++){
    		//output VRML face data...
    		WRL.println( "       #Face #" + (j+1) );
    		for(i=1; i<=face_calc; i++){
    			//Data saved is the point coordinate index for the given face
    			sX = (polyface[i].cornerA + (vertex_calc * j)) - 1;
    			sY = (polyface[i].cornerB + (vertex_calc * j)) - 1;
    			sZ = (polyface[i].cornerC + (vertex_calc * j)) - 1;
    			if ( j<= end_face/2 )
    		        WRL.println( sX + ", " + sY + ", " + sZ + ", -1," );
    		    else
    		        WRL.println( sZ + ", " + sY + ", " + sX + ", -1," );
    		}
    	}
    	WRL.println( "               ] #end coordIndex" );
    	WRL.println( "          } #end IndexedFaceSet" );
    	WRL.println( "      } #end Shape" );
    } catch( Exception e ) {
		e.printStackTrace();
	}
}

//-------------------save WRL file as indexed line set
void save_WRL_wire(String filename)
{

	//save chord data for symmetry triangle to WRL file
	//filename should include .WRL extension on MS-DOS systems
	try {
	PrintStream WRL = new PrintStream( new FileOutputStream(filename));
	//Set field widths



    	StringWriter sw = new StringWriter();
    	PrintWriter pw  = new PrintWriter( sw );
    	pw.println( "#VRML V2.0 utf8" );
    	pw.println( "" );
    	pw.println( "#WRL File Generated by DOME 4.60: " );
    	pw.println( "#modified by Trapezium for VRML 2.0" );
    	pw.println( "" );
    	pw.println( "Transform { children [ " );
    	saveILS( pw );
    	pw.println( "] }" );
    	pw.flush();
    	WRL.println( sw.toString() );
    } catch ( Exception e ) {
    }
}

void saveILS( PrintWriter WRL )
{
	int start_pnt, end_pnt;
	double dX, dY, dZ;
	int end_face = 19;
    try {
	WRL.println( "Shape {" );
	WRL.println( "  appearance Appearance { material Material {" );
	WRL.println( "    emissiveColor .9 0 .8 }}" );
	WRL.println( "  geometry IndexedLineSet {" );
	WRL.println( "    coord Coordinate {" );
	WRL.println( "      point [" );

	//output VRML data...start with vertexia points...
	for(j=0; j<=end_face; j++){
		icosa_sphere(j);

		WRL.println( "       #Face #" + (j+1) );
		for(i=1; i<=vertex_calc; i++){

			//convert spherical to cartesian
			dX = sphere_pnt[i].radius *
				  clean_float(Math.cos(sphere_pnt[i].phi * DEG_TO_RAD) *
								  Math.sin(sphere_pnt[i].theta * DEG_TO_RAD));
			dY = sphere_pnt[i].radius *
				  clean_float(Math.sin(sphere_pnt[i].phi * DEG_TO_RAD) *
								  Math.sin(sphere_pnt[i].theta * DEG_TO_RAD));
			dZ = sphere_pnt[i].radius *
				  clean_float(Math.cos(sphere_pnt[i].theta * DEG_TO_RAD));

			//Save data
			WRL.println( dX + " " + dY + " " + dZ + "," );
		}
	}
	WRL.println( "            ]" );
	WRL.println( "          }" );

	//Now define chords...
	WRL.println( "       coordIndex [" );

	for(j=0; j<=end_face; j++){
		//output VRML line data...
		WRL.println( "       #Face #" + (j+1) );
		for(i=1; i<=edges_calc; i++){
			//Data saved is the point coordinate index for the given chord
			start_pnt = (edgepts[i].start + (vertex_calc * j)) - 1;
			end_pnt = (edgepts[i].end + (vertex_calc * j)) - 1;
		WRL.println( start_pnt + ", " + end_pnt + ", -1," );
		}
	}
	WRL.println( "               ] #end coordIndex" );
	WRL.println( "          } #end IndexedLineSet" );
	WRL.println( "      } #end Shape" );
	WRL.close();
	} catch( Exception e ) {
		e.printStackTrace();
	}
}
}
//--------------------End of class Geodesic


