
//package ThreeDimensionalJava;
package com.trapezium.test;

// Import the Java 3D classes and the vector mathmetics classes.
import javax.media.j3d.*;
import javax.vecmath.*;

/**
 * The Rotator class is used to set up a bevavior that continously
 * rotates about a given axis.  The Rotator class acts just like
 * a TransformGroup - add a Rotator to the scene graph and then any
 * children added to the Rotator will be rotated.
 *
 * @(#)Rotator.java 1.3 98/02/23 17:09:19
 * @version 1.0
 */
public class Rotator extends TransformGroup {

    /**
     * Creates and initializes the objects used to create
     * the rotation bevior.  This includes creating the Alpha,
     * the RotationInterpolator and the bounds object.
     *
     * @param axis The axis to rotate about.  The default
     *  RotationInterpolator rotates about the Y-axis.
     * @param speed The time for one complete roatation to
     *  occur (in milliseconds).
     */
    protected void initRotator(Transform3D axis ) {

        // Must set this to allow the TransformGroup's Transform
        // to be updated when the scene graph is made live.
        // If this value is not set, a CapabilityNotSetException
        // will be thrown once the node is made live.
        setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);

        // Create an alpha to use with the RotationInterpolator.
        // The following values are used for each parameter:
        //   loopCount
        //     Specifies the number of times to cycle the alpha.
        //     A -1 means to cycle it infinitely.
        //   mode
        //     Specifies what part of the wave form (going from
        //     0.0 to 1.0 or 1.0 to 0.0 or both) to use.  The Rotator
        //     alpha will go from 0.0 to 1.0 and then start over at 0.0.
        //   triggerTime
        //     Defines when (in milliseconds) to begin the
        //     alpha waveform.  For the Rotator class no delay is
        //     necessary.
        //   phaseDelayDuration
        //     Defines a time period when the alpha value does not
        //     change.  For this time alpha will be either 0.0 or 1.0
        //     depending on the waveform of the alpha created.
        //   increasingAlphaDuration
        //     The time (in milliseconds) to vary alpha from 0.0
        //     to its 1.0 value.  This value is changed in the Rotator
        //     class to change how fast an object is rotating.
        //   increasingAlphaRampDuration
        //     Used to change the rate at which alpha increases over
        //     time.  For the Rotator class, a constant rate is desired
        //     so this value is set to 0.
        //   alphaAtOneDuration
        //     The time (in milliseconds) to keep alpha at its 1.0 value.
        //     The Rotator class wants a continous rotation so this value
        //     is set to 0 so no delay occurs.
        //   decreasingAlphaDuration
        //     The time (in milliseconds) to vary alpha from 1.0 to 0.0.
        //     The Rotator class does not use decreasing alpha so this
        //     value is ignored.
        //   decreasingAlphaRampDuration
        //     Used to change the rate at which alpha decreases over
        //     time.  The Rotator class does not use decreasing alpha so
        //     this value is ignored.
        //   alphaAtZeroDuration
        //     The time (in milliseconds) to keep alpha at its 0.0 value.
        //     The Rotator class wants a continous rotation so this value
        //     is set to 0 so no delay occurs.
        Alpha rotationAlpha = new Alpha(-1, Alpha.INCREASING_ENABLE,
          0, 0, 8000, 0, 0, 0, 0, 0);

        // Create the RotationInterpolator.  The following values
        // are used for each parameter:
        //   alpha
        //     The alpha value to use for this interpolator.
        //   target
        //     The TransformGroup to modify by this interpolator.
        //     Because Rotator is a subclass of TransformGroup,
        //     the Rotator class will be modifing its own
        //     TransformGroup.
        //   axisOfRotation
        //     The RotationInterpolator always does its rotation
        //     about the Y-axis.  This transform is used to
        //     orient this local coordinate system in the desired
        //     manner before performing the rotation.
        //   minimumAngle
        //     The starting angle of the interpolator.  The Rotator
        //     class will be generating a complete 360-degree rotation
        //     so the start value is 0.0 (in radians).
        //   maximumAngle
        //     The final angle of the interpolator.  Since the Rotator
        //     class wants one rotation to happen for every alpha cycle,
        //     The final angle is 2*PI.
        RotationInterpolator rotator = new RotationInterpolator(
          rotationAlpha, this, axis, 0.0f, (float)Math.PI*2.0f);

        // Allocate the bounding sphere to be used with this Rotator.
        // The assumption is that all Rotators should be active all
        // the time.  The bounds center is set to the origin and
        // then a radius of 100 is assumed to be "sufficiently large" so
        // that this interpolator will always be active.  For large
        // scene graphs this bounds should be adjusted so that the
        // Interpolator is only active when the viewer is close enough
        // to see the object being rotated.
        BoundingSphere bounds =
          new BoundingSphere(new Point3d(0.0, 0.0, 0.0), 100.0);
        rotator.setSchedulingBounds(bounds);

        // Add the Rotator Interpolator to this TransformGroup.  Now
        // all objects added to the TransformGroup will be rotated.
        addChild(rotator);
    }

    /**
     * Creates a Rotator object rotating about the given axis, rotating
     * at the specified speed.
     *
     * @param axis The transform used to orient the RotationInterpolator's
     *  Y-axis (the axis the rotation is performed on).
     * @param speed The time (in milliseconds) for one complete rotation
     *  to occur.  Three predefined values are provided:SLOW, MEDIUM and
     *  FAST.
     */
    public Rotator(Transform3D axis) {
        // Call the initRotator routine with the given parameters.
        initRotator(axis);
    }

    /**
     * Creates an uninitialized Rotator object.  Only routines that
     * subclass the Rotator need this method.
     */
    Rotator() {
        // initRotator must be called manually
    }
}
