/*
 * @(#)Access.java
 *
 * Copyright (c) 1999 by Trapezium Development LLC.  All Rights Reserved.
 *
 * The information in this file is the property of Trapezium Development LLC
 * and may be used only in accordance with the terms of the license granted
 * by Trapezium.
 *
 */
package com.trapezium.util;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.URL;
import java.util.*;

/** Generalized access manager for applications.
 *
 *  There are four kinds of access:
 *
 *       ALLOWED      Registered copy, full access allowed
 *       EVALUTAION   Evaluation copy, temporary access allowed
 *       REFUSED      Evaluation copy, evaluation has expired
 *       BETA         Beta copy, full or temporary access allowed
 *
 *
 *
 *
 *  @author          Michael St. Hippolyte
 *  @version         1.0, 24 Feb 1999
 *
 *  @since           1.0
 */


public class Access {

    public static final int ACCESS_BETA = -2;
    public static final int ACCESS_ALLOWED = -1;
    public static final int ACCESS_REFUSED = 0;
    public static final int ACCESS_EVALUATION = 1;

    private String appname;

    private int access;
    private int days;
    private String username;
    private String email;
    private String key;

    private String checkFilename;   // not the real key file, but must be present
    private String keyFilename;     // the real key file

    /** create a beta access object */
    public Access(String appname, int days) {
        this.appname = appname;
        this.days = days;
        access = ACCESS_BETA;
        username = "";
        email = "";
        key = "";
    }
 
    
    public Access(String appname, String keyDirectory) {

        this.appname = appname;

        if (keyDirectory == null) {
            keyDirectory = "";
        }
        checkFilename = keyDirectory + appname + ".key";
        keyFilename = keyDirectory + ProgramYek.getVidSys();
        ProgramYek yek = getYek(keyFilename);
        if ( yek != null && yek.isValid() ) {
            initFields(yek);
        } else {
            days = 0;
            access = ACCESS_REFUSED;
            username = "";
            email = "";
            key = "";
        }
    }

    public int daysLeft() {
        return days;
    }

    public int getAccess() {
        return access;
    }

    public String getUserName() {
        return username;
    }

    public String getUserEmail() {
        return email;
    }

    public String getKey() {
        return key;
    }

    public boolean accessAllowed() {
        return (access != ACCESS_REFUSED);
    }

    public boolean fullAccessAllowed() {
        return (access == ACCESS_ALLOWED || access == ACCESS_BETA);
    }

    public boolean evaluationAccess() {
        return (access == ACCESS_EVALUATION);
    }

    public URL getResource(String name) {
        return getClass().getResource("/" + name);
    }

    public boolean setFields(String name, String email, String key) {

        // validate fields
        if (email == null || email.length() < 5) {
             return false;
         }
        int n = email.indexOf('@');
        if (n == -1 || email.indexOf('.', n) == -1) {
            return false;
        }
        if (name == null || name.length() < 1) {
            return false;
        }
        if (key == null || key.length() != ProgramYek.KEYLEN) {
            return false;
        }
        // define a new ProgramYek object
        ProgramYek newYek = new ProgramYek(appname, name, email);
        if (newYek.isValid() && newYek.isKey(key)) {

            initFields(newYek);

            // save to disk
            File f = new File( keyFilename );
            try {
                FileOutputStream fos = new FileOutputStream( f );
                newYek.write(fos);
                fos.close();
            } catch ( Exception e ) {
                System.out.println( "** Exception writing key: " + e );
                return false;
            }
            return true;
        } else {
            return false;
        }
    }

    private void initFields(ProgramYek yek) {
        days = yek.daysLeft();
        if (days > 0) {
            access = ACCESS_EVALUATION;
        } else if (days == 0) {
            access = ACCESS_REFUSED;
        } else {
            access = ACCESS_ALLOWED;
        }
        username = yek.getUserName();
        email = yek.getUserEmail();
        key = yek.toString();
    }

    private ProgramYek getYek(String keyFilename) {
        File f = new File(keyFilename);
        InputStream is = null;
        try {
            is = new FileInputStream( f );
        } catch ( java.io.FileNotFoundException ee ) {
            System.out.println( "Key file " + keyFilename + " not found" );
        } catch ( Exception e ) {
            System.out.println( "** Exception reading key: " + e );
        }
        if (is == null) {
            try {
                URL url = getResource(ProgramYek.getVidSys());
                if ( url != null ) {
                    is = url.openStream();
                }
            } catch ( Exception e ) {
                System.out.println( "** Exception reading key: " + e );
            }
        }

        ProgramYek y = null;
        if (is != null) {
            try {
                y = ProgramYek.read(is);
            } catch ( Exception e ) {
                System.out.println( "** Exception reading key: " + e );
            }
        } else {
            // if we haven't yet found a key, and the check file is present,
            //we can generate one
            File chk = new File(checkFilename);
            if (chk.exists()) {
                System.out.println("This is the evaluation version of " + appname);
                y = new ProgramYek(appname, "Evaluation", "", 15);
                ProgramYek.serialize(y, keyFilename);
            }
        }
        return y;
    }
}

