/*
 * BarLayout.java
 */

package com.trapezium.util;

import java.awt.*;
import javax.swing.AbstractButton;

/**
 * FlowLayout with a little BorderLayout thrown in, for vertical or horizontal
 * control bars.
 */
public class BarLayout implements LayoutManager {

    // alignment
    public static final int LEFT = 0;
    public static final int CENTER = 1;
    public static final int RIGHT = 2;
    public static final int TOP = 3;
    public static final int VCENTER = 4;
    public static final int BOTTOM = 5;

    // flag added to alignment indicating all buttons should have the same
    // vertical size (for horizontal bars) or horizontal size (for vertical bars)
    public static final int EQUALIZE_BUTTONS = 16;

    // controls at ends
    public static final String LEFT_END = "Left";
    public static final String TOP_END = "Top";
    public static final String RIGHT_END = "Right";
    public static final String BOTTOM_END = "Bottom";

    Component end1;     // top or left
    Component end2;     // bottom or right
    
    int align;
    int hgap;
    int vgap;
    
    public BarLayout() {
    	this(CENTER, 2, 2);
    }
    public BarLayout(int align) {
    	this(align, 2, 2);
    }
    public BarLayout(int align, int hgap, int vgap) {
    	this.hgap = hgap;
	    this.vgap = vgap;
        setAlignment(align);
    }

    public int getAlignment() {
    	return align;
    }

    public void setAlignment(int align) {
        if ((align & (~EQUALIZE_BUTTONS)) < LEFT || (align & (~EQUALIZE_BUTTONS)) > BOTTOM) {
            throw new IllegalArgumentException("Invalid align value for BarLayout");
        }
    	this.align = align;
    }

    public final boolean isHorizontal() {
        int a = align & (~EQUALIZE_BUTTONS);
        return (a == LEFT || a == CENTER || a == RIGHT);
    }


    public int getHgap() {
    	return hgap;
    }
    public void setHgap(int hgap) {
    	this.hgap = hgap;
    }

    public int getVgap() {
    	return vgap;
    }

    public void setVgap(int vgap) {
    	this.vgap = vgap;
    }
    
    

    public void addLayoutComponent(String name, Component comp) {
        if (LEFT_END.equals(name) || TOP_END.equals(name)) {
            end1 = comp;
        } else if (RIGHT_END.equals(name) || BOTTOM_END.equals(name)) {
            end2 = comp;
        }
    }

    public void removeLayoutComponent(Component comp) {
        if (end1 == comp) {
            end1 = null;
        }
        if (end2 == comp) {
            end2 = null;
        }
    }

    public Dimension preferredLayoutSize(Container target) {
        synchronized (target.getTreeLock()) {
	        Dimension dim = new Dimension(0, 0);
	        int nmembers = target.getComponentCount();

            if (isHorizontal()) { 
    	        for (int i = 0 ; i < nmembers ; i++) {
    	            Component m = target.getComponent(i);
    	            if (m.isVisible()) {
    		            Dimension d = m.getPreferredSize();
    		            dim.height = Math.max(dim.height, d.height);
   		                dim.width += hgap;
    		            dim.width += d.width;
    	            }
    	        }
    	        
    	        // The following adjusts the hgaps according to whether end 
    	        // components are present.  It works out to adding an hgap 
    	        // if neither end is present, subtracting an hgap if both
    	        // ends are present, and doing nothing if exactly one end
    	        // component is present.
        	    if (end1 == null) {
        	        dim.width += hgap;
        	    }
    	        if (end2 != null) {
        	        dim.width -= hgap;
        	    }
                dim.height += 2 * vgap;

    	    } else {
    	        for (int i = 0 ; i < nmembers ; i++) {
    	            Component m = target.getComponent(i);
    	            
    	            // don't filter out invisible members, or they may not fit
    	            // in the control when/if they become visible
    	            if (m.isVisible()) {
    		            Dimension d = m.getPreferredSize();
    		            dim.width = Math.max(dim.width, d.width);
   		                dim.height += vgap;
    		            dim.height += d.height;
    	            }
    	        }
    	        // See explanation above on adjusting hgaps
        	    if (end1 == null) {
        	        dim.height += vgap;
        	    }
    	        if (end2 != null) {
        	        dim.height -= vgap;
        	    }
                dim.width += 2 * hgap;
    	    }
        	Insets insets = target.getInsets();
        	dim.width += insets.left + insets.right;
        	dim.height += insets.top + insets.bottom;
        	return dim;
        }
    }

    public Dimension minimumLayoutSize(Container target) {
        return preferredLayoutSize(target);
    }

    public void layoutContainer(Container target) {
        synchronized (target.getTreeLock()) {
    	    int nmembers = target.getComponentCount();
    	    Insets insets = target.getInsets();
    	    int width = target.getWidth();
    	    int height = target.getHeight();
    	    int x, y;
    	    
    	    int end1width, end1height;
    	    Dimension pref;
    	    if (end1 != null) {
    	        pref = end1.getPreferredSize();
    	        end1width = pref.width;
    	        end1height = pref.height;
            } else {
                end1width = end1height = 0;
    	    }
    	    
    	    int end2width, end2height;
    	    if (end2 != null) {
    	        pref = end2.getPreferredSize();
    	        end2width = pref.width;
    	        end2height = pref.height;
            } else {
                end2width = end2height = 0;
    	    }

    	    pref = preferredLayoutSize(target);
    	    
            int a = align & (~EQUALIZE_BUTTONS);
            if (isHorizontal()) {
                if (a == LEFT) {
                    x = insets.left + end1width + hgap;
                } else if (a == CENTER) {
                    x = (width - pref.width) / 2;
                } else {                
                    x = width - insets.right - hgap - pref.width + end1width;
                }
                int eqheight = 0;
                if ((align & EQUALIZE_BUTTONS) != 0) {
                    eqheight = pref.height - insets.top - insets.bottom;
                    if (end1 != null && end1 instanceof AbstractButton) {
                        end1height = eqheight;
                    }
                    if (end2 != null && end1 instanceof AbstractButton) {
                        end2height = eqheight;
                    }
                }                    
                
                if (end1 != null) {
              		y = (height - end1height) / 2;
                    end1.setBounds(insets.left, y, end1width, end1height);
                }
                if (end2 != null) {
              		y = (height - end2height) / 2;
                    end2.setBounds(width - insets.right - end2width, y, end2width, end2height);
                }
                for (int i = 0; i < nmembers; i++) {
            	    Component m = target.getComponent(i);
            	    if (m.isVisible() && m != end1 && m != end2) {
                		Dimension d = m.getPreferredSize();
                		int mheight = d.height;
                		if (eqheight > 0 && m instanceof AbstractButton) {
                		    mheight = eqheight;
                		}
                		y = (height - mheight) / 2;
                		m.setBounds(x, y, d.width, mheight);
                		x += d.width + hgap;
                    }
                }

            } else {
                if (a == TOP) {
                    y = insets.top + end1height + vgap;
                } else if (a == VCENTER) {
                    y = (height - pref.height) / 2;
                } else {                
                    y = height - insets.bottom - vgap - pref.height + end1height;
                }

                int eqwidth = 0;
                if ((align & EQUALIZE_BUTTONS) != 0) {
                    eqwidth = pref.width - insets.left - insets.right;
                    if (end1 != null && end1 instanceof AbstractButton) {
                        end1width = eqwidth;
                    }
                    if (end2 != null && end2 instanceof AbstractButton) {
                        end2width = eqwidth;
                    }
                }                    
                
                if (end1 != null) {
              		x = (width - end1width) / 2;
                    end1.setBounds(x, insets.top, end1width, end1height);
                }
                if (end2 != null) {
              		x = (width - end2width) / 2;
                    end2.setBounds(x, height - insets.bottom - end2height, end2width, end2height);
                }
                for (int i = 0; i < nmembers; i++) {
            	    Component m = target.getComponent(i);
            	    if (m.isVisible() && m != end1 && m != end2) {
                		Dimension d = m.getPreferredSize();
                		int mwidth = d.width;
                		if (eqwidth > 0 && m instanceof AbstractButton) {
                		    mwidth = eqwidth;
                		}
                		x = (width - mwidth) / 2;
                		m.setBounds(x, y, mwidth, d.height);
                		y += d.height + vgap;
                    }
                }
            }    	    
        }
    }

    public String toString() {
    	String str = "";
    	switch (align & (~EQUALIZE_BUTTONS)) {
    	    case LEFT:      str = ",align=left"; break;
    	    case CENTER:    str = ",align=center"; break;
    	    case RIGHT:     str = ",align=right"; break;
    	    case TOP:       str = ",align=top"; break;
    	    case VCENTER:   str = ",align=vcenter"; break;
    	    case BOTTOM:    str = ",align=bottom"; break;
    	}
    	if ((align & EQUALIZE_BUTTONS) != 0) {
    	    str = str + ",equalize";
    	}
    	return getClass().getName() + "[hgap=" + hgap + ",vgap=" + vgap + str + "]";
    }
}