/*
 * BasicWheelControlUI.java
 */

package com.trapezium.util;

import java.awt.Component;
import java.awt.Container;
import java.awt.Adjustable;
import java.awt.event.*;
import java.awt.Graphics;
import java.awt.Dimension;
import java.awt.Rectangle;
import java.awt.Point;
import java.awt.Insets;
import java.awt.Color;
import java.awt.IllegalComponentStateException;
import java.awt.Polygon;
import java.beans.*;
import java.util.*;

import javax.swing.border.AbstractBorder;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.plaf.*;
import javax.swing.plaf.basic.BasicGraphicsUtils;


public class BasicWheelControlUI extends SliderUI {

    public final int POSITIVE_SCROLL = +1;
    public final int NEGATIVE_SCROLL = -1;
    public final int MIN_SCROLL = -2;
    public final int MAX_SCROLL = +2;

    protected JWheelControl wheel;

    protected Insets focusInsets = null;
    protected Insets insetCache = null;
    protected Rectangle focusRect = null;
    protected Rectangle contentRect = null;
    protected Rectangle wheelRect = null;

    private static final Dimension PREFERRED_HORIZONTAL_SIZE = new Dimension(141, 15);
    private static final Dimension PREFERRED_VERTICAL_SIZE = new Dimension(15, 141);
    private static final Dimension MINIMUM_HORIZONTAL_SIZE = new Dimension(79, 15);
    private static final Dimension MINIMUM_VERTICAL_SIZE = new Dimension(15, 79);

    private static final Insets HORIZONTAL_PADDING = new Insets(0, 2, 2, 2);
    private static final Insets VERTICAL_PADDING = new Insets(2, 0, 2, 2);

    private transient boolean isDragging;

    protected TrackListener trackListener;
    protected ChangeListener changeListener;
    protected ComponentListener componentListener;
    protected FocusListener focusListener;
    protected PropertyChangeListener propertyChangeListener;

    // Colors
    private Color shadowColor;
    private Color highlightColor;
    private Color focusColor;


    protected Color getShadowColor() {
        return shadowColor;
    }

    protected Color getHighlightColor() {
        return highlightColor;
    }

    protected Color getFocusColor() {
        return focusColor;
    }

    /////////////////////////////////////////////////////////////////////////////
    // ComponentUI Interface Implementation methods
    /////////////////////////////////////////////////////////////////////////////
    public static ComponentUI createUI(JComponent w)    {
        return new BasicWheelControlUI((JWheelControl) w);
    }

    public BasicWheelControlUI(JWheelControl w) {
    }

    public void installUI(JComponent c)   {
        wheel = (JWheelControl) c;

        wheel.setEnabled(wheel.isEnabled());
        wheel.setOpaque(false);
        wheel.setBorder(BorderFactory.createLoweredBevelBorder());

        isDragging = false;
        trackListener = createTrackListener( wheel );
        changeListener = createChangeListener( wheel );
        componentListener = createComponentListener( wheel );
        focusListener = createFocusListener( wheel );
    	propertyChangeListener = createPropertyChangeListener( wheel );

	    installDefaults( wheel );
    	installListeners( wheel );
	    installKeyboardActions( wheel );

    	insetCache = wheel.getInsets();
	    focusRect = new Rectangle();
    	contentRect = new Rectangle();
	    wheelRect = new Rectangle();

	    calculateGeometry(); // This figures out where the labels, ticks, track, and thumb are.
    }

    public void uninstallUI(JComponent c) {
        if ( c != wheel ) {
            throw new IllegalArgumentException("Unable to uninstallUI for components other than wheel controls");
        }

        LookAndFeel.uninstallBorder(wheel);


    	uninstallListeners( wheel );
	    uninstallKeyboardActions(wheel);

    	focusInsets = null;
	    insetCache = null;
    	focusRect = null;
	    contentRect = null;
	    wheelRect = null;
        trackListener = null;
        changeListener = null;
        componentListener = null;
        focusListener = null;
	    propertyChangeListener = null;
        wheel = null;
    }

    protected void installDefaults( JWheelControl wheel ) {
        // use slider default properties
        LookAndFeel.installBorder(wheel, "Slider.border");
        LookAndFeel.installColors(wheel, "Slider.background", "Slider.foreground");
        highlightColor = UIManager.getColor("Slider.highlight");
        shadowColor = UIManager.getColor("Slider.shadow");
        focusColor = UIManager.getColor("Slider.focus");
    	focusInsets = (Insets)UIManager.get( "Slider.focusInsets" );
    }

    protected TrackListener createTrackListener( JWheelControl wheel ) {
        return new TrackListener();
    }

    protected ChangeListener createChangeListener( JWheelControl wheel ) {
        return new ChangeHandler();
    }

    protected ComponentListener createComponentListener( JWheelControl wheel ) {
        return new ComponentHandler();
    }

    protected FocusListener createFocusListener( JWheelControl wheel ) {
        return new FocusHandler();
    }

    protected PropertyChangeListener createPropertyChangeListener( JWheelControl wheel ) {
        return new PropertyChangeHandler();
    }

    protected void installListeners( JWheelControl wheel ) {
        wheel.addMouseListener(trackListener);
        wheel.addMouseMotionListener(trackListener);
        wheel.addFocusListener(focusListener);
        wheel.addComponentListener(componentListener);
        wheel.addPropertyChangeListener( propertyChangeListener );
        wheel.getModel().addChangeListener(changeListener);
    }

    protected void uninstallListeners( JWheelControl wheel ) {
        wheel.removeMouseListener(trackListener);
        wheel.removeMouseMotionListener(trackListener);
        wheel.removeFocusListener(focusListener);
        wheel.removeComponentListener(componentListener);
        wheel.removePropertyChangeListener( propertyChangeListener );
        wheel.getModel().removeChangeListener(changeListener);
    }

    protected void installKeyboardActions( JWheelControl wheel ) {
        wheel.registerKeyboardAction(new ActionScroller(wheel, POSITIVE_SCROLL, false),
                                      KeyStroke.getKeyStroke(KeyEvent.VK_RIGHT,0),
                                      JComponent.WHEN_FOCUSED);

        wheel.registerKeyboardAction(new ActionScroller(wheel, NEGATIVE_SCROLL, false),
                                      KeyStroke.getKeyStroke(KeyEvent.VK_DOWN,0),
                                      JComponent.WHEN_FOCUSED);

        wheel.registerKeyboardAction(new ActionScroller(wheel, NEGATIVE_SCROLL, true),
                                      KeyStroke.getKeyStroke(KeyEvent.VK_PAGE_DOWN,0),
                                      JComponent.WHEN_FOCUSED);

        wheel.registerKeyboardAction(new ActionScroller(wheel, NEGATIVE_SCROLL, false),
                                      KeyStroke.getKeyStroke(KeyEvent.VK_LEFT,0),
                                      JComponent.WHEN_FOCUSED);

        wheel.registerKeyboardAction(new ActionScroller(wheel, POSITIVE_SCROLL, false),
                                      KeyStroke.getKeyStroke(KeyEvent.VK_UP,0),
                                      JComponent.WHEN_FOCUSED);

        wheel.registerKeyboardAction(new ActionScroller(wheel, POSITIVE_SCROLL, true),
                                      KeyStroke.getKeyStroke(KeyEvent.VK_PAGE_UP,0),
                                      JComponent.WHEN_FOCUSED);

        wheel.registerKeyboardAction(new ActionScroller(wheel, MIN_SCROLL, true),
                                      KeyStroke.getKeyStroke(KeyEvent.VK_HOME,0),
                                      JComponent.WHEN_FOCUSED);

        wheel.registerKeyboardAction(new ActionScroller(wheel, MAX_SCROLL, true),
                                      KeyStroke.getKeyStroke(KeyEvent.VK_END,0),
                                      JComponent.WHEN_FOCUSED);
    }

    protected void uninstallKeyboardActions( JWheelControl wheel ) {
        wheel.unregisterKeyboardAction( KeyStroke.getKeyStroke( KeyEvent.VK_RIGHT, 0 ) );
        wheel.unregisterKeyboardAction( KeyStroke.getKeyStroke( KeyEvent.VK_DOWN, 0 ) );
        wheel.unregisterKeyboardAction( KeyStroke.getKeyStroke( KeyEvent.VK_PAGE_DOWN, 0 ) );
        wheel.unregisterKeyboardAction( KeyStroke.getKeyStroke( KeyEvent.VK_LEFT, 0 ) );
        wheel.unregisterKeyboardAction( KeyStroke.getKeyStroke( KeyEvent.VK_UP, 0 ) );
        wheel.unregisterKeyboardAction( KeyStroke.getKeyStroke( KeyEvent.VK_PAGE_UP, 0 ) );
        wheel.unregisterKeyboardAction( KeyStroke.getKeyStroke( KeyEvent.VK_HOME, 0 ) );
        wheel.unregisterKeyboardAction( KeyStroke.getKeyStroke( KeyEvent.VK_END, 0 ) );
    }

    public Dimension getPreferredHorizontalSize() {
        return PREFERRED_HORIZONTAL_SIZE;
    }

    public Dimension getPreferredVerticalSize() {
        return PREFERRED_VERTICAL_SIZE;
    }

    public Dimension getMinimumHorizontalSize() {
        return MINIMUM_HORIZONTAL_SIZE;
    }

    public Dimension getMinimumVerticalSize() {
        return MINIMUM_VERTICAL_SIZE;
    }

    public Dimension getPreferredSize(JComponent c)    {
        recalculateIfInsetsChanged();
        Dimension d;
        if ( wheel.getOrientation() == JWheelControl.VERTICAL ) {
            d = new Dimension(getPreferredVerticalSize());
    	    //d.width = insetCache.left + insetCache.right + focusInsets.left + focusInsets.right + wheelRect.width;
        } else {
            d = new Dimension(getPreferredHorizontalSize());
    	    //d.height = insetCache.top + insetCache.bottom + focusInsets.top + focusInsets.bottom + wheelRect.height;
        }
        return d;
    }

    public Dimension getMinimumSize(JComponent c)  {
        recalculateIfInsetsChanged();
        Dimension d;

        if ( wheel.getOrientation() == JWheelControl.VERTICAL ) {
            d = new Dimension(getMinimumVerticalSize());
    	    d.width = insetCache.left + insetCache.right + focusInsets.left + focusInsets.right + wheelRect.width;
        } else {
            d = new Dimension(getMinimumHorizontalSize());
    	    d.height = insetCache.top + insetCache.bottom + focusInsets.top + focusInsets.bottom + wheelRect.height;
        }

        return d;
    }

    public Dimension getMaximumSize(JComponent c) {
        Dimension d = getPreferredSize(c);
        if ( wheel.getOrientation() == JWheelControl.VERTICAL ) {
            d.height = Short.MAX_VALUE;
        } else {
            d.width = Short.MAX_VALUE;
        }

        return d;
    }

    protected void calculateGeometry() {
        calculateFocusRect();
    	calculateWheelRect();
    	calculateFacets();
    }

    protected void calculateFocusRect() {
        focusRect.x = insetCache.left;
    	focusRect.y = insetCache.top;
	    focusRect.width = wheel.getWidth() - (insetCache.left + insetCache.right);
    	focusRect.height = wheel.getHeight() - (insetCache.top + insetCache.bottom);
    }


    protected void calculateWheelRect() {
        Insets padding = ( wheel.getOrientation() == JWheelControl.VERTICAL ? VERTICAL_PADDING : HORIZONTAL_PADDING );

        wheelRect.x = focusRect.x + focusInsets.left + padding.left;
        wheelRect.y = focusRect.y + focusInsets.top + padding.top;
        wheelRect.width = focusRect.width - (focusInsets.left + focusInsets.right + padding.left +  + padding.right);
        wheelRect.height = focusRect.height - (focusInsets.top + focusInsets.bottom + padding.top + padding.bottom);
    }


    public class PropertyChangeHandler implements PropertyChangeListener {
        public void propertyChange( PropertyChangeEvent e ) {
	        String propertyName = e.getPropertyName();
            if (propertyName.equals( "orientation" ) || propertyName.equals( "inverted" ) || propertyName.equals( "labelTable" )
                   || propertyName.equals( "majorTickSpacing" ) || propertyName.equals( "minorTickSpacing" ) || propertyName.equals( "paintTicks" )
                   || propertyName.equals( "paintTrack" ) || propertyName.equals( "paintLabels" ) ) {
 
	            calculateGeometry();

    	    } else if ( propertyName.equals( "model" ) ) {
    	        ((BoundedRangeModel)e.getOldValue()).removeChangeListener( changeListener );
        		((BoundedRangeModel)e.getNewValue()).addChangeListener( changeListener );

        		wheel.repaint();
    	    }
	    }
    }


    public void paint( Graphics g, JComponent c )   {
    }
    
    
    /** This method does the real painting, as opposed to paint().  This is 
        so in order to paint the wheel is painted after the border so it can
        overlap the border a little to look more 3D.  This method is called 
        by the wheel component after the border has been painted. */
    public void postBorderPaint( Graphics g, JComponent c )   {
        recalculateIfInsetsChanged();
	    Rectangle clip = g.getClipBounds();
    	if (clip.intersects( wheelRect ) ) {
    	    paintWheel( g );
    	}
    }

    protected void recalculateIfInsetsChanged() {
        Insets newInsets = wheel.getInsets();
        if ( !newInsets.equals( insetCache ) ) {
    	    insetCache = newInsets;
	        calculateGeometry();
    	}
    }

    public void paintFocus(Graphics g)  {
    	g.setColor( getFocusColor() );

	    BasicGraphicsUtils.drawDashedRect( g, focusRect.x, focusRect.y, focusRect.width, focusRect.height );
    }
    

    final static int numFacets = 34;
    final static double angleIncrement = Math.PI / numFacets;
    final static double anglePairIncrement = 2 * angleIncrement;
    final static double angleOffsetIncrement = angleIncrement / 8;
    final static double angleOffset = 0.0;
    double angleStartOffset = 0.0;
    
    final static Color onZeroColor = new Color(219, 219, 51);
    final static Color nearZeroColor = new Color(51, 202, 51);
    final static Color offZeroColor = new Color(51, 51, 153);
    
    Color[] facetColor = new Color[numFacets];
    int[] facetRightX = new int[numFacets];
    int[] facetBottomY = new int[numFacets];
    int zeroFacet = numFacets / 2;

    transient int[] facetX = new int[4];
    transient int[] facetY = new int[4];
    
    public void valueChanged(boolean increase) {
        
        /*****
        if (increase) {
            angleStartOffset += angleOffsetIncrement;
            if (angleStartOffset > 2 * angleIncrement) {
                angleStartOffset -= 2 * angleIncrement;
            }
        } else {
            angleStartOffset -= angleOffsetIncrement;
            if (angleStartOffset < 0.0) {
                angleStartOffset += 2 * angleIncrement;
            }
        }****/
        
        int valuePerFacetPair = Math.max((wheel.getMaximum() - wheel.getMinimum()) / (numFacets / 2), 2);
        
        double newOffset = angleOffset + (double)(wheel.getValue() % valuePerFacetPair) / (double) valuePerFacetPair * angleIncrement * 2; 
        if (newOffset == angleStartOffset) {
            newOffset += (increase ? angleOffsetIncrement : -angleOffsetIncrement);
            if (newOffset < 0.0) newOffset += anglePairIncrement;
            else if (newOffset >= anglePairIncrement) newOffset -= anglePairIncrement;
        }
        
        angleStartOffset = newOffset;

        calculateFacets();
        wheel.repaint();
    }
    

    void calculateFacets() {

        boolean isHoriz = (wheel.getOrientation() == JWheelControl.HORIZONTAL);
        double radius = (double) (isHoriz ? wheelRect.width : wheelRect.height) * 0.5;
        double shortRadius = (double) (isHoriz ? wheel.getHeight() - wheelRect.y - 1 : wheel.getWidth() - wheelRect.x - 1) * 0.4;

        // control not initialized yet, skip calculations
        if (radius <= 0.0 || shortRadius <= 0.0) {
            return;
        }

        // figure out where 0 is
        int min = wheel.getMinimum();
        int valuePerFacet = Math.max((wheel.getMaximum() - min) / numFacets, 1);
        zeroFacet = (wheel.getValue() - min + (valuePerFacet/2)) / valuePerFacet;

        int longOffset;
        int[] longCoord;
        int shortOffset;
        int[] shortCoord;
        if (isHoriz) {
            longOffset = wheelRect.x;
            longCoord = facetRightX;
            shortOffset = wheelRect.y;
            shortCoord = facetBottomY;
             
        } else {
            longOffset = wheelRect.y;
            longCoord = facetBottomY;
            shortOffset = wheelRect.x;
            shortCoord = facetRightX;
        }
        int highlightRed = 204; //highlightColor.getRed();
        int highlightGreen = 204; //highlightColor.getGreen();
        int highlightBlue = 204; //highlightColor.getBlue();

        double redExtent = 153; //highlightRed - shadowColor.getRed();
        double greenExtent = 153; //highlightGreen - shadowColor.getGreen();
        double blueExtent = 153; //highlightBlue - shadowColor.getBlue();

        // do the facets in pairs, each pair comprising a single tooth on the wheel
        int i;
        int r, g, b;
        double angle;
        double colorScale;
        for (i = 0; i < numFacets; i++) {

            angle = (double) i * angleIncrement + angleStartOffset;

            longCoord[i] = longOffset + (int) Math.round(radius - radius * Math.cos(angle));
            shortCoord[i] = shortOffset + (int) Math.floor(1.5 * shortRadius + shortRadius * Math.sin(angle));

            // create wheel texture by offest odd/even color brightness in opposite directions
            if ((i & 1) == 0) {
                colorScale = Math.abs( Math.sin(angle - 0.1875 * Math.PI) );
            } else {
                colorScale = Math.abs( Math.sin(angle - 0.375 * Math.PI) ); // 3/16ths behind
            }

            r = highlightRed - (int) (colorScale * redExtent);
            g = highlightGreen - (int) (colorScale * greenExtent);
            b = highlightBlue - (int) (colorScale * blueExtent);

            // color the facets surrounding the zero line
            if ( (i == zeroFacet) || (i == zeroFacet + 1) || (i == zeroFacet - numFacets + 1) ) {
                r = (3 * r / 4 + 102) % 255;
                g = (3 * g / 4 + 102) % 255;
                b = b / 2;
            }

            facetColor[i] = new Color(r, g, b);
        }
    }

    public void paintWheel(Graphics g)  {

        int val = wheel.getValue();
        int x = wheelRect.x;
        int y = wheelRect.y;
        int i;
        int iprev = numFacets - 1;
        Color color;

        if ( wheel.getOrientation() == JWheelControl.HORIZONTAL ) {
            
            for (i = 0; i < numFacets; i++) {
                
                facetX[0] = x;
                facetY[0] = y;
                
                facetX[1] = facetRightX[i];
                facetY[1] = y;
                
                facetX[2] = facetRightX[i];
                facetY[2] = facetBottomY[i];
                
                facetX[3] = x;
                facetY[3] = facetBottomY[iprev]; 
                
                g.setColor(facetColor[i]);
                g.fillPolygon(facetX, facetY, 4);
                
                x = facetRightX[i];
                
                iprev = i;
            }

            //g.setColor(Color.yellow);
            //g.drawLine(facetRightX[zeroFacet], y, facetRightX[zeroFacet], facetBottomY[zeroFacet]);

            x = focusRect.x + focusRect.width/2;   // midpoint
            y = focusRect.y;
            color = (val == 0 ? onZeroColor : (val > -3 && val < 3 ? nearZeroColor : offZeroColor));
            g.setColor(color.brighter());
            g.drawLine(x - 1, y, x - 1, y + focusRect.height);
            g.setColor(color);
            g.drawLine(x, y, x, y + focusRect.height);
            g.setColor(color.darker());
            g.drawLine(x + 1, y, x + 1, y + focusRect.height);

        } else {

            for (i = 0; i < numFacets; i++) {
                
                facetX[0] = x;
                facetY[0] = y;
                
                facetX[1] = facetRightX[i];
                facetY[1] = y;
                
                facetX[2] = facetRightX[iprev];
                facetY[2] = facetBottomY[i];
                
                facetX[3] = x;
                facetY[3] = facetBottomY[i]; 
                
                g.setColor(facetColor[i]);
                g.fillPolygon(facetX, facetY, 4);
                
                y = facetBottomY[i];

                iprev = i;
            }
            //g.setColor(Color.green);
            //g.drawLine(x, facetBottomY[zeroFacet], facetRightX[zeroFacet], facetBottomY[zeroFacet]);


            x = focusRect.x;
            y = focusRect.y + focusRect.height/2;   // midpoint
            color = (val == 0 ? onZeroColor : (val > -3 && val < 3 ? nearZeroColor : offZeroColor));
            g.setColor(color.brighter());
            g.drawLine(x, y - 1, x + focusRect.width, y - 1);
            g.setColor(color);
            g.drawLine(x, y, x + focusRect.width, y);
            g.setColor(color.darker());
            g.drawLine(x, y + 1, x + focusRect.width, y + 1);
        }
    }


    public void scrollByBlock(int direction)    {
        synchronized(wheel)    {

            int oldValue = wheel.getValue();
            int blockIncrement = (wheel.getMaximum() - wheel.getMinimum()) / 10;
            int delta = blockIncrement * ((direction > 0) ? POSITIVE_SCROLL : NEGATIVE_SCROLL);

            wheel.setValue(oldValue + delta);
        }
    }

    public void scrollByUnit(int direction) {
        synchronized(wheel)    {

            int oldValue = wheel.getValue();
            int delta = 1 * ((direction > 0) ? POSITIVE_SCROLL : NEGATIVE_SCROLL);

            wheel.setValue(oldValue + delta);
        }
    }

    /**
     * Return the value of a distance in the y direction (assumes orientation is vertical).
     */
    public int valueForYDelta( int yDelta ) {
	    double valuePerPixel = (double) (wheel.getMaximum() - wheel.getMinimum()) / (double) wheelRect.height;
	    int value = (int) Math.round( (double) yDelta * valuePerPixel );

	    if (value == 0) {
	        if (yDelta > 0) value = 1;
	        else if (yDelta < 0) value = -1;
	    }

	    if ( wheel.getInverted()) value = -value;

        return value;
    }

    /**
     * Return the value of a distance in the x direction (assumes orientation is horizontal).
     */
    public int valueForXDelta( int xDelta ) {
	    double valuePerPixel = (double) (wheel.getMaximum() - wheel.getMinimum()) / (double) wheelRect.width;
	    int value = (int) Math.round( (double) xDelta * valuePerPixel );

	    if (value == 0) {
	        if (xDelta > 0) value = 1;
	        else if (xDelta < 0) value = -1;
	    }
	        
	    if ( wheel.getInverted()) value = -value;

        return value;
    }


    /////////////////////////////////////////////////////////////////////////
    /// Model Listener Class
    /////////////////////////////////////////////////////////////////////////
    /**
     * Data model listener.
     *
     * This inner class is marked &quot;public&quot; due to a compiler bug.
     * This class should be treated as a &quot;protected&quot; inner class.
     * Instantiate it only within subclasses of <Foo>.
     */
    public class ChangeHandler implements ChangeListener {
        public void stateChanged(ChangeEvent e) {

	    }
    }

    /////////////////////////////////////////////////////////////////////////
    /// Track Listener Class
    /////////////////////////////////////////////////////////////////////////
    /**
     * Track mouse movements.
     *
     * This inner class is marked &quot;public&quot; due to a compiler bug.
     * This class should be treated as a &quot;protected&quot; inner class.
     * Instantiate it only within subclasses of <Foo>.
     */
    public class TrackListener extends MouseInputAdapter {
        protected transient int offset;
        protected transient int currentMouseX, currentMouseY;

        public void mouseReleased(MouseEvent e)                {
            if ( !wheel.isEnabled() ) {
                return;
            }
            offset = 0;
            isDragging = false;
        }

        /**
        * If the mouse is pressed above the "thumb" component
        * then reduce the scrollbars value by one page ("page up"),
        * otherwise increase it by one page.  If there is no
        * thumb then page up if the mouse is in the upper half
        * of the wheel.
        */
        public void mousePressed(MouseEvent e)                {
            if ( !wheel.isEnabled() ) {
                return;
            }

            currentMouseX = e.getX();
            currentMouseY = e.getY();

            wheel.requestFocus();

            if ( wheelRect.contains(currentMouseX, currentMouseY) ) {
                isDragging = true;
                wheel.setValueIsAdjusting(true);
                switch ( wheel.getOrientation() ) {
                    case JWheelControl.VERTICAL:
                        int ycenter = wheel.getHeight() / 2;
                        if (currentMouseY >= ycenter - 2 && currentMouseY <= ycenter + 2) {
                            wheel.setValue(0);
                        } else if (currentMouseY < ycenter) {
                            scrollByUnit( wheel.getInverted() ? POSITIVE_SCROLL : NEGATIVE_SCROLL ); 
                        } else {
                            scrollByUnit( wheel.getInverted() ? NEGATIVE_SCROLL : POSITIVE_SCROLL ); 
                        }
                        break;
                    case JWheelControl.HORIZONTAL:
                        int xcenter = wheel.getWidth() / 2;
                        if (currentMouseX >= xcenter - 2 && currentMouseX <= xcenter + 2) {
                            wheel.setValue(0);
                        } else if (currentMouseX < xcenter) {
                            scrollByUnit( wheel.getInverted() ? POSITIVE_SCROLL : NEGATIVE_SCROLL ); 
                        } else {
                            scrollByUnit( wheel.getInverted() ? NEGATIVE_SCROLL : POSITIVE_SCROLL ); 
                        }
                        break;
                }
            }
        }


        /**
        * Set the models value to the position of the top/left
        * of the thumb relative to the origin of the wheel.
        */
        public void mouseDragged( MouseEvent e ) {
            if ( !wheel.isEnabled() ) {
                return;
            }

            int oldMouseX = currentMouseX;
            int oldMouseY = currentMouseY;

            currentMouseX = e.getX();
            currentMouseY = e.getY();

            if ( !isDragging ) {
                return;
            }

            switch ( wheel.getOrientation() ) {
                case JWheelControl.VERTICAL:
                    wheel.setValue( wheel.getValue() + valueForYDelta( currentMouseY - oldMouseY ) );
                    break;
                case JWheelControl.HORIZONTAL:
                    wheel.setValue( wheel.getValue() + valueForXDelta( currentMouseX - oldMouseX ) );
                    break;
                default:
                    return;
            }
        }

        public void mouseMoved(MouseEvent e) {
        }

    }


    /**
     * Listener for resizing events.
     * <p>
     * This inner class is marked &quot;public&quot; due to a compiler bug.
     * This class should be treated as a &quot;protected&quot; inner class.
     * Instantiate it only within subclasses of <Foo>.
     */
    public class ComponentHandler extends ComponentAdapter {
        public void componentResized(ComponentEvent e)  {
    	    calculateGeometry();
	        wheel.repaint();
        }
    };

    /**
     * Focus-change listener.
     * <p>
     * This inner class is marked &quot;public&quot; due to a compiler bug.
     * This class should be treated as a &quot;protected&quot; inner class.
     * Instantiate it only within subclasses of <Foo>.
     */
    public class FocusHandler implements FocusListener {
        public void focusGained(FocusEvent e) {
            wheel.repaint();
        }
        public void focusLost(FocusEvent e) {
            wheel.repaint();
        }
    }

    /**
     * Defines the action to take when scrolled.
     * <p>
     * This inner class is marked &quot;public&quot; due to a compiler bug.
     * This class should be treated as a &quot;protected&quot; inner class.
     * Instantiate it only within subclasses of <Foo>.
     */
    public class ActionScroller extends AbstractAction {
        int dir;
        boolean block;
        JWheelControl wheel;

        public ActionScroller( JWheelControl wheel, int dir, boolean block) {
            this.dir = dir;
            this.block = block;
            this.wheel = wheel;
        }

        public void actionPerformed(ActionEvent e) {
            if ( wheel.isEnabled() ) {
                if ( dir == NEGATIVE_SCROLL || dir == POSITIVE_SCROLL ) {
                    int realDir = dir;
                    if ( wheel.getInverted() ) {
                        realDir = (dir == NEGATIVE_SCROLL ? POSITIVE_SCROLL : NEGATIVE_SCROLL);
                    }

                    if ( block ) {
                        scrollByBlock(realDir);
                    } else {
                        scrollByUnit(realDir);
                    }
                } else {
                    if ( wheel.getInverted() ) {
                        if ( dir == MIN_SCROLL ) {
                            wheel.setValue(wheel.getMaximum());
                        } else if ( dir == MAX_SCROLL ) {
                            wheel.setValue(wheel.getMinimum());
                        }
                    } else {
                        if ( dir == MIN_SCROLL ) {
                            wheel.setValue(wheel.getMinimum());
                        } else if ( dir == MAX_SCROLL ) {
                            wheel.setValue(wheel.getMaximum());
                        }
                    }
                }
            }
        }

        public boolean isEnabled() {
            return true;
        }
    }
}
