/*
 * @(#)GoalPostLayout.java
 *
 * Copyright (c) 1999 by Trapezium Development LLC.  All Rights Reserved.
 *
 * The information in this file is the property of Trapezium Development LLC
 * and may be used only in accordance with the terms of the license granted
 * by Trapezium.
 */
package com.trapezium.util;

import java.awt.*;

/** GoalPostLayout lays out up to four components in a pattern like this:

                    1   2
                    33333
                      4

    The first component is positioned at the top left, the second at the top
    right, the third sized horizontally to the parent width (minus insets)
    and the fourth at bottom center.  The vertical position of the third
    component depends on whether there is a fourth component.  If the 
    container has four ot more components, the third component is centered 
    vertically, otherwise it is positioned at the bottom.

*/
public class GoalPostLayout implements LayoutManager {

    int hmargin = 0;
    int vmargin = 0;

    public GoalPostLayout() {
    }

    public GoalPostLayout(int hmargin, int vmargin) {
        this.hmargin = hmargin;
        this.vmargin = vmargin;
    }

	public Dimension preferredLayoutSize(Container parent) {
        Insets insets = parent.getInsets();
        Dimension pref = new Dimension(insets.left + insets.right + 2 * hmargin, insets.top + insets.bottom + 2 * vmargin);
        int num = parent.getComponentCount();

        if (num > 1) {
            Dimension p0 = parent.getComponent(0).getPreferredSize();
            Dimension p1 = parent.getComponent(1).getPreferredSize();
            pref.width += p0.width + p1.width;
            pref.height += Math.max(p0.height, p1.height);
            if (num > 2) {
                Dimension p2 = parent.getComponent(2).getPreferredSize();
                pref.width = Math.max(pref.width, p2.width + insets.left + insets.right + 2 * hmargin);
                pref.height += p2.height;

                if (num > 3) {
                    Dimension p3 = parent.getComponent(3).getPreferredSize();
                    pref.width = Math.max(pref.width, p3.width + insets.left + insets.right + 2 * hmargin);
                    pref.height += p3.height;
                }
            }
        } else if (num == 1) {
            pref = parent.getComponent(0).getPreferredSize();
        }

        return pref;
    }

    public Dimension minimumLayoutSize(Container parent) {
        Insets insets = parent.getInsets();
        Dimension min = new Dimension(insets.left + insets.right + 2 * hmargin, insets.top + insets.bottom + 2 * vmargin);
        int num = parent.getComponentCount();

        if (num > 1) {
            Dimension m0 = parent.getComponent(0).getMinimumSize();
            Dimension m1 = parent.getComponent(1).getMinimumSize();
            min.width += m0.width + m1.width;
            min.height += Math.max(m0.height, m1.height);
            if (num > 2) {
                Dimension m2 = parent.getComponent(2).getMinimumSize();
                min.width = Math.max(min.width, m2.width + insets.left + insets.right + 2 * hmargin);
                min.height += m2.height;

                if (num > 3) {
                    Dimension m3 = parent.getComponent(3).getMinimumSize();
                    min.width = Math.max(min.width, m3.width + insets.left + insets.right + 2 * hmargin);
                    min.height += m3.height;
                }
            }
        } else if (num == 1) {
            min = parent.getComponent(0).getMinimumSize();
        }

        return min;
    }

    public void layoutContainer(Container parent) {
        synchronized (parent.getTreeLock()) {
        	Insets insets = parent.getInsets();
        	int num = parent.getComponentCount();

            if (num > 1) {
                
                Dimension size = parent.getSize();
                
                Component c0 = parent.getComponent(0);
                Dimension p0 = c0.getPreferredSize();

                Component c1 = parent.getComponent(1);
                Dimension p1 = c1.getPreferredSize();
                
                int w0 = p0.width;
                int w1 = p1.width;
                
                int wtot = size.width - insets.left - insets.right - 2 * hmargin;
                if (w0 + w1 > wtot) {
                    // it would be pretty strange for the following test to
                    // fail (size.width would have to be negative) but nobody
                    // enjoys a DBZ
                    if (w0 + w1 > 0) {
                        float ratio = (float) w0 / (float) (w0 + w1);
                        w0 = (int) (wtot * ratio);
                        w1 = wtot - w0;
                    }
                }
                
                c0.setBounds(insets.left + hmargin, insets.top, w0, p0.height);
                c1.setBounds(size.width - insets.right - hmargin - w1, insets.top + vmargin, w1, p1.height);
                

                if (num > 2) {
                    Component c2 = parent.getComponent(2);
                    Dimension p2 = c2.getPreferredSize();
                    if (num == 3) {
                        c2.setBounds(insets.left + hmargin, size.height - insets.bottom - vmargin - p2.height, wtot, p2.height);

                    } else {                
                        c2.setBounds(insets.left + hmargin, (size.height - p2.height)/2, wtot, p2.height);
                        Component c3 = parent.getComponent(3);
                        Dimension p3 = c3.getPreferredSize();
                        c3.setBounds((size.width - p3.width)/2, size.height - insets.bottom - vmargin - p3.height, p3.width, p3.height);
                    }
                }
            } else if (num == 1) {
                Component c = parent.getComponent(0);
                Dimension pref = c.getPreferredSize();
                c.setBounds(insets.left + hmargin, insets.top + vmargin, pref.width, pref.height);
            }
        }
    }


    /**
     * Not used.
     *
     * @param name the name of the component
     * @param component the component
     */
    public void addLayoutComponent(String name, Component component) {
    }

    /**
     * Not used.
     *
     * @param component the component
     */
    public void removeLayoutComponent(Component component) {
    }

}
