/*
 * @(#)ResourceManager.java
 *
 * Copyright (c) 1999 by Trapezium Development LLC.  All Rights Reserved.
 *
 * The information in this file is the property of Trapezium Development LLC
 * and may be used only in accordance with the terms of the license granted
 * by Trapezium.
 *
 */
package com.trapezium.util;

import com.trapezium.factory.FactoryDataUpdater;

import java.util.*;
import java.io.File;
import java.awt.*;
import java.awt.event.*;
import java.net.URL;

import javax.swing.*;

/**
 *  Localizable resources for browser applications.
 *
 *  @author          Michael St. Hippolyte
 *  @version         1.0, 23 Feb 1999
 *
 *  @since           1.0
 */
abstract public class ResourceManager extends ListResourceBundle {

    // default display colors
	final static Color DEFAULT_TOOLBARCOLOR = new Color(102, 96, 96);
	final static Color DEFAULT_STATUSBARCOLOR = DEFAULT_TOOLBARCOLOR;
	final static Color DEFAULT_TEXTCOLOR = new Color(255, 224, 192);

    //
    // private stuff
    //

    private static Hashtable imageHash = new Hashtable();
	private static ResourceBundle resources;

	public static ResourceBundle getDefaultResourceBundle() {
		return resources;
	}


    /** Preload button images for toolbar, since this seems related to jdk1.1.7a hang
     *  and possibly SGI & Solaris java hangs on startup
     */
    public static void preloadImages() {
        String[] toolKeys = tokenize(getResourceString("toolbar"));
        for (int i = 0; i < toolKeys.length; i++) {
            if (!toolKeys[i].equals("-")) {
                //System.out.println("ToolBar creating button for " + toolKeys[i]);
                String key = toolKeys[i];
                Image image = getImage(key);
                if (image == null) {
                  System.out.println(key + " image not found.");
                }
                imageHash.put(key,image);
            }
        }
    }

    public static String getResourceString(String nm) {
        String str;
        if (nm == null) {
            str = null;
        } else try {
            str = resources.getString(nm);
        } catch (MissingResourceException mre) {
            str = null;
        }
        return str;
    }

    /** Get a URL for a resource, look in JAR file if possible, or disk if not */
    public static URL getResource(String key) {
        String name = getResourceString(key);
        return( getResourceByName( name ));
    }

    public static URL getResourceByName( String name ) {
        if (name != null) {
            URL url = resourceURL(name);
            // couldn't find it?  look in the local directory instead of the classpath
            if (url == null) {
                url = resourceURL(name, false);
            }
            return url;
        }
        return null;
    }


    private static URL base = null;
    private static String resourceName(String file) {
        if (file.startsWith("~")) {
            file = System.getProperty("user.home") + file.substring(1);
        }
        return "file:" + (new File(file)).getAbsolutePath().replace(File.separatorChar, '/');
    }

    private static String resourceURLString(String file) {
        URL url = resourceURL(file);
        if (url == null) {
            return null;
        }
        return url.toExternalForm();
    }

    /** Get a URL of a file */
    public static URL resourceURL(String file) {
        return resourceURL(file, true);
    }

    public static URL resourceURL(String file, boolean loadFromClasspath) {
        if (loadFromClasspath) {
            return file.getClass().getResource("/" + file);

        } else {
            String resource = resourceName(file);
            URL url;
            try {
                url = new URL(base, resource);
            } catch (Exception e) {
                url = null;
            }
            return url;
        }
    }

    /**
     * Suffix applied to the key used in resource file
     * lookups for an image.
     */
    private static final String imageSuffix = "Image";

    /**
     * Suffix applied to the key used in resource file
     * lookups for a label.
     */
    private static final String labelSuffix = "Label";

    /**
     * Suffix applied to the key used in resource file
     * lookups for an action.
     */
    private static final String actionSuffix = "Action";

    /**
     * Suffix applied to the key used in resource file
     * lookups for tooltip text.
     */
    private static final String tipSuffix = "Tooltip";

    /**
     * Create a JMenuBar for the app.  By default this pulls the
     * definition of the menu from the associated resource file.
     */
    public static JMenuBar createMenuBar(JButton censusData) {
        String[] menuKeys = tokenize(getResourceString("menubar"));
        JMenuBar jbar = new JMenuBar();
        try {
            for (int i = 0; i < menuKeys.length; i++) {
                JMenu m = createJMenu( menuKeys[i]);
                if (m != null) {
                    jbar.add(m);
                }
            }
        } catch (Exception e) {
            System.out.println("Exception in createMenuBar: " + e);
        }
        if (censusData != null ) {
            jbar.add(Box.createHorizontalGlue());
            jbar.add(censusData);
        }
        return jbar;
    }

    /** if the entry for a menu item in the resource file has this prefix, the 
      * menu item is interpreted as the name of a submenu rather than a command.
      */
    static final String SUBMENU_PREFIX = ">>";

     /* if the entry for a menu item in the resource file has one of these 
      * prefixes, the menu item is interpreted as a checkable item, initially
      * checked or unchecked depending on the prefix
      */
    static final String CHECKEDMENUITEM_PREFIX = "!";
    static final String UNCHECKEDMENUITEM_PREFIX = "?";

     /* radiobutton-sytle menu items can be created using these prefixes.
      */
    static final String FIRSTRADIOMENUITEM_PREFIX = "((";
    static final String CHECKEDFIRSTRADIOMENUITEM_PREFIX = "(!";
    static final String LASTRADIOMENUITEM_PREFIX = "))";
    static final String CHECKEDLASTRADIOMENUITEM_PREFIX = ")!";
    
    /**
     * Create a JMenu based on a key
     */
    public static JMenu createJMenu(String key) {
        String[] itemKeys = tokenize(getResourceString(key));
        JMenu menu = new JMenu(getResourceString(key + labelSuffix));
        ButtonGroup radioGroup = null;
        
        for (int i = 0; i < itemKeys.length; i++) {
            String itemKey = itemKeys[i];
            if (itemKey.equals("-")) {
                menu.addSeparator();
            
            } else if (itemKey.startsWith(SUBMENU_PREFIX)) {
                itemKey = itemKey.substring(SUBMENU_PREFIX.length());
                JMenu submenu = createJMenu(itemKey);
                menu.add(submenu);
            
            } else try {
                boolean checkable = false;
                boolean checked = false;
                boolean endradio = false;
                if (itemKey.startsWith(CHECKEDMENUITEM_PREFIX)) {
                    itemKey = itemKey.substring(CHECKEDMENUITEM_PREFIX.length());
                    checkable = true;
                    checked = true;
                } else if (itemKey.startsWith(UNCHECKEDMENUITEM_PREFIX)) {
                    itemKey = itemKey.substring(UNCHECKEDMENUITEM_PREFIX.length());
                    checkable = true;
                } else if (itemKey.startsWith(FIRSTRADIOMENUITEM_PREFIX)) {
                    itemKey = itemKey.substring(FIRSTRADIOMENUITEM_PREFIX.length());
                    radioGroup = new ButtonGroup();
                    checkable = true;
                } else if (itemKey.startsWith(CHECKEDFIRSTRADIOMENUITEM_PREFIX)) {
                    itemKey = itemKey.substring(CHECKEDFIRSTRADIOMENUITEM_PREFIX.length());
                    radioGroup = new ButtonGroup();
                    checkable = true;
                    checked = true;
                } else if (itemKey.startsWith(LASTRADIOMENUITEM_PREFIX)) {
                    itemKey = itemKey.substring(LASTRADIOMENUITEM_PREFIX.length());
                    checkable = true;
                    endradio = true;
                } else if (itemKey.startsWith(CHECKEDLASTRADIOMENUITEM_PREFIX)) {
                    itemKey = itemKey.substring(CHECKEDLASTRADIOMENUITEM_PREFIX.length());
                    checkable = true;
                    endradio = true;
                    checked = true;
                } 
            
                JMenuItem mi = createJMenuItem(itemKey, checkable, radioGroup);
                if (checked) {
                    mi.setSelected(true);
                }
                if (endradio) {
                    radioGroup = null;
                }
                menu.add(mi);
            } catch (Exception e) {
                System.out.println("Exception in createJMenu for item '" + itemKey + "': " + e);
                e.printStackTrace();
            }
        }
        menu.validate();
        return menu;
    }

    /** Create the JMenuItem for a command.  The owning application is passed in
        to get the Action for the command */
    public static JMenuItem createJMenuItem(String cmd, boolean checkable, ButtonGroup group) {
        String label = getResourceString(cmd + labelSuffix);
        JMenuItem mi;
        if (checkable) {
            if (group == null) {
                mi = new JCheckBoxMenuItem(label);
            } else {
                mi = new JRadioButtonMenuItem(label);
                group.add(mi);
            }
        } else {
            mi = new JMenuItem(label);
        }
        URL url = getResource(cmd + imageSuffix);
        if (url != null) {
            mi.setHorizontalTextPosition(JButton.RIGHT);
            mi.setIcon(new ImageIcon(url));
        }
        String astr = getResourceString(cmd + actionSuffix);
        if (astr == null) {
            astr = cmd;
        }
        mi.setActionCommand(astr);

        ActionListener a = BaseAction.getAction(astr);
        if (a != null) {
            mi.addActionListener(a);
            if (a instanceof ItemListener) {
                mi.addItemListener( (ItemListener) a );
            }
            
        } else {
            System.out.println("No support for menu item " + cmd + " in this application");
            mi.setEnabled(false);
        }
        return mi;
    }

    /**
     * Create a toolbar.
     */
    public static Container createToolBar() {

        Container bar = null;
        if ( bar != null ) {
            return( bar );
        }

        bar = new JToolBar();
        bar.setBackground(DEFAULT_TOOLBARCOLOR);
        bar.setForeground(DEFAULT_TEXTCOLOR);
        String[] toolKeys = tokenize(getResourceString("toolbar"));
        for (int i = 0; i < toolKeys.length; i++) {
            if (toolKeys[i].equals("-")) {
                bar.add(Box.createHorizontalStrut(5));
            } else {
                Component b = createButton(toolKeys[i]);
                if (b != null) {
                    bar.add(b);
                }
            }
        }
        bar.add(Box.createHorizontalGlue());
        bar.setSize(bar.getPreferredSize());

        return bar;
    }
    public static Container createStatusBar() {
        StatusBar bar = new StatusBar();
        bar.setBackground(DEFAULT_STATUSBARCOLOR);
        bar.setForeground(DEFAULT_TEXTCOLOR);
        FactoryDataUpdater.setDefaultProgressIndicator(bar);
        return bar;
    }

    /**
     * Create a button suitable for a toolbar.
     */
    public static Component createButton(String key) {

        Image image = (Image)imageHash.get(key);
        if ( image == null ) {
            image = getImage(key);
        }
        if (image == null) {
            System.err.println("Image not available for " + key);
            return null;
        }
        String astr = getActionCommand(key);
        if (astr == null) {
            astr = key;
        }
        ActionListener a = BaseAction.getAction(astr);

        JButton button = new JButton(new ImageIcon(image)) {
                public void requestFocus() {}  // not allowed on toolbar buttons
                public float getAlignmentY() { return 0.5f; }
            };
        
        String tip = getResourceString(key + tipSuffix);
        if (tip != null) {
            button.setToolTipText(tip);
        }
        if (a != null) {
            button.setActionCommand(astr);
            button.addActionListener(a);
        } else {
            button.setEnabled(false);
        }
        return button;
    }


    /**
     * Take the given string and chop it up into a series
     * of strings on whitespace boundries.  This is useful
     * for trying to get an array of strings out of the
     * resource file.
     */
    private static String[] tokenize(String input) {
        Vector v = new Vector();
        StringTokenizer t = new StringTokenizer(input);
        String cmd[];

        while (t.hasMoreTokens()) {
            v.addElement(t.nextToken());
        }
        cmd = new String[v.size()];
        for (int i = 0; i < cmd.length; i++) {
            cmd[i] = (String) v.elementAt(i);
        }

        return cmd;
    }


    public static String getImageName(String key) {
        return getResourceString(key + imageSuffix);
    }

    public static String getLabel(String key) {
        return getResourceString(key + labelSuffix);
    }

    public static String getActionCommand(String key) {
        return getResourceString(key + actionSuffix);
    }

    public static String getTip(String key) {
        return getResourceString(key + tipSuffix);
    }

    public static Image getImage(String key) {
        URL url = getResource(key + imageSuffix);
        if (url == null) {
            System.out.println("No image resource for " + key);
            return null;
        } else {
            //System.out.println("Loading image resource for " + key + " URL: " + url.toExternalForm());
            Image image = null;
            try {
                image = Toolkit.getDefaultToolkit().createImage((java.awt.image.ImageProducer) url.getContent());
            } catch (Exception e) {
                System.out.println("Error loading image " + url.toExternalForm());
                image = null;
            }
            return image;
        }
    }



    public static void init(String baseName) {
		System.out.println("Loading resources..." + baseName);
		try {
            resources = ResourceBundle.getBundle(baseName, Locale.getDefault());
        } catch (MissingResourceException mre) {
            System.err.println("Resource class file " + baseName + ".class not found");
   			System.exit(1);
        }
        preloadImages();
    }

}