/*
 * @(#)ComposableObject.java
 *
 * Copyright (c) 1999 by Trapezium Development LLC.  All Rights Reserved.
 *
 * The information in this file is the property of Trapezium Development LLC
 * and may be used only in accordance with the terms of the license granted
 * by Trapezium.
 *
 */
package com.trapezium.virtualworld;

import com.trapezium.browser.Viewable;
import com.trapezium.util.GlobalProgressIndicator;
import java.util.Vector;
import java.util.Enumeration;

/**
 *  A ComposableObject is the base class for any object in the virtualworld object hierarchy.
 *
 *  Each ComposableObject contains zero or more children, which are also ComposableObjects.
 *
 *  @author          Michael St. Hippolyte
 *  @version         1.0, 15 June 1999
 *
 *  @since           1.0
 */

public class ComposableObject implements java.io.Serializable {

	/** Used to track number of ComposableObjects created */
	static public int createCount = 0;

	/** parent */
	ComposableObject parent = null;

	/** children, can be a Vector, an array or a single child.  Does not include
	    children which are temporary. */
	Object children;

	boolean temporary;

	/** default constructor */
	public ComposableObject() {
	    this(true);
	}

    /** constructor for use by convenience objects generated by the program, which
        don't get counted */
    protected ComposableObject(boolean persistent) {
        if (persistent) {
            createCount++;
            temporary = false;
        } else {
            temporary = true;
        }
    }


    public boolean isTemporary() {
        return temporary;
    }

    /** Write out the object for serialization, mark progress if GlobalProgressIndicator
     *  is set up.
     */
    private void writeObject(java.io.ObjectOutputStream oos) throws java.io.IOException {
        GlobalProgressIndicator.markProgress();
        oos.defaultWriteObject();
    }

    /** Read in serialized object, mark progress if GlobalProgressIndicator is set up.
     */
    private void readObject(java.io.ObjectInputStream oos) throws java.io.IOException {
        try {
            GlobalProgressIndicator.markProgress();
            oos.defaultReadObject();
        } catch ( ClassNotFoundException e ) {
            e.printStackTrace();
            throw new java.io.IOException();
        }
    }


	/** get the class name without package */
	public String getBaseName() {
		String className = getClass().getName();
		int index = className.lastIndexOf( '.' ) + 1;
		return( className.substring( index ));
	}

	/** Add a child */
	public void addChild( ComposableObject child ) {

		if ( child == null ) {
			return;
		}

		// optimization, one child added directly, if more than one
		// create Vector
		if ( children == null ) {
			children = child;

        } else {
            if ( children instanceof ComposableObject[]) {
                int len = ((ComposableObject[])children).length;
                Vector v = new Vector(len + 1);
                for (int i = 0; i < len; i++) {
                    v.addElement(((ComposableObject[])children)[i]);
                }
                v.addElement(child);
                children = v;

    		} else if ( children instanceof Vector ) {
    		    ((Vector)children).addElement( child );

    		} else {
    		    Object original = children;
    		    children = new Vector();
    		    ((Vector)children).addElement( original );
    		    ((Vector)children).addElement( child );
    		}
    		child.setParent(this);
		}

	}

	/** Remove a child.
	 *
	 *  @param child child to remove
	 *  @throws ObjectNotFoundException if the child is not part
	 *     of this Object
	 */
	public void removeChild( Object child ) throws ObjectNotFoundException {
	    if ( !contains( child )) {
	        throw new ObjectNotFoundException("Cannot remove " + child.toString() + "; not a child of this object (" + toString() + ")");
	    }
        if ( children instanceof Vector ) {
			((Vector)children).removeElement( child );

        } else if ( children instanceof ComposableObject[]) {
            int len = ((ComposableObject[])children).length;
            Vector v = new Vector(len);
            for (int i = 0; i < len; i++) {
                ComposableObject obj = ((ComposableObject[])children)[i];
                if (!obj.equals(child)) {
                    v.addElement(obj);
                }
            }
            children = v;


    	} else {
    	    children = null;
    	}
	}

	/** Check if a specific child exists.
	 *
	 *  @param child child to check for
	 *  @return true if child found, otherwise false
	 */
	public boolean contains( Object child ) {
	    if ( children != null ) {
	        if ( children instanceof Vector ) {
	            return( ((Vector)children).contains( child ));

            } else if ( children instanceof ComposableObject[]) {
                ComposableObject[] array = (ComposableObject[]) children;
                for (int i = 0; i < array.length; i++) {
                    if (array[i].equals(child)) {
                        return( true );
                    }
                }
                return( false );

	        } else {
	            return( children == child );
	        }
	    } else {
	        return( false );
	    }
	}

    public Enumeration getChildren() {
        if (children instanceof Vector) {
            return ((Vector) children).elements();
        }
        return new ChildrenEnumerator();
    }

    class ChildrenEnumerator implements Enumeration {
        int index = 0;
        public ChildrenEnumerator() {
        }
        public boolean hasMoreElements() {
            if (children == null) {
                return false;
            } else if (children instanceof ComposableObject[]) {
                return (index < ((ComposableObject[]) children).length);
            } else {
                return index == 0;
            }
        }
        public java.lang.Object nextElement() {
            if (children instanceof ComposableObject[]) {
                return ((ComposableObject[]) children)[index++];
            } else {
                index++;
                return children;
            }
        }
    }

	/** Get the specific child.
	 *
	 * @param offset the offset of the child, from 0 to (numberChlidren-1)
	 * @return the child, null if none or the offset value out
	 *     of range.
	 */
	public ComposableObject getChildAt( int offset ) {
		if ( children == null ) {
			return( null );

		} else if ( children instanceof Vector ) {
		    Vector vchildren = (Vector)children;
            if ( vchildren.size() <= offset ) {
    			return( null );
    		} else {
    			return( (ComposableObject)vchildren.elementAt( offset ));
    		}

        } else if ( children instanceof ComposableObject[]) {
            ComposableObject[] array = (ComposableObject[]) children;
            if (offset >= array.length) {
                return( null );
            } else {
                return( array[offset] );
            }

		} else if ( offset == 0 ) {
		    return( (ComposableObject)children );

		} else {
		    return( null );
		}
	}
	/** Get the number of children */
	public int numberChildren() {
		if ( children == null ) {
			return( 0 );
		} else if ( children instanceof Vector ) {
		    Vector vchildren = (Vector)children;
			return( vchildren.size() );
        } else if ( children instanceof ComposableObject[]) {
            return( ((ComposableObject[]) children).length );
		} else {
		    return( 1 );
		}
	}

	/** set a child's parent */
	public void setParent( ComposableObject p ) {
		parent = p;
	}

	/** get this element's parent */
	public ComposableObject getParent() {
		return( parent );
	}


    /** Get the root of the object tree */
	public ComposableObject getRoot() {
		if ( parent == null ) {
			return( this );
		} else {
			return( parent.getRoot() );
		}
	}
}
