/*
 * @(#)VisualObject.java
 *
 * Copyright (c) 1998-1999 by Trapezium Development LLC.  All Rights Reserved.
 *
 * The information in this file is the property of Trapezium Development LLC
 * and may be used only in accordance with the terms of the license granted
 * by Trapezium.
 */
package com.trapezium.virtualworld;

import com.trapezium.browser.Visual;
import com.trapezium.browser.Viewable;
import com.trapezium.browser.vrml.node.RuntimeField;
import com.trapezium.browser.vrml.node.RealizableNode;
import com.trapezium.generator.Generator;
import com.trapezium.vrml.Scene;
import com.trapezium.vrml.VrmlElement;
import com.trapezium.vrml.node.Node;
import com.trapezium.vrml.node.DEFUSENode;


import java.awt.*;
import java.awt.image.*;
import java.io.PrintWriter;
import java.util.Vector;
import java.util.Enumeration;
import javax.vecmath.*;


/** a VisualObject is an object capable of generating its own visual
    representation. */

public class VisualObject extends ComposableObject implements Viewable, Visual, java.io.Serializable {

    /** constructor */
    public VisualObject() {
    }
    
    /** get the most immediate ancestor which is a VisualObject, or null if this object
        doesn't have an ancestor which is a VisualObject. */
    public VisualObject getVisualParent() {
        ComposableObject obj = getParent();
        while (obj != null && !(obj instanceof VisualObject)) {
            obj = obj.getParent();
        }
        
        return (VisualObject) obj;
    }

    /** the display resolution of this object.  Resolution is expressed as
        a percentage of depth, so 100 is full resolution. */
    int resolution = 100;

    int getDepth() {
        // the depth is equal to the depth of the deepest child plus one
        int depth = 0;
        Enumeration children = getChildren();
        while (children.hasMoreElements()) {
            ComposableObject obj = (ComposableObject) children.nextElement();
            if (obj instanceof VisualObject) {
                depth = Math.max( depth, ((VisualObject) obj).getDepth() );
            }
        }
        depth++;
        return depth;
    }


    Generator generator = null;

    /** the generator for recreating the node tree */
	public Generator getGenerator() {
	    if (generator == null) {
            generator = new TransformGenerator();
        }
        return generator;
	}

    /** set the generator for this object */
    public void setGenerator( Generator gen ) {
        generator = gen;
    }

    /** the node tree representing the object, stored as a Scene object */
    Scene scene = null;

    /** the scene preloader thread. */
    Thread preloader = null;

    /** define the scene for this object */
    public void setScene(Scene scene) {
        this.scene = scene;
        long timestamp = scene.getTime();
        setFieldValue(scene, "translation", loc, timestamp);
        setFieldValue(scene, "orientation", orient, timestamp);
        setFieldValue(scene, "scale", scale, timestamp);
        setFieldValue(scene, "transparency", new Float(trans), timestamp);
    }

    /** Set the value for a field for all the children of a subgraph.  If no children
     *  have this field, recursively checks grandchildren.  Stops recursing when the
     *  field is found or gets to a leaf node.
     */
    static public void setFieldValue(VrmlElement subgraph, String fieldname, Object value, long timestamp ) {
        int nChildren = subgraph.numberChildren();
        // first look for the field at this level
        boolean found = false;
        for ( int i = 0; i < nChildren; i++ ) {
            VrmlElement e = subgraph.getChildAt( i );
            if ( e instanceof Node ) {
                Node n = (Node)e;
                if ( n instanceof DEFUSENode ) {
                    DEFUSENode dun = (DEFUSENode)n;
                    n = n.getNode();
                }
                if ( n != null ) {
                    RuntimeField rf = RuntimeField.getRuntimeField(n, fieldname);
                    if (rf != null) {
                        found = true;
                        rf.setValue(value);
                        if (n instanceof RealizableNode) {
                            ((RealizableNode) n).updateField(fieldname, timestamp);
                        }
                    }
                }
            }
        }
        
        // not found yet; recurse to next level
        for ( int i = 0; i < nChildren; i++ ) {
            VrmlElement e = subgraph.getChildAt( i );
            if ( e instanceof Node ) {
                Node n = (Node)e;
                if ( n instanceof DEFUSENode ) {
                    DEFUSENode dun = (DEFUSENode)n;
                    n = n.getNode();
                }
                if ( n != null ) {
                    setFieldValue( n, fieldname, value, timestamp );
                }
            }
        }
    }

    /** return the scene, blocking if necessary.  The base clase assembles the
        scenes of its children. */
    public Scene getScene() {
        return getScene(true);
	}


    /** get the scene if available; block only if waitForScene is true.  If
        waitForScene is false and the scene has not yet been generated,
        trigger preloading, which is done by a background thread. */
    public Scene getScene(boolean waitForScene) {
        if (scene == null) {
            if (!waitForScene) {
                if (preloader == null) {
                    preloader = new Thread(getGenerator());
                    preloader.setDaemon(true);
                    preloader.start();
                } else {
                    // done yet?
                    if (!preloader.isAlive()) {
                        scene = getGenerator().getScene();
                    }
                }
            } else {
                scene = getGenerator().getScene();
            }
        }
        return( scene );
    }


    // Viewable interface

    /** return the depiction type to use by default */
    public int getPreferredDepiction() {
        return GRAPHIC3D;
    }

    /** Resolution ranges from -1.0 to 1.0, and is combined with
     *  the level parameter to determine how much detail to depict
     *
     *      -1.0  Minimal depiction, no detail at any level
     *      -0.5  Half the detail as normal
     *       0.0  Normal depiction, detail varies directly with level
     *       0.5  More detail than normal (halfway between normal and maximal)
     *       1.0  Maximal depiction, maximum detail at every level
     */
    public void setResolution(double res) {
        resolution = (int) ((res * 50.0) + 50.0);   // convert to percentage
    }



    /** Get the text depiction of this object at a particular level of
     *  detail, in the form of a string.
     */
    public String getTextDepiction() {
        return "";
    }



    /** Get the XML depiction of this object, in the form of a DOM node.
     */
    public org.w3c.dom.Node getXMLDepiction() {
        return null;
    }

    /** Get the 2D depiction of this object at a particular level of
     *  detail, in the form of a Java2D Raster.
     */
    public java.awt.image.Raster get2DDepiction() {
        return null;
    }

    /** Get the 3D depiction of this object, in the form of a Scene object.
     */
    public Scene get3DDepiction() {
        // wait for scene to load
        return getScene(true);
    }

    /** get the title of this object as a String */
    public String getTitle() {
        Scene scene = getScene(false);
        if (scene != null) {
            return scene.getText();
        }
        return "";
    }

    /** get the metainformation available for this object as a String */
    public String getMetaInfo() {
        return "";
    }


    public String getUrl() {
        if (generator != null) {
            return generator.getUrl();
        } else {
            return null;
        }
    } 

    // Visual interface

	/* visual properties */
    float[] loc = { 0f, 0f, 0f };
    float[] orient = { 0f, 1f, 0f, 0f };
    float[] scale = { 0f, 0f, 0f };;
    float trans = 0.0f;


    public Point3f getLocation() {
        Point3f pt = new Point3f();
        getLocation(pt);
        return pt;
    }

    /** Get the location of this object.  If the object is in a scene, returns
     *  the current location of the object in the scene, otherwise returns the
     *  value most recently set by setLocation or 0,0,0 if setLocation has
     *  never been called.
     */
    public void getLocation( Point3f returnPt ) {
        returnPt.x = loc[0];
        returnPt.y = loc[1];
        returnPt.z = loc[2];
    }

    public void setLocation( Point3f pt ) {
        setLocation(pt.x, pt.y, pt.z);
    }

    /** Set the location of this object.  If the object is in a scene, changes
     *  the current location of the object; otherwise, changes the initial location
     *  which the object will have when it is added to a scene.
     */
    public void setLocation( float x, float y, float z ) {
        loc[0] = x;
        loc[1] = y;
        loc[2] = z;
    }

    public Vector4f getOrientation() {
        Vector4f vec = new Vector4f();
        getOrientation(vec);
        return vec;
    }

    public void getOrientation( Vector4f returnOrient ) {
        returnOrient.x = orient[0];
        returnOrient.y = orient[1];
        returnOrient.z = orient[2];
        returnOrient.w = orient[3];
    }

    public void setOrientation( Vector4f vec ) {
        orient[0] = vec.x;
        orient[1] = vec.y;
        orient[2] = vec.z;
        orient[3] = vec.w;
    }

    public void setOrientation( float yaw, float pitch, float roll ) {
        // have to cut and paste in the math from VizModelInfo or wherever it is
        System.out.println("setOrientation( float yaw, float pitch, float roll ) not supported");
    }

    public Color getColor() {
        VisualObject obj = getVisualParent();
        if (obj == null) {
            return null;
        } else {
            return obj.getColor();
        }
    }
       
    public Color getColor(String fieldName) {
        VisualObject obj = getVisualParent();
        if (obj == null) {
            return null;
        } else {
            return obj.getColor(fieldName);
        }
    }

    public void setColor(Color color) {
        setColor( "diffuseColor", color );
    }

    public void setColor(String fieldName, Color color) {
        Enumeration children = getChildren();
        while (children.hasMoreElements()) {
            ComposableObject obj = (ComposableObject) children.nextElement();
            if (obj instanceof VisualObject) {
                ((VisualObject) obj).setColor(fieldName, color);
            }
        }
        
    }

    public Vector3f getScale() {
        Vector3f vec = new Vector3f();
        getScale(vec);
        return vec;
    }
    
    public void getScale( Vector3f returnScale ) {
        returnScale.x = scale[0];
        returnScale.y = scale[1];
        returnScale.z = scale[2];
    }
        

    public void setScale( Vector3f vec ) {
        scale[0] = vec.x;
        scale[1] = vec.y;
        scale[2] = vec.z;
    }
        
    public void setScale( float xScale, float yScale, float zScale ) {
        scale[0] = xScale;
        scale[1] = yScale;
        scale[2] = zScale;
    }
        

    public float getTransparency() {
        return trans;
    }
    
    public void setTransparency( float transparency ) {
        trans = transparency;
    }
    

    /** the default generator for VisualObjects generates a Transform from
        the visual properties */
	class TransformGenerator extends Generator {

	    public void generate( PrintWriter pw ) {
            int depth = ((resolution * getDepth()) + 50) / 100;
	        generate( pw, depth );
	    }

        /** generate VRML to a particular depth.  If depth is greater than zero,
            VRML is generated for the children of this object as well */

        public void generate( PrintWriter pw, int depth ) {

            depth--;
            if (depth < 0) {
                return;
            }

            pw.println( "Transform {" );
            if (loc != null) {
                pw.println( "  translation " + loc[0] + " " + loc[1] + " " + loc[2] );
            }
            if (orient != null) {
                pw.println( "  orientation " + orient[0] + " " + orient[1] + " " + orient[2] + " " + orient[3] );
            }
            if (scale != null) {
                pw.println( "  scale " + scale[0] + " " + scale[1] + " " + scale[2] );
            }
            pw.println( "  children [" );
            Enumeration children = getChildren();
            while (children.hasMoreElements()) {
                ComposableObject obj = (ComposableObject) children.nextElement();
                if (obj instanceof VisualObject) {
	                Generator gen = ((VisualObject) obj).getGenerator();
	                if (gen != null) {
	                    gen.generate( pw, depth );
	                }
	            }
	        }
            pw.println( "  ]" );
            pw.println( "}" );
        }
    }
}



