/*
 * @(#)ControlBox.java
 *
 * Copyright (c) 1999 by Trapezium Development LLC.  All Rights Reserved.
 *
 * The information in this file is the property of Trapezium Development LLC
 * and may be used only in accordance with the terms of the license granted
 * by Trapezium.
 *
 */
package com.trapezium.vizwiz;

import com.trapezium.browser.FontPool;
import com.trapezium.util.Glyph;
import com.trapezium.util.GlyphButton;
import com.trapezium.util.GlyphToggleButton;
import com.trapezium.util.FocuslessButton;
import java.awt.*;
import java.awt.event.*;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyVetoException;
import java.util.Vector;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;
import javax.swing.plaf.ComponentUI;
import javax.accessibility.*;

/** the base class of the various control boxes owned by the
    VizwizContollerPanel */
public class ControlBox extends JPanel {

    /** Constrained property name indicated that this frame has selected status. */
    public final static String IS_SELECTED_PROPERTY = "selected";
    /** Constrained property name indicating that the frame is minimized. */
    public final static String IS_MINIMIZED_PROPERTY = "minimized";

    public static final String MINMAX_CMD = "Minimize/Maximize";

    static final Border emptyBorder = new EmptyBorder(0,0,0,0);
    static final Insets buttonMargin = new Insets(2, 2, 2, 2);

    ContentPanel contentPane;
    ControlTitleBox titleBar;
    
    boolean minimized = false;
    boolean selected = false;

    public ControlBox(String title) {
        this(title, true);
    }
    
    public ControlBox(String title, boolean hasToggle) {
        setBorder(BorderFactory.createEtchedBorder());
        setLayout(null);
        titleBar = new ControlTitleBox(title, hasToggle);
        add(titleBar);

        contentPane = new ContentPanel();
        add(contentPane);
        invalidate();
    }
    
    public String getTitle() {
        return( titleBar.getTitle() );
    }

    public Component[] getRow(int row) {
        return contentPane.getRow(row);
    }
    
    public void setRow(int pos, Component c) {
        contentPane.setRow(pos, c);
    }
        
    public void setRow(int pos, Component cLeft, Component cMiddleRight) {
        contentPane.setRow(pos, cLeft, cMiddleRight);
    }

    public void setRow(int pos, Component cLeft, Component cMiddle, Component cRight) {
        contentPane.setRow(pos, cLeft, cMiddle, cRight);
    }

    public void addRow(Component c) {
        contentPane.addRow(c);
    }
        
    public void addRow(Component cLeft, Component cMiddleRight) {
        contentPane.addRow(cLeft, cMiddleRight);
    }

    public void addRow(Component cLeft, Component cMiddle, Component cRight) {
        contentPane.addRow(cLeft, cMiddle, cRight);
    }

    public void setTitleRow(Component c) {
        contentPane.setTitleRow(c);
    }
        
    public void setTitleRow(Component cLeft, Component cMiddleRight) {
        contentPane.setTitleRow(cLeft, cMiddleRight);
    }

    public void setTitleRow(Component cLeft, Component cMiddle, Component cRight) {
        contentPane.setTitleRow(cLeft, cMiddle, cRight);
    }
    
    public void insertRow(int pos, Component c) {
        contentPane.insertRow(pos, c);
    }
        
    public void insertRow(int pos, Component cLeft, Component cMiddleRight) {
        contentPane.insertRow(pos, cLeft, cMiddleRight);
    }

    public void insertRow(int pos, Component cLeft, Component cMiddle, Component cRight) {
        contentPane.insertRow(pos, cLeft, cMiddle, cRight);
    }
    
    public void addTool(Component c) {
        titleBar.addToToolBar(c);
    }
    
    public Container getContentPane() {
        return contentPane;
    }

    public boolean isSelected() {
        return selected;
    }


    public Dimension getPreferredSize() {
        Dimension pref = getMinimumSize();
        Dimension cppref = contentPane.getPreferredSize();
        Insets insets = getInsets();
        
        pref.width = Math.max(pref.width, cppref.width + insets.left + insets.right);
        pref.height += cppref.height;
        return pref;
    }
    

    public Dimension getMinimumSize() {
        Dimension min = titleBar.getMinimumSize();
        Insets insets = getInsets();
        min.width += insets.left + insets.right;
        min.height += insets.top + insets.bottom;
        return min;
    }
        

    /**
     * Minimizes or maximizes the box.
     *
     * @param minimize  a boolean, where true maximizes the frame and false
     *                  restores it
     * @exception PropertyVetoException when the attempt to set the
     *            property is vetoed by the receiver.
     * @beaninfo
     *     constrained: true
     *     description: Indicates whether the frame is maximized.
     */
    public void setMinimize(boolean minimize) throws PropertyVetoException {
        if (minimize != minimized) {

            Boolean oldValue = minimized ? Boolean.TRUE : Boolean.FALSE;
            Boolean newValue = minimize ? Boolean.TRUE : Boolean.FALSE;
            fireVetoableChange(IS_MINIMIZED_PROPERTY, oldValue, newValue);
            minimized = minimize;
            firePropertyChange(IS_MINIMIZED_PROPERTY, oldValue, newValue);
            
            invalidate();
            Container parent = getParent();
            while (parent != null && !(parent instanceof VizwizControllerPanel)) {
                parent = parent.getParent();
            }
            if (parent != null) {
                parent.invalidate();
                parent.validate();
            }
        }
    }
    
    public boolean isMinimized() {
        return minimized;
    }

    public void doLayout() {
	    Insets insets = getInsets();
        int width = getWidth() - insets.left - insets.right;
        int height = getHeight() - insets.top - insets.bottom;
        int x = insets.left;
        int y = insets.top;
        int tbheight = titleBar.getMinimumSize().height;
        titleBar.setBounds(x, y, width, tbheight);
        contentPane.setBounds(x, y + tbheight, width, height - tbheight);
    } 




    /** add property change listener */
    void installPropertyChangeListener(PropertyChangeListener listener) {
        addPropertyChangeListener(listener);
    }

    class ControlTitleBox extends JComponent {

        Color selectedTitleColor;
        Color selectedTextColor;
        Color notSelectedTitleColor;
        Color notSelectedTextColor;

        PropertyChangeListener propertyChangeListener;


        Action minmaxAction;
        Icon minIcon;
        Icon maxIcon;

        String title;
        JToggleButton minmaxButton;
        JToolBar toolBar;

        public ControlTitleBox(String title, boolean hasToggle) {
            this.title = title;
            setLayout(null);
            setFont( FontPool.getSerifFont(0)); //UIManager.getFont("InternalFrame.font") );
            
            
            toolBar = new JToolBar();
            toolBar.setVisible(false);
            toolBar.setFloatable(false);
            toolBar.setBorder(BorderFactory.createEmptyBorder());
            add(toolBar);

        	selectedTitleColor = UIManager.getColor("InternalFrame.activeTitleBackground");
        	selectedTextColor = UIManager.getColor("InternalFrame.activeTitleForeground");
        	notSelectedTitleColor = UIManager.getColor("InternalFrame.inactiveTitleBackground");
        	notSelectedTextColor = UIManager.getColor("InternalFrame.inactiveTitleForeground");
            
            if (hasToggle) {
                minmaxAction = new MinMaxAction();
    
                minmaxButton = new GlyphToggleButton( Glyph.DOWN, Glyph.UP );
                minmaxButton.putClientProperty("paintActive", Boolean.TRUE);
                minmaxButton.setSelected(true);
                
            /***************
                getIcons();
                minmaxButton = new FocuslessButton( minIcon );
                minmaxButton.setMargin(buttonMargin);
                minmaxButton.putClientProperty("paintActive", Boolean.TRUE);
                minmaxButton.setFocusPainted(false);
                //minmaxButton.setBorder(emptyBorder);
                minmaxButton.setOpaque(false);
            ********************/
            
                minmaxButton.addActionListener(minmaxAction);
                minmaxButton.setActionCommand(MINMAX_CMD);
                minmaxButton.getAccessibleContext().setAccessibleName(MINMAX_CMD);
                add(minmaxButton);
    
                // install listeners
                   installPropertyChangeListener(new PropertyChangeHandler());
            }
        }
        
        public String getTitle() {
            return( title );
        }
        
        public void addToToolBar(Component c) {
            toolBar.add(c);
            toolBar.setVisible(true);
        }
        
        /**
         * Notification from the UIManager that the L&F has changed.
         * Replaces the current UI object with the latest version from the
         * UIManager.
         *
         * @see JComponent#updateUI
         */
        public void updateUI() {
            //setUI((ComponentUI)UIManager.getUI(this));
        	selectedTitleColor = UIManager.getColor("InternalFrame.activeTitleBackground");
        	selectedTextColor = UIManager.getColor("InternalFrame.activeTitleForeground");
        	notSelectedTitleColor = UIManager.getColor("InternalFrame.inactiveTitleBackground");
        	notSelectedTextColor = UIManager.getColor("InternalFrame.inactiveTitleForeground");
            //getIcons();
            //setButtonIcon();
            invalidate();
        }
    
        protected void getIcons() {
            int w = 19;
            int h = 19;
            Graphics g;

            minIcon = (Icon) UIManager.get("InternalFrame.minimizeIcon");
            // unfortunate hack 
            if ("Motif".equals(UIManager.getLookAndFeel().getID())) {
                minIcon = null;
            }
            if (minIcon == null || minIcon.getIconWidth() == 0 || minIcon.getIconHeight() == 0) {
                Image minImage = createImage(w, h);
                g = minImage.getGraphics();
                g.setColor(getBackground());
                g.fillRect(0, 0, w, h);
                g.setColor(getBackground() == Color.black ? Color.white : Color.black);
                Glyph.drawMinimize(g, 6, 6, w - 12, h - 12);
                minIcon = new ImageIcon(minImage);
            }
                
            maxIcon = (Icon) UIManager.get("InternalFrame.maximizeIcon");
            // unfortunate hack 
            if ("Motif".equals(UIManager.getLookAndFeel().getID())) {
                maxIcon = null;
            } else System.out.println(UIManager.getLookAndFeel().getID() + " detected");
            if (maxIcon == null || maxIcon.getIconWidth() == 0 || maxIcon.getIconHeight() == 0) {
                Image maxImage = createImage(w, h);
                g = maxImage.getGraphics();
                g.setColor(getBackground());
                g.fillRect(0, 0, w, h);
                g.setColor(getBackground() == Color.black ? Color.white : Color.black);
                Glyph.drawMaximize(g, 6, 6, w - 12, h - 12);
                g.dispose();
                maxIcon = new ImageIcon(maxImage);
            }
        }

/*******************        
        protected void setButtonIcon() {
        	if (isMinimized()) {
        	    System.out.println("deselecting minmax");
        	    minmaxButton.setSelected(false);
            } else {
        	    System.out.println("SELECTING minmax");
        	    minmaxButton.setSelected(true);
            }
            System.out.println("minmax selected = " + minmaxButton.isSelected());
        }************/


        public void paintComponent(Graphics g)  {
        	boolean isSelected = isSelected();

            Color shadowColor;
        	if (isSelected) {
        	    g.setColor(selectedTitleColor);
        	    if (!selectedTitleColor.equals(getBackground())) {
            	    shadowColor = selectedTitleColor.darker();
            	} else {
            	    int red = Math.min(selectedTitleColor.getRed(), 192);
            	    int green = Math.min(selectedTitleColor.getBlue() + 16, 192);
            	    int blue = Math.max(selectedTitleColor.getGreen() + 16, 192);
            	    shadowColor = new Color( red, green, blue );
            	}
        	} else {
        	    g.setColor(notSelectedTitleColor);
        	    if (!notSelectedTitleColor.equals(getBackground())) {
            	    shadowColor = notSelectedTitleColor.darker();
            	} else {
            	    int red = Math.min(notSelectedTitleColor.getRed(), 192);
            	    int green = Math.min(notSelectedTitleColor.getBlue() + 16, 192);
            	    int blue = Math.max(notSelectedTitleColor.getGreen() + 16, 192);
            	    shadowColor = new Color( red, green, blue );
            	}
        	}

        	int width = getWidth();
        	int height = getHeight();
        	g.fillRect(0, 0, width, height - 1);
        	g.setColor(shadowColor);
        	g.drawLine(0, height - 1, width - 1, height - 1);

        	if (title != null) {
        	    Font f = g.getFont();
        	    g.setFont(FontPool.getSerifFont(0)); //UIManager.getFont("InternalFrame.titleFont"));
        	    if (isSelected) {
            		g.setColor(selectedTextColor);
        	    } else {
            		g.setColor(notSelectedTextColor);
                }

                    // Center text vertically.
        	    FontMetrics fm = g.getFontMetrics();
                int fmHeight = fm.getHeight() - fm.getLeading();
                int baseline = (height - fmHeight) / 2 + fm.getAscent() + fm.getLeading();
                    
        	    g.drawString(title, 4, baseline);
        	    g.setFont(f);
        	}
        }
        
        public Dimension getMinimumSize() {
            Dimension min = new Dimension(200, 0);
      	    FontMetrics fm = getFontMetrics(FontPool.getSerifFont(0));//UIManager.getFont("InternalFrame.titleFont"));
            int mmheight = 0;
            if ( minmaxButton != null ) {
                mmheight = minmaxButton.getPreferredSize().height;
            }
            min.height +=  Math.max(mmheight, fm.getHeight());
            return min;
        }


        public Dimension getPreferredSize() {
            return getMinimumSize();
        }


        public class PropertyChangeHandler implements PropertyChangeListener {
            public void propertyChange(PropertyChangeEvent evt) {

           	    String prop = (String) evt.getPropertyName();

        	    if (ControlBox.IS_SELECTED_PROPERTY.equals(prop)) {
        	        repaint();
            		return;
        	    } 

        	    if (ControlBox.IS_MINIMIZED_PROPERTY.equals(prop)) {
            		//setButtonIcon();
            		return;
    	        } 
        	}
        }

        public class MinMaxAction extends AbstractAction {
            public MinMaxAction() {
        	    super(MINMAX_CMD);
            }

            public void actionPerformed(ActionEvent e) {
       		    try {
       		        setMinimize(!isMinimized());
       		    } catch (PropertyVetoException pve) {
       		    }
    	    }
    	}      


        public void doLayout() {
            int x = getWidth();
            int y = getHeight();
            if ( minmaxButton != null ) {
                Dimension buttonpref = minmaxButton.getPreferredSize();
                int buttonWidth = buttonpref.width;
                int buttonHeight = buttonpref.height;
            
                x = getWidth() - buttonWidth - 2;
                y = (getHeight() - buttonHeight) / 2;
                minmaxButton.setBounds(x, y, buttonWidth, buttonHeight);
            } 
   
            Dimension tbpref = toolBar.getPreferredSize();
    	    x -= tbpref.width + 4;
    	    
    	    int tbHeight = Math.min(tbpref.height, getHeight() - 4);
            y = (getHeight() - tbHeight) / 2;
    	    
    	    toolBar.setBounds(x, y, tbpref.width, tbHeight);
        } 
    }


    private static int vmargin = 4;
    private static int hmargin = 2;
    private static int vgap = 2;
    
    private class ContentPanel extends JPanel {

        boolean varRowHeight = true;

        Component[] titleRow;
        Vector rows;
        int leftWidth = 0;
        int rightWidth = 0;
        int centerMin = 0;
        int titleHeight = 0;
        
        public ContentPanel() {
            super();

            titleRow = null;
            rows = new Vector();
            setLayout(null);
        }
        
        public Component[] getRow(int row) {
            return (Component[]) rows.elementAt(row);
        }
        
        public void setRow(int pos, Component c) {
            Component[] oldrow = (Component[]) rows.elementAt(pos);
            for (int i = 0; i < oldrow.length; i++) {
                super.remove(oldrow[i]);
            }
            super.add(c);
            Component[] rowParts = new Component[1];
            rowParts[0] = c;
            rows.setElementAt(rowParts, pos);
        }
        
        public void setRow(int pos, Component cLeft, Component cMiddleRight) {
            Component[] oldrow = (Component[]) rows.elementAt(pos);
            for (int i = 0; i < oldrow.length; i++) {
                super.remove(oldrow[i]);
            }
            super.add(cLeft);
            super.add(cMiddleRight);
            Component[] rowParts = new Component[2];
            rowParts[0] = cLeft;
            rowParts[1] = cMiddleRight;
            rows.setElementAt(rowParts, pos);
        }

        public void setRow(int pos, Component cLeft, Component cMiddle, Component cRight) {
            Component[] oldrow = (Component[]) rows.elementAt(pos);
            for (int i = 0; i < oldrow.length; i++) {
                super.remove(oldrow[i]);
            }
            super.add(cLeft);
            super.add(cMiddle);
            super.add(cRight);
            Component[] rowParts = new Component[3];
            rowParts[0] = cLeft;
            rowParts[1] = cMiddle;
            rowParts[2] = cRight;
            rows.setElementAt(rowParts, pos);
        }

        public void addRow(Component c) {
            super.add(c);
            Component[] rowParts = new Component[1];
            rowParts[0] = c;
            rows.addElement(rowParts);
        }
        
        public void addRow(Component cLeft, Component cMiddleRight) {
            super.add(cLeft);
            super.add(cMiddleRight);
            Component[] rowParts = new Component[2];
            rowParts[0] = cLeft;
            rowParts[1] = cMiddleRight;
            rows.addElement(rowParts);
        }

        public void addRow(Component cLeft, Component cMiddle, Component cRight) {
            super.add(cLeft);
            super.add(cMiddle);
            super.add(cRight);
            Component[] rowParts = new Component[3];
            rowParts[0] = cLeft;
            rowParts[1] = cMiddle;
            rowParts[2] = cRight;
            rows.addElement(rowParts);
        }


        public void insertRow(int pos, Component c) {
            super.add(c);
            Component[] rowParts = new Component[1];
            rowParts[0] = c;
            rows.insertElementAt(rowParts, pos);
        }
        
        public void insertRow(int pos, Component cLeft, Component cMiddleRight) {
            super.add(cLeft);
            super.add(cMiddleRight);
            Component[] rowParts = new Component[2];
            rowParts[0] = cLeft;
            rowParts[1] = cMiddleRight;
            rows.insertElementAt(rowParts, pos);
        }

        public void insertRow(int pos, Component cLeft, Component cMiddle, Component cRight) {
            super.add(cLeft);
            super.add(cMiddle);
            super.add(cRight);
            Component[] rowParts = new Component[3];
            rowParts[0] = cLeft;
            rowParts[1] = cMiddle;
            rowParts[2] = cRight;
            rows.insertElementAt(rowParts, pos);
        }

        public void setTitleRow(Component c) {
            super.add(c);
            titleRow = new Component[1];
            titleRow[0] = c;
        }
        
        public void setTitleRow(Component cLeft, Component cMiddleRight) {
            super.add(cLeft);
            super.add(cMiddleRight);
            titleRow = new Component[2];
            titleRow[0] = cLeft;
            titleRow[1] = cMiddleRight;
        }

        public void setTitleRow(Component cLeft, Component cMiddle, Component cRight) {
            super.add(cLeft);
            super.add(cMiddle);
            super.add(cRight);
            titleRow = new Component[3];
            titleRow[0] = cLeft;
            titleRow[1] = cMiddle;
            titleRow[2] = cRight;
        }

        public Dimension getPreferredSize() {

            Dimension pref = new Dimension(2 * hmargin, 0);
            
            if (isMinimized()) {
                return pref;
            }
            
            pref.height = 2 * vmargin;
            
            int numrows = rows.size();
            
            leftWidth = 0;
            rightWidth = 0;
            int centerWidth = 0;
            int totHeight = 0;
            for (int i = 0; i < numrows; i++) {
                Component[] row = (Component[]) rows.elementAt(i);
                int numparts = row.length;
                if (numparts > 0) {
                    Dimension partpref = row[0].getPreferredSize();
                    leftWidth = Math.max(partpref.width, leftWidth);
                    int rowHeight = partpref.height;
                    if (numparts > 1) {
                        partpref = row[1].getPreferredSize();
                        rowHeight = Math.max(partpref.height, rowHeight);
                        // presume all components have fonts of the same size and only calculate once
                        if (centerWidth == 0) {
                            FontMetrics fm = getFontMetrics(getFont());
                            //centerWidth = 32 * fm.charWidth('M');
                            centerWidth = 10 * fm.charWidth('M');
                            centerMin = centerWidth / 4;
                        }
                        
                        if (numparts > 2) {
                            partpref = row[2].getPreferredSize();
                            rightWidth = Math.max(partpref.width, rightWidth);
                            rowHeight = Math.max(partpref.height, rowHeight);
                        }
                    }
                    totHeight += rowHeight;
                }
            }
            pref.width += leftWidth + centerWidth + rightWidth;
            pref.height += totHeight + (numrows > 0 ? (numrows - 1) * vgap : 0);

            // add the title row height if present
            titleHeight = 0;
            if (titleRow != null) {
                for (int n = 0; n < titleRow.length; n++) {
                    titleHeight = Math.max(titleRow[n].getPreferredSize().height, titleHeight);
                }
            }
            pref.height += titleHeight;

            Insets insets = getInsets();
            pref.width += insets.left + insets.right;
            pref.height += insets.top + insets.bottom;
            
            return pref;
        }
        

        public Dimension getMinimumSize() {
            Dimension min = new Dimension(2 * hmargin, 0);
            if (isMinimized()) {
                return min;
            }
            
            min.height = 2 * vmargin;
            
            int numrows = rows.size();
            
            int leftMin = 0;
            int rightMin = 0;
            int totHeight = 0;
            for (int i = 0; i < numrows; i++) {
                Component[] row = (Component[]) rows.elementAt(i);
                int numparts = row.length;
                if (numparts > 0) {
                    Dimension partmin = row[0].getMinimumSize();
                    leftMin = Math.max(partmin.width, leftWidth);
                    int rowHeight = partmin.height;
                    if (numparts > 1) {
                        partmin = row[1].getMinimumSize();
                        rowHeight = Math.max(partmin.height, rowHeight);
                        // presume all components have fonts of the same size and only calculate once
                        if (centerMin == 0) {
                            FontMetrics fm = getFontMetrics(getFont());
                            centerMin = 8 * fm.charWidth('M');
                        }
                        
                        if (numparts > 2) {
                            partmin = row[2].getMinimumSize();
                            rightMin = Math.max(partmin.width, rightWidth);
                            rowHeight = Math.max(partmin.height, rowHeight);
                        }
                    }
                    totHeight += rowHeight;
                }
            }

            min.width += leftMin + centerMin + rightMin;
            min.height += totHeight + (numrows > 0 ? (numrows - 1) * vgap : 0);;

            // add the title row height if present
            int th = 0;
            if (titleRow != null) {
                for (int n = 0; n < titleRow.length; n++) {
                    th = Math.max(th, titleRow[n].getMinimumSize().height);
                }
            }
            min.height += th;

            Insets insets = getInsets();
            min.width += insets.left + insets.right;
            min.height += insets.top + insets.bottom;
            
            return min;
        }
        
        public void doLayout() {

            int numrows = rows.size();
            if (numrows == 0) {
                return;
            }

            // make sure leftWidth has been calculated
            if (leftWidth == 0) {
                getPreferredSize();
            }
            
            Insets insets = getInsets();
            int width = getWidth() - insets.left - insets.right;
            int y = insets.top + vmargin;
            int height = getHeight() - insets.top - insets.bottom - titleHeight;
            int centerWidth = Math.max(width - leftWidth - rightWidth - 2 * hmargin, centerMin);

            if (titleRow != null && titleRow.length > 0) {
                if (titleRow.length == 1) {
                    titleRow[0].setBounds(hmargin, y, width - 2 * hmargin, titleHeight);

                } else if (titleRow.length == 2) {
                    titleRow[0].setBounds(hmargin, y, leftWidth, titleHeight);
                    titleRow[1].setBounds(hmargin + leftWidth, y, width - leftWidth - 2 * hmargin, titleHeight);
                    
                } else {
                    titleRow[0].setBounds(hmargin, y, leftWidth, titleHeight);
                    titleRow[1].setBounds(hmargin + leftWidth, y, centerWidth, titleHeight);
                    titleRow[2].setBounds(hmargin + leftWidth + centerWidth, y, width - leftWidth - centerWidth - 2 * hmargin, titleHeight);
                }
            }                
            y += titleHeight;
            
            for (int i = 0; i < numrows; i++) {
                Component[] row = (Component[]) rows.elementAt(i);
                int numparts = row.length;
                int rowHeight = 0;
                if (numparts == 1) {
                    rowHeight = row[0].getPreferredSize().height;
                    row[0].setBounds(hmargin, y, width - 2 * hmargin, rowHeight);
                    

                } else if (numparts == 2) {
                    rowHeight = Math.max(row[0].getPreferredSize().height, row[1].getPreferredSize().height);
                    row[0].setBounds(hmargin, y, leftWidth, rowHeight);
                    row[1].setBounds(hmargin + leftWidth, y, width - leftWidth - 2 * hmargin, rowHeight);
                    
                } else if (numparts == 3) {
                    rowHeight = Math.max(row[0].getPreferredSize().height, Math.max(row[1].getPreferredSize().height, row[2].getPreferredSize().height));
                    
                    row[0].setBounds(hmargin, y, leftWidth, rowHeight);
                    row[1].setBounds(hmargin + leftWidth, y, centerWidth, rowHeight);
                    row[2].setBounds(hmargin + leftWidth + centerWidth, y, width - leftWidth - centerWidth - 2 * hmargin, rowHeight);
                }

                y += rowHeight + vgap;
            }
        } 
    }
}


