/*
 * @(#)UpdateObject.java
 *
 * Copyright (c) 1999 by Trapezium Development LLC.  All Rights Reserved.
 *
 * The information in this file is the property of Trapezium Development LLC
 * and may be used only in accordance with the terms of the license granted
 * by com.trapezium.
 *
 */

package com.trapezium.vizwiz;

import java.lang.reflect.*;

/** object for passing around data updates in vizwiz */
abstract public class UpdateObject {
    
    
    static public UpdateObject createUpdateObject(Object obj) {
        if (obj.getClass().isArray()) {
            Object item = Array.get(obj, 0);
            Class c = item.getClass();
            if (c == Integer.TYPE) {
                return new PointUpdate((int[]) obj);
            } else if (c == Float.TYPE) {
                return new PointUpdate((float[]) obj);
            } else if (c.isArray()) {
                item = Array.get(item, 0);
                c = item.getClass();
                if (c == Integer.TYPE) {
                    return new MultiPointUpdate((int[][]) obj);
                } else if (c == Float.TYPE) {
                    return new MultiPointUpdate((float[][]) obj);
                }
            }
        }

        throw new IllegalArgumentException("Unsupported type for UpdateObject");
    }
                

    static public UpdateObject createUpdateObject(int[] obj) {
        // stream of triplets -- x, y, val
        return new PointUpdate(obj);        
    }

    static public UpdateObject createUpdateObject(int[][] obj) {
        return new MultiPointUpdate(obj);        
    }

    static public UpdateObject createUpdateObject(float[] obj) {
        // stream of triplets -- x, y, val
        return new PointUpdate(obj);        
    }

    static public UpdateObject createUpdateObject(float[][] obj) {
        return new MultiPointUpdate(obj);        
    }

    protected UpdateObject() {
    }

}

class PropertyUpdate extends UpdateObject {
    String name;
    String value;
    
    /* a PropertyUpdate consists of a property name and a value.  Both
       are represented as Strings */
    public PropertyUpdate(String name, String value) {
        this.name = name;
        this.value = value;
    }

    public void update(DataModel model) {
        model.setProperty(name, value);
    }
}
    

class PointUpdate extends UpdateObject {

    float x;
    float y;
    float value;

    /* a PointUpdate consists of a 2D point and a value.  In 3D
       the value is usually represented vertically, which means
       the y direction.  So the x, y values will be converted to
       x, z values. */
    public PointUpdate(float x, float y, float value) {
        this.x = x;
        this.y = y;
        this.value = value;
    }

    public PointUpdate(float[] val) {
        try {
            this.x = val[0];
            this.y = val[1];
            this.value = val[2];
        } catch (ArrayIndexOutOfBoundsException e) {
            throw new IllegalArgumentException("array too short for PointUpdate");
        }
    }

    public PointUpdate(int[] val) {
        try {
            this.x = (float) val[0];
            this.y = (float) val[1];
            this.value = (float) val[2];
        } catch (ArrayIndexOutOfBoundsException e) {
            throw new IllegalArgumentException("array too short for PointUpdate");
        }
    }

    public void update(PointModel model) {
        model.addPoint(x, y, value);
    }
}

class MultiPointUpdate extends UpdateObject {
    PointUpdate[] pts;

    public MultiPointUpdate(float[][] vals) {
        int len = vals.length;
        pts = new PointUpdate[len];
        for (int i = 0; i < len; i++) {
            pts[i] = new PointUpdate(vals[i]);
        }
    }

    public MultiPointUpdate(int[][] vals) {
        int len = vals.length;
        pts = new PointUpdate[len];
        for (int i = 0; i < len; i++) {
            pts[i] = new PointUpdate(vals[i]);
        }
    }

    public void update(PointModel model) {
        for (int i = 0; i < pts.length; i++) {
            model.addPoint(pts[i].x, pts[i].y, pts[i].value);
        }
    }

}



