package com.trapezium.vizwiz.command;

import java.io.InputStream;
import java.io.ObjectInputStream;

public class CommandReader {
    static public final int PLAYBACK_CHANNEL_INDICATOR = 0x1000;
    
    // either Command.BYTE_FORMAT or Command.INT_FORMAT
    int format;
    
    // input read from an InputStream or an ObjectInputStream
    InputStream inStream;
    ObjectInputStream objectInputStream;
   
    // kind of weird optimization, preset sized fixed arrays for handling
    // different sized int array commands, otherwise we get into all the
    // overhead of creating/destroying objects
    static final int MAX_INTS = 30;
    int[][] arrayOfArrays = new int[ MAX_INTS + 1 ][];
 
    /** Class constructor
     *
     *  @param inStream input stream that is the source of the commands
     */   
    public CommandReader( InputStream inStream, int format ) throws java.io.StreamCorruptedException, java.io.IOException {
        if ( format == Command.INT_FORMAT ) {
            this.inStream = null;
            objectInputStream = new ObjectInputStream( inStream );
        } else {
            this.inStream = inStream;
            objectInputStream = null;
        }
        this.format = format;
        for ( int i = 1; i <= MAX_INTS; i++ ) {
            arrayOfArrays[ i ] = new int[i];
        }
    }
 
    
    /** Read the next command from the InputStream.  
     *
     *  NOTE: only works for Command.INT_FORMAT data streams.  Command.BYTE_FORMAT 
     *        data streams are not supported.
     *
     *  @return an int array representing command.  Offsets are somewhat hardcoded:
     *     offset 0:  channel number
     *     offset 1:  command
     *     remaining offsets:  command data, varies depending on command
     */
     
    public int[] readCommand() {
        try {
            if (( objectInputStream != null ) && ( format == Command.INT_FORMAT )) {
                // read first int to see type of command
                int commandSize = objectInputStream.readInt();
                int[] result = arrayOfArrays[ commandSize ];
                for ( int i = 0; i < commandSize; i++ ) {
                    result[i] = objectInputStream.readInt();
                }
                result[0] |= PLAYBACK_CHANNEL_INDICATOR;
                return( result );
            }
        } catch ( Exception e ) {
            return( null );
        }
        return( null );
    }
    
    
    /** Close any open streams.
     */
    public void close() {
        try {
            if ( objectInputStream != null ) {
                objectInputStream.close();
                objectInputStream = null;
            }
        } catch ( Exception e ) {
            System.out.println( "Exception closing CommandReader objectInputStream: " + e );
            objectInputStream = null;
        }
    }
    
    
    /** read a command of known size from the InputStream.
     *  This type of read only supports Command.BYTE_FORMAT
     */
    final int BUF_SIZE = 1024;
    byte[] buffer = new byte[BUF_SIZE];
    int[] ubuffer = new int[4];
    public int[] readCommand( int size ) throws java.io.IOException {
        if (( inStream != null ) && ( format == Command.BYTE_FORMAT )) {
            int readCount = inStream.read( buffer, 0, size );
            while ( readCount != size ) {
                if ( readCount < 0 ) {
                    return( null );
                }
                int temp = inStream.read( buffer, readCount, size - readCount );
                if ( temp < 0 ) {
                    return( null );
                }
                readCount += temp;
            }
            if (readCount >= 2) {
                // have to add channel support back into neattools
                // currently always the same channel
                int channel = buffer[1];
                int numInts = (readCount+1)/4; //(readCount - 2)/4;
                int idx = 2;
                int[] val;
                if (( numInts > 0 ) && ( numInts <= MAX_INTS )) {
                    val = arrayOfArrays[ numInts + 1 ];
                } else {
                    System.out.println( "Unsupported int count! " + numInts );
                    return( null );
                }
                val[0] = channel;
                for ( int i = 1; i <= numInts; i++ ) {
                    ubuffer[0] = buffer[idx++];
                    ubuffer[1] = buffer[idx++];
                    ubuffer[2] = buffer[idx++];
                    ubuffer[3] = buffer[idx++];
                    if ( ubuffer[0] < 0 ) ubuffer[0] += 256;
                    if ( ubuffer[1] < 0 ) ubuffer[1] += 256;
                    if ( ubuffer[2] < 0 ) ubuffer[2] += 256;
                    if ( ubuffer[3] < 0 ) ubuffer[3] += 256;
                    val[i] = ubuffer[0] + (ubuffer[1] << 8) + (ubuffer[2] << 16) + (ubuffer[3] << 24);
                }
                return( val );
            }
            return( null );
        } else {
            return( null );
        }
    }
}