package com.trapezium.vizwiz.command;

import java.util.Vector;
import java.io.FileInputStream;
import java.io.FileOutputStream;

/**
 *  The CommandStreamProcessor is the first pass at taking a recorded stream
 *  and transforming it into a playback stream with pause markers.  Really it
 *  is a first pass at an on-the-fly modification of a playback stream, but
 *  this pass does it as a command line file processing utility.
 */

public class CommandStreamProcessor {
    
    Vector filesToProcess;
    boolean split;
    boolean pause;
    int pauseDuration;
    int pauseDistance;
    boolean jump;
    boolean collide;
    Vector collideInfo;
    boolean timeout;
    int timeoutDuration;
    boolean skip;
    int skipType;
    int skipCount;
    boolean dump;
    
    int nextSetting;
    static final int SET_PAUSE_DURATION = 1;
    static final int SET_PAUSE_DISTANCE = 2;
    static final int SET_COLLIDE_O1 = 3;
    static final int SET_COLLIDE_O2 = 4;
    static final int SET_COLLIDE_DISTANCE = 5;
    static final int SET_TIMEOUT_DURATION = 6;
    static final int SET_SKIP_TYPE = 7;
    static final int SET_SKIP_COUNT = 8;
    
    class CollideInfo {
        public int o1;
        public int o2;
        public int distance;
    }
    
    CollideInfo collideInstance;
          
    
    public CommandStreamProcessor() {
        split = false;
        pause = false;
        jump = false;
        collide = false;
        timeout = false;
        nextSetting = 0;
        collideInstance = new CollideInfo();
    }
    
    public void setupArg( String arg ) {
        if ( arg.charAt( 0 ) == '-' ) {
            setupFlag( arg );
        } else if ( nextSetting > 0 ) {
            setupSetting( arg );
        } else {
            addFileToProcess( arg );
        }
    }
    
    void addFileToProcess( String fileName ) {
        if ( filesToProcess == null ) {
            filesToProcess = new Vector();
        }
        filesToProcess.addElement( fileName );
    }
    
    void usage() {
        System.out.println( "java com.trapezium.vizwiz.command.CommandStreamProcessor [options]" );
        System.out.println( "-split" );
        System.out.println( "    Splits recorded data into per-object streams" );
        System.out.println( "-pause <duration> <distance>" );
        System.out.println( "    Injects pause commands into stream whenever a set of " );
        System.out.println( "    locations are within 'distance' for 'duration' milliseconds'" );
        System.out.println( "-jump" );
        System.out.println( "    Removes time markers from stream, playback is then jumpy" );
        System.out.println( "    or so fast it seems instant" );
        System.out.println( "-collide <o1> <o2> <distance>" );
        System.out.println( "    Adds command used by playback, that registers playback" );
        System.out.println( "    with viz to be notified of collisions between two objects" );
        System.out.println( "-skip <command> <count>" );
        System.out.println( "    Reduces playback file size by skipping 'count' commands" );
        System.out.println( "    of a specific type." );
        System.out.println( "-dump" );
        System.out.println("     Dumps out each command as encountered" );
    }
    
    void setupFlag( String flag ) {
        if ( flag.compareTo( "-split" ) == 0 ) {
            split = true;
        } else if ( flag.compareTo( "-pause" ) == 0 ) {
            pause = true;
            nextSetting = SET_PAUSE_DURATION;
        } else if ( flag.compareTo( "-jump" ) == 0 ) {
            jump = true;
        } else if ( flag.compareTo( "-collide" ) == 0 ) {
            collide = true;
            nextSetting = SET_COLLIDE_O1;
        } else if ( flag.compareTo( "-timeout" ) == 0 ) {
            timeout = true;
            nextSetting = SET_TIMEOUT_DURATION;
        } else if ( flag.compareTo( "-skip" ) == 0 ) {
            skip = true;
            nextSetting = SET_SKIP_TYPE;
        } else if ( flag.compareTo( "-dump" ) == 0 ) {
            dump = true;
        } else if ( flag.compareTo( "-help" ) == 0 ) {
            usage();
            System.exit( 0 );
        }
    }
    
    void setupSetting( String value ) {
        if ( nextSetting == SET_PAUSE_DURATION ) {
            pauseDuration = Integer.parseInt( value );
            nextSetting = SET_PAUSE_DISTANCE;
        } else if ( nextSetting == SET_PAUSE_DISTANCE ) {
            pauseDistance = Integer.parseInt( value );
            nextSetting = 0;
        } else if ( nextSetting == SET_COLLIDE_O1 ) {
            collideInstance.o1 = Integer.parseInt( value );
            nextSetting = SET_COLLIDE_O2;
        } else if ( nextSetting == SET_COLLIDE_O2 ) {
            collideInstance.o2 = Integer.parseInt( value );
            nextSetting = SET_COLLIDE_DISTANCE;
        } else if ( nextSetting == SET_COLLIDE_DISTANCE ) {
            collideInstance.distance = Integer.parseInt( value );
            if ( collideInfo == null ) {
                collideInfo = new Vector();
            }
            collideInfo.addElement( collideInstance );
            collideInstance = new CollideInfo();
            nextSetting = 0;
        } else if ( nextSetting == SET_TIMEOUT_DURATION ) {
            timeoutDuration = Integer.parseInt( value );
            nextSetting = 0;
        } else if ( nextSetting == SET_SKIP_TYPE ) {
            skipType = Integer.parseInt( value );
            nextSetting = SET_SKIP_COUNT;
        } else if ( nextSetting == SET_SKIP_COUNT ) {
            skipCount = Integer.parseInt( value );
            nextSetting = 0;
        }
    }
    
    public void processFiles() {
        for ( int i = 0; i < filesToProcess.size(); i++ ) {
            String fileName = (String)filesToProcess.elementAt( i );
            try {
                processFile( fileName );
            } catch ( Exception e ) {
                System.out.println( "Exception processing file: " + fileName );
                System.out.println( ".." + e );
                e.printStackTrace();
            }
        }
    }
    
    
    void processFile( String fileName ) throws java.io.StreamCorruptedException, java.io.IOException {
        
        System.out.println( "Processing file: " + fileName );
        
        CommandReader cr = new CommandReader( new FileInputStream( fileName ), Command.INT_FORMAT );
        
        CommandWriter cw;
        
        if ( split ) {
            cw = new SplitWriter( fileName );
        } else {
            cw = new CommandWriter( new FileOutputStream( fileName + ".out" ), Command.INT_FORMAT );
        }
        
        if ( pause ) {
            cw = new PauseInjector( cw );
        }
        
        if ( jump ) {
            cw = new SkipWriter( cw, Command.TIME_MARKER );
        }
        
        if ( collide ) {
            for ( int i = 0; i < collideInfo.size(); i++ ) {
                CollideInfo ci = (CollideInfo)collideInfo.elementAt( i );
                int[] collisionData = new int[4];
                collisionData[0] = Command.REGISTER_COLLISION;
                collisionData[1] = ci.o1;
                collisionData[2] = ci.o2;
                collisionData[3] = ci.distance;
                cw = new FirstTimeWriter( cw, collisionData );
            }
        }
        
        if ( timeout ) {
            int[] toData = new int[2];
            toData[0] = Command.REGISTER_TIMEOUT;
            toData[1] = timeoutDuration;
            cw = new FirstTimeWriter( cw, toData );
        }
        
        if ( skip ) {
            cw = new SkipWriter( cw, skipType, skipCount );
        }
                
        System.out.println( "main loop" );
        int lineNo = 1;
        while ( true ) {
            int[] command = cr.readCommand();
            if ( command == null ) {
                if ( dump ) {
                    System.out.println( "null command" );
                }
                break;
            }
            if ( dump ) {
                System.out.println( lineNo + ": " + Command.toString( command ));
                lineNo++;
            }
            cw.writeCommand( command );
        }
        
        cw.close();
        System.out.println( "done" );
    }
        
    
    static public void main( String[] args ) {
        CommandStreamProcessor csp = new CommandStreamProcessor();
        for ( int i = 0; i < args.length; i++ ) {
            csp.setupArg( args[i] );
        }
        csp.processFiles();
    }    
}