/*
 * tngpdb.java
 *      Extracts TNG data from a pdb file
 *
 *	Usage: java tngpdb foo.pdb bar
 *     based on pdb2files.java by Ken Shirriff  ken.shirriff@eng.sun.com
 *
 * Joh Johannsen         johj@pacbell.net
 *
 * Copyright (c) 2001 MindTel, LLC. All Rights Reserved.
 *
 */
package mindtel.pdb.parse;

import java.io.*;
import java.util.*;

public class tngpdb {
    
    Vector recordList;
    short numRecords;
    boolean fileExists;
    Date created;
    String dataName;

    /**
     convert a 4 byte PDB date into a Java Date object
    
     PDB files base their date on January 1, 1904
     Java Date objects are based on January 1, 1970
     
     This method converts pdb date bytes into an integer, doing some weird
     stuff because Java does not support unsigned numbers.
     */
   static Date calcDate( byte[] pdbDate ) {
       int unsignedDate[] = new int[4];
       for ( int i = 0; i < 4; i++ ) {
           unsignedDate[i] = pdbDate[i];
           if ( unsignedDate[i] < 0 ) {
               unsignedDate[i] = unsignedDate[i] + 256;
           }
       }
       unsignedDate[0] -= 124;  // highest byte value diff of 1970 - 1904 (to prevent sign problems in following line)
       long val = (unsignedDate[0]<<24) + (unsignedDate[1]<<16) +(unsignedDate[2]<<8) +unsignedDate[3];
       val -= 2470016;  // remaining bytes diff of 1970 - 1904
       return( new Date(val*1000) );
   }
   
   /** Get the name from a byte array that may have garbage after the null terminator
    */
   static String calcName( byte[] name ) {
       StringBuffer sb = new StringBuffer();
       for ( int i = 0; i < name.length; i++ ) {
           if ( name[i] == 0 ) {
               break;
           }
           sb.append((char)name[i]);
       }
       return( new String( sb ));
   }
   
   public static void main(String args[]) {
       if ( args.length == 0 ) {
           System.out.println("Usage: java tngpdb <filename> <outfile>");
           System.out.println("   or: java tngpdb <filename> -html");
           System.out.println("");
           System.out.println("The [outfile] parameter creates comma separated file.");
           System.out.println("The -html parameter generates an HTML view of the PDB data.");
           System.exit(0);
       } else {
           
           tngpdb tngData = new tngpdb( args[0] );
           if ( !tngData.exists() ) {
               System.out.println("File not found.");
               System.exit(0);
           } 
           if ( args.length == 2 ) {
               if ( args[1].compareTo("-html") == 0 ) {
                   tngData.genHTML();
               } else {
                   tngData.dump( args[1] );
               }
           } else {
               tngData.dump2stream( new DataOutputStream( System.out ));
           }
       }
   }

   public tngpdb( String fileName ) {
       recordList = new Vector();
       fileExists = false;
   
       try {
           File f = new File(fileName);
           DataInputStream is = new DataInputStream(new FileInputStream(f));
           
           // read pdb header 
           byte name[] = new byte[32];
           is.read(name);
           short attributes = is.readShort();
           short version = is.readShort();
           byte creationDate[] = new byte[4];
           is.read(creationDate);
           byte modificationDate[] = new byte[4];
           is.read(modificationDate);
           int lastBackupDate = is.readInt();
           int modificationNumber = is.readInt();
           int appInfoID = is.readInt();
           int sortInfoID = is.readInt();
           byte type[] = new byte[4];
           is.read(type);
           byte creator[] = new byte[4];
           is.read(creator);
           int uniqueIDSeed = is.readInt();
           created = calcDate(creationDate);
           dataName = calcName(name);
           
           /*
           System.out.println("name "+dataName);
           System.out.println("version "+version);
           System.out.println("attributes "+attributes);
           System.out.println("type "+(new String(type)));
           System.out.println("creator "+(new String(creator)));
           System.out.println("creationDate "+created.toString());
           System.out.println("modificationDate "+calcDate(modificationDate));
           System.out.println("appInfoID "+appInfoID);
           */
           

           // RecordListType
           int nextRecordListID = is.readInt();
           numRecords = is.readShort();
           //System.out.println("numRecords "+numRecords);
           //System.out.println("nextRecordListID "+nextRecordListID);
           
           // read all the records
           for (int i=0; i<numRecords; i++) {
               readRecord(is);
           }
           is.readShort(); // pad
           
           // now read the record data
           int offset = 80+numRecords*8;
           int totNumRecords = numRecords;
           //System.out.println("totNumRecords "+totNumRecords);
           for (int i=0; i<totNumRecords; i++) {
               int end;
               Record r = (Record)recordList.elementAt(i);
               int start = r.localChunkID;
               if (i == totNumRecords-1) {
                   end = (int)f.length();
               } else {
                   end = ((Record)recordList.elementAt(i+1)).localChunkID;
               }
               int read = is.read(r.setData(end-start),0,end-start);
               //System.out.println("read " +read + " bytes");
           }
           fileExists = true;
       } catch (IOException e) {
           System.out.println("IOException "+e);
       }

   }
   
   public boolean exists() {
       return( fileExists );
   }
   
   void readRecord(DataInputStream is) throws IOException {
       Record r = new Record();
       recordList.addElement(r);
       r.localChunkID = is.readInt();
       r.attributes = is.readByte();
       is.read(r.uniqueID);
   }
  
   public void dump( String fileName ) {
       DataOutputStream os = null;
       try {
           os = new DataOutputStream(new FileOutputStream(fileName));
       } catch (IOException e) {
           System.out.println("Failed opening output file " + fileName);
           System.exit(0);
       }
       dump2stream( os );
   }
   
   public void dump2stream( DataOutputStream os ) {
       int totNumRecords = numRecords;
       //System.out.println("totNumRecords "+totNumRecords);
       try {
           for (int i=0; i<totNumRecords; i++) {
               Record r = (Record)recordList.elementAt(i);
               os.writeBytes(Integer.toString(i)+",");
               os.writeBytes(r.getData());
           }
       } catch (IOException e) {
           System.out.println("IOException "+e);
       }
   }

   public int getNumberRecords() {
        return( numRecords );
   }

   public int getTime( int recordNumber ) {
       Record r = (Record)recordList.elementAt( recordNumber );
       return( r.getIntTime() );
   }

   public int getNumberChannels( int recordNumber ) {
       Record r = (Record)recordList.elementAt( recordNumber );
       return( r.getNumberChannels() );
   }

   public int getDataValue( int recordNumber, int dataOffset ) {
       Record r = (Record)recordList.elementAt( recordNumber );
       return( r.getDataValue( dataOffset ));
   }

   public void genHTML() {
       Record r = (Record)recordList.elementAt(0);
       System.out.println("<HTML><HEAD><TITLE>" + dataName + "</TITLE></HEAD>");
       System.out.println("<BODY>");
       System.out.println("<h2>"+dataName+", "+created.toString()+"</h2>");
       System.out.println("<table border=1><tr>");
       System.out.println("<th>#</th><th>Time</th>");
       int nchannels = r.getNumberChannels();
       for ( int i = 0; i < nchannels; i++ ) {
           System.out.println("<th>ch"+(i+1)+"</th>");
       }
       System.out.println("</tr>");
       for ( int i = 0; i < numRecords; i++ ) {
           r = (Record)recordList.elementAt(i);
           System.out.println("<tr><td>"+(i+1)+"</td><td>"+r.getTime()+"</td>");
           for ( int j = 0; j < nchannels; j++ ) {
               System.out.println("<td>"+r.getChannel(j)+"</td>");
           }
           System.out.println("</tr>");
       }
       System.out.println("</table></BODY></HTML>");
   }
}

class Record {
    int localChunkID;
    byte attributes;
    byte uniqueID[];
    byte data[];

    Record() {
      uniqueID = new byte[3];
    }
    
    byte[] setData( int length ) {
        data = new byte[ length ];
        return( data );
    }
    
    int convbyte(byte b ) {
        int result = b;
        if ( result < 0 ) {
            result = result + 256;
        }
        return( result );
    }
    
    String getValue( int offset ) {
        int d1 = convbyte(data[offset]);
        int d2 = convbyte(data[offset+1]);
        return( Integer.toString((d1<<8) + d2 ));
    }
    
    int getIntValue( int offset ) {
        int d1 = convbyte(data[offset]);
        int d2 = convbyte(data[offset+1]);
        return( (d1<<8) + d2 );
    }
    
    int getIntTime() {
        return getIntValue(2);
    }
    
    String getTime() {
        return getValue(2);
    }
    
    int getDataValue( int channel ) {
        return( getIntValue( 4 + channel*2 ));
    }

    String getChannel( int channel ) {
        return getValue(4+channel*2);
    }
    
    void appendValues( StringBuffer result, int offset, int count ) {
        for ( int i = 0; i < count; i++ ) {
            result.append( getValue( offset ));
            if ( i == (count - 1)) {
                result.append("\n");
            } else {
                result.append(",");
            }
            offset = offset + 2;
        }
    }
    
    int getNumberChannels() {
        int l = data.length;
        if ( l == 15 ) {
            return( 2 );
        } else if ( l == 37 ) {
            return( 4 );
        } else if ( l == 105 ) {
            return( 8 );
        } else {
            System.out.println("data length is "+l);
            return( 0 );
        }
    }

    String getData() {
        StringBuffer result = new StringBuffer();
        appendValues( result, 2, getNumberChannels() + 1 );
        return( new String(result));
  /*          
        int dl[] = new int[l];
        for (int i = 0; i < l; i++ ) {
            dl[i] = convbyte(data[i]);
            System.out.println("dl["+i+"] = "+dl[i]);
        }
        return( Integer.toString(l) + "\n");*/
    }
}