#
#  grabs information dimensions out of outbox files
#

import sys
import os
import string

directory = "Outbox"

if len(sys.argv) > 1:
  directory = sys.argv[1]

files = os.listdir(directory)

#
# base class for all traversable objects, works in cooperation with XVisitor
# which implements the Visitor Pattern 
#

class XTraversable:
  def __init__(self):
    self.children = []
    self.parent = None

  def addChild(self,tag):
    self.children.append(tag)
    tag.setParent(self)

  def setParent(self,parent):
    self.parent = parent

  def getParent(self):
    return self.parent

  def traverse(self, visitor):
    if visitor.visit(self):
      for x in self.children:
        x.traverse(visitor)
    visitor.done()


#
# Visitor Pattern abstract base class
#

class XVisitor:
  def __init__(self):
    self.level = 0

  #
  # convention is that returning 1 means children get visited
  # otherwise children do not get visited
  #
  def visit(self,traversible):
    self.level = self.level + 1
    return(self.visitObject(traversible))
  
  #
  # slightly weird, but XVisitor and XTraversable must totally cooperate
  # here to make sure the level is set correctly.  
  #
  def done(self):
    self.level = self.level - 1

  # 
  # abstract base class method, does nothing, and asks that traversal stop
  #
  def visitObject(self,traversible):
    return 0



#
#  Very simple XML parsing, we don't actually match tags, just collect names & values
#  in a traversable hierarchy
#

class XTag(XTraversable):
  def __init__(self,name,value=''):
    XTraversable.__init__(self)
    self.name = name
    self.value = value

  def getName(self):
    return self.name

  def getValue(self):
    return self.value


#
#  Builds the hierarchy of XTags
#

class XParser:
  def __init__(self,root):
    self.root = root
    self.root.setParent(self.root)
    self.currentTag = self.root

  def parseLine(self,line):
    tags = string.split(line,"<")
    for x in tags:
      xtags = string.split(x,">")
      if len(xtags)==2:
        if xtags[0][0] == '/':
          self.currentTag = self.currentTag.getParent()
        elif xtags[0][0] != '?':
          newtag = XTag(xtags[0],xtags[1])
          self.currentTag.addChild(newtag)
          self.currentTag = newtag


#
#  DumpVisitor traverses tree, dumps out hierarchy with appropriate indentation
#  useful for debugging
#

class DumpVisitor(XVisitor):
  #
  #  nothing special needed here
  #
  def __init__(self):
    XVisitor.__init__(self)

  #
  #  print current node, dump all children (by visiting them)
  #
  def visitObject(self, traversible):
    print ' '*self.level,traversible.getName(), traversible.getValue()
    return 1

#
#  DataVisitor traverses tree, but only dumps out nodes that have values
#

class DataVisitor(XVisitor):
  #
  #  nothing special needed here
  #
  def __init__(self):
    XVisitor.__init__(self)

  #
  #  print current node, dump all children (by visiting them)
  #
  def visitObject(self, traversible):
    if len(traversible.getValue()):
      print ' '*self.level,traversible.getName(), traversible.getValue()
    return 1



#
#  VizObject collects information to pipe over to viz through a socket connection
#  This object does NOT communicate with viz, just collects the info needed from
#  each file
#

parsedFileList = []
for file in files:
  root = XTag("root")
  fp = XParser(root)
  parsedFileList.append(root)
  filetoparse = open(directory + '/' + file)
  for x in filetoparse.readlines():
    x = x[:-1]
    fp.parseLine(x)

for x in parsedFileList:
  dv = DataVisitor()
  x.traverse(dv)
